#
# Include file test_for_error
#
# This include file executes either a backup or restore of the
# backup_test database. It is designed to test for a specific
# error using either debug insertion or error condition insertion
# via the DBUG_EXECUTE_IF() macro.
#
# Prerequisites
#  (1) There exists a database named 'backup_test'
#  (2) There exists a backup image file named 'backup_test_orig.bak'
#
# Parameters (required)
#   $caseno      The test case number
#   $operation   'BACKUP'  = execute backup database w/o options
#                'RESTORE' = execute restore from
#   $errname     The name of the error. e.g. ER_SOME_ERROR
#   $errno       The numeric value of the error 
#                (see error_name_to_number.inc)
#   $compress    Use the compression option for backup
#                Valid values: ON or OFF
#   $suppress    Suppress error numbers in warning or error
#                messages on backup. "ON" = suppress
#                Valid values: ON or OFF
#

--echo #
--echo # Test case $caseno - testing error $errname
--echo #
--echo # Set debug SESSION variable for $errname
--echo #
eval SET SESSION debug="+d,$errname";

# Check to see operation is for backup
if (`SELECT STRCMP("$operation", "BACKUP") = 0`)
{
  --echo #
  --echo # Execute backup
  --echo #
  --disable_abort_on_error
  --replace_column 1 #
  --replace_regex /Default backup driver|MyISAM backup driver/DEBUG TEST backup driver/ /Failed to obtain metadata for grant .*/Failed to obtain metadata for grant 'Debug insertion test'./
  # Suppress error numbers if turned on.
  if (`SELECT STRCMP("$suppress", "ON") = 0`)
  {
    --replace_regex /[0-9]+/####/
  }
  # Use compression if turned on.
  if (`SELECT STRCMP("$compress", "ON") = 0`)
  {
    BACKUP DATABASE backup_test TO 'backup_test_err.bak' WITH COMPRESSION;
  }
  if (`SELECT STRCMP("$compress", "ON") <> 0`)
  {
    BACKUP DATABASE backup_test TO 'backup_test_err.bak';
  }
  --error 0,1
  remove_file $MYSQLD_BACKUPDIR/backup_test_err.bak;
  --enable_abort_on_error
}

# Check to see operation is for restore
if (`SELECT STRCMP("$operation", "RESTORE") = 0`)
{
  --echo #
  --echo # Execute restore
  --echo #
  --disable_abort_on_error
  --replace_column 1 # 2 #
  --replace_regex /Default restore driver|MyISAM restore driver/DEBUG TEST restore driver/ /Could not execute grant .*/Could not execute grant 'Debug insertion test'./
  RESTORE FROM 'backup_test_orig.bak' OVERWRITE;
  --enable_abort_on_error
}

let $my_err = $mysql_errno;

# If error number matches the error we are testing, the case passes.
if (`SELECT $my_err = $errno`)
{
  --echo #
  --echo # Test case for error $errname PASSED.
  --echo #

  if (`SELECT $errno = 0`)
  {
    --replace_regex /Default backup driver|MyISAM backup driver/DEBUG TEST backup driver/ /Default restore driver|MyISAM restore driver/DEBUG TEST restore driver/
    --replace_column 2 ####
    SHOW WARNINGS;
  }
  
  --echo #
  --echo # Show warning/error from progress and history log
  --echo # if they exist.
  --echo #

  # Mask out error number from select statement.
  --replace_regex /Default backup driver|MyISAM backup driver/DEBUG TEST backup driver/ /Default restore driver|MyISAM restore driver/DEBUG TEST restore driver/ /Could not execute grant .*/Could not execute grant 'Debug insertion test'./ /Failed to obtain metadata for grant .*/Failed to obtain metadata for grant 'Debug insertion test'./
  --replace_column 1 # 3 ####
  SELECT * FROM mysql.backup_progress WHERE error_num <> 0;

  PURGE BACKUP LOGS;
}

# If error number does not match, the test case has failed.
if (`SELECT $my_err <> $errno`)
{
  --echo #
  --echo # Test case for error $errname FAILED. Aborting test.
  --echo # Test returned error code $my_err.
  --echo #
  exit;
}

--echo #
--echo # Turn off debug SESSION.
--echo #
SET SESSION debug="-d";

--echo #
--echo # Now demonstrate that the command will work without the
--echo # debug session tag.

# Check to see operation is for backup
if (`SELECT STRCMP("$operation", "BACKUP") = 0`)
{
  --echo #
  --echo # Running backup - should not fail.
  --echo #
  --replace_column 1 #
  # Use compression if turned on.
  if (`SELECT STRCMP("$compress", "ON") = 0`)
  {
    BACKUP DATABASE backup_test TO 'backup_test_err.bak' WITH COMPRESSION;
  }
  if (`SELECT STRCMP("$compress", "ON") <> 0`)
  {
    BACKUP DATABASE backup_test TO 'backup_test_err.bak';
  }
  remove_file $MYSQLD_BACKUPDIR/backup_test_err.bak;
}

# Check to see operation is for restore
if (`SELECT STRCMP("$operation", "RESTORE") = 0`)
{
  --echo #
  --echo # Running restore - should not fail.
  --echo #
  --replace_column 1 # 2 #
  RESTORE FROM 'backup_test_orig.bak' OVERWRITE;
}
--echo #
--echo # Done.
--echo #
