#
# List various objects existing in the server and their properties.
#
# This file is used in backup_object_coverage.test to check if all objects
# were correctly restored.
#
# Note: Some properties are not listed because they are not preserved
# by backup/restore (BUG#44514/15). When these bugs are fixed, the 
# corresponding properties should be uncommented (assuming that they have
# to be preserved).
#

SELECT '# databases' AS '####';

USE information_schema;

SELECT
  CATALOG_NAME,
  SCHEMA_NAME,
  DEFAULT_CHARACTER_SET_NAME,
  DEFAULT_COLLATION_NAME,
  SQL_PATH
FROM schemata WHERE schema_name LIKE 'db%'
ORDER BY schema_name;

SELECT '# tables' AS '####';

SELECT 
  TABLE_CATALOG,
  TABLE_SCHEMA,
  TABLE_NAME,
  TABLE_TYPE,
  ENGINE,
  VERSION,
  ROW_FORMAT,
  TABLE_ROWS,
  AVG_ROW_LENGTH,
  DATA_LENGTH,
  MAX_DATA_LENGTH,
  INDEX_LENGTH,
  DATA_FREE,
  AUTO_INCREMENT,
#  CREATE_TIME,		# BUG#44514
#  UPDATE_TIME,		# BUG#44514
  CHECK_TIME,
  TABLE_COLLATION,
  CHECKSUM,
  CREATE_OPTIONS,
  TABLE_COMMENT,
  TABLESPACE_NAME
FROM tables WHERE table_schema LIKE 'db%'
ORDER BY table_schema, table_name;

SELECT '# table columns' AS '####';

SELECT 
  c.TABLE_CATALOG,
  c.TABLE_SCHEMA,
  c.TABLE_NAME,
  c.COLUMN_NAME,
  c.ORDINAL_POSITION,
  c.COLUMN_DEFAULT,
  c.IS_NULLABLE,
  c.DATA_TYPE,
  c.CHARACTER_MAXIMUM_LENGTH,
  c.CHARACTER_OCTET_LENGTH,
  c.NUMERIC_PRECISION,
  c.NUMERIC_SCALE,
  c.CHARACTER_SET_NAME,
  c.COLLATION_NAME,
  c.COLUMN_TYPE,
  c.COLUMN_KEY,
  c.EXTRA,
  c.PRIVILEGES,
  c.COLUMN_COMMENT,
  c.STORAGE,
  c.FORMAT
FROM columns as c, tables as t 
WHERE c.table_schema LIKE 'db%'
      AND c.table_schema = t.table_schema
      AND c.table_name = t.table_name
      AND t.table_type = 'BASE_TABLE'
ORDER BY table_schema, table_name, ordinal_position;

#
# In some cases column's default value is not correctly reported 
# for restored views: see BUG#45669.
# 
SELECT '# view columns' AS '####';

SELECT 
  c.TABLE_CATALOG,
  c.TABLE_SCHEMA,
  c.TABLE_NAME,
  c.COLUMN_NAME,
  c.ORDINAL_POSITION,
#  c.COLUMN_DEFAULT,    # BUG#45669
  c.IS_NULLABLE,
  c.DATA_TYPE,
  c.CHARACTER_MAXIMUM_LENGTH,
  c.CHARACTER_OCTET_LENGTH,
  c.NUMERIC_PRECISION,
  c.NUMERIC_SCALE,
  c.CHARACTER_SET_NAME,
  c.COLLATION_NAME,
  c.COLUMN_TYPE,
  c.COLUMN_KEY,
  c.EXTRA,
  c.PRIVILEGES,
  c.COLUMN_COMMENT,
  c.STORAGE,
  c.FORMAT
FROM columns as c, tables as t 
WHERE c.table_schema LIKE 'db%'
      AND c.table_schema = t.table_schema
      AND c.table_name = t.table_name
      AND t.table_type = 'VIEW'
ORDER BY table_schema, table_name, ordinal_position;

SELECT '# views' AS '####';

SELECT
  TABLE_CATALOG,
  TABLE_SCHEMA,
  TABLE_NAME,
  VIEW_DEFINITION,
  CHECK_OPTION,
  IS_UPDATABLE,
  DEFINER,
  SECURITY_TYPE,
  CHARACTER_SET_CLIENT,
  COLLATION_CONNECTION
FROM views WHERE table_schema LIKE 'db%'
ORDER BY table_schema, table_name;

SELECT '# stored routines' AS '####';

SELECT
  SPECIFIC_NAME,
  ROUTINE_CATALOG,
  ROUTINE_SCHEMA,
  ROUTINE_NAME,
  ROUTINE_TYPE,
  DATA_TYPE,
#  CHARACTER_MAXIMUM_LENGTH,	# BUG#44515
#  CHARACTER_OCTET_LENGTH,	# BUG#44515
  NUMERIC_PRECISION,
  NUMERIC_SCALE,
  CHARACTER_SET_NAME,
  COLLATION_NAME,
#  DTD_IDENTIFIER,	# BUG#44515
  ROUTINE_BODY,
  ROUTINE_DEFINITION,
  EXTERNAL_NAME,
  EXTERNAL_LANGUAGE,
  PARAMETER_STYLE,
  IS_DETERMINISTIC,
  SQL_DATA_ACCESS,
  SQL_PATH,
  SECURITY_TYPE,
#  CREATED,		# BUG#44514
#  LAST_ALTERED,	# BUG#44514
  SQL_MODE,
  ROUTINE_COMMENT,
  DEFINER,
  CHARACTER_SET_CLIENT,
  COLLATION_CONNECTION
  DATABASE_COLLATION
FROM routines WHERE routine_schema LIKE 'db%'
ORDER BY routine_schema, routine_name;

SELECT '# events' AS '####';

SELECT
  EVENT_CATALOG,
  EVENT_SCHEMA,
  EVENT_NAME,
  DEFINER,
  TIME_ZONE,
  EVENT_BODY,
  EVENT_DEFINITION,
  EVENT_TYPE,
  EXECUTE_AT,
  INTERVAL_VALUE,
  INTERVAL_FIELD,
  SQL_MODE,
  STARTS,
  ENDS,
  STATUS,
  ON_COMPLETION,
#  CREATED,		# BUG#44514
#  LAST_ALTERED,	# BUG#44514
  LAST_EXECUTED,
  EVENT_COMMENT,
  ORIGINATOR,
  CHARACTER_SET_CLIENT,
  COLLATION_CONNECTION
  DATABASE_COLLATION
FROM events WHERE event_schema LIKE 'db%'
ORDER BY event_schema, event_name;

SELECT '# triggers' AS '###';

SELECT
  TRIGGER_CATALOG,
  TRIGGER_SCHEMA,
  TRIGGER_NAME,
  EVENT_MANIPULATION,
  EVENT_OBJECT_CATALOG,
  EVENT_OBJECT_SCHEMA,
  EVENT_OBJECT_TABLE,
  ACTION_ORDER,
  ACTION_CONDITION,
  ACTION_STATEMENT,
  ACTION_ORIENTATION,
  ACTION_TIMING,
  ACTION_REFERENCE_OLD_TABLE,
  ACTION_REFERENCE_NEW_TABLE,
  ACTION_REFERENCE_OLD_ROW,
  ACTION_REFERENCE_NEW_ROW,
  CREATED,
  SQL_MODE,
  DEFINER,
  CHARACTER_SET_CLIENT,
  COLLATION_CONNECTION
  DATABASE_COLLATION
FROM triggers WHERE trigger_schema LIKE 'db%'
ORDER BY trigger_schema, trigger_name;


SELECT '# privileges' AS '####';

SELECT
  GRANTEE,
  TABLE_CATALOG,
  TABLE_SCHEMA,
  PRIVILEGE_TYPE,
  IS_GRANTABLE
FROM schema_privileges WHERE table_schema LIKE 'db%'
ORDER BY grantee, table_schema, privilege_type;

SELECT
  GRANTEE,
  TABLE_CATALOG,
  TABLE_SCHEMA,
  TABLE_NAME,
  PRIVILEGE_TYPE,
  IS_GRANTABLE
FROM table_privileges WHERE table_schema LIKE 'db%'
ORDER BY grantee, table_schema, table_name, privilege_type;

SELECT
  GRANTEE,
  TABLE_CATALOG,
  TABLE_SCHEMA,
  TABLE_NAME,
  COLUMN_NAME,
  PRIVILEGE_TYPE,
  IS_GRANTABLE
FROM column_privileges WHERE table_schema LIKE 'db%'
ORDER BY grantee, table_schema, table_name, column_name, privilege_type;

SELECT '# Keys & Constraints' AS '####';

SELECT
  CONSTRAINT_CATALOG,
  CONSTRAINT_SCHEMA,
  CONSTRAINT_NAME,
  TABLE_SCHEMA,
  TABLE_NAME,
  CONSTRAINT_TYPE
FROM table_constraints WHERE table_schema LIKE 'db%'
ORDER BY table_schema, table_name, constraint_name;

SELECT
  CONSTRAINT_CATALOG,
  CONSTRAINT_SCHEMA,
  CONSTRAINT_NAME,
  TABLE_CATALOG,
  TABLE_SCHEMA,
  TABLE_NAME,
  COLUMN_NAME,
  ORDINAL_POSITION,
  POSITION_IN_UNIQUE_CONSTRAINT,
  REFERENCED_TABLE_SCHEMA,
  REFERENCED_TABLE_NAME,
  REFERENCED_COLUMN_NAME
FROM key_column_usage WHERE table_schema LIKE 'db%'
ORDER BY table_schema, table_name, column_name, constraint_name, ordinal_position;

SELECT
  CONSTRAINT_CATALOG,
  CONSTRAINT_SCHEMA,
  CONSTRAINT_NAME,
  UNIQUE_CONSTRAINT_CATALOG,
  UNIQUE_CONSTRAINT_SCHEMA,
  UNIQUE_CONSTRAINT_NAME,
  MATCH_OPTION,
  UPDATE_RULE,
  DELETE_RULE,
  TABLE_NAME,
  REFERENCED_TABLE_NAME
FROM referential_constraints WHERE constraint_schema LIKE 'db%'
ORDER BY unique_constraint_schema, unique_constraint_name;

