#!/usr/bin/perl 

use Getopt::Long;
use POSIX;
use Cwd qw ( cwd abs_path );
use File::Basename;
use File::Copy;
use File::Path;
use File::Spec;
# Needed by platform independent diff_dumps(), but is not available in PB
#use Text::Diff; 

my $mtr_build_thread_def=74;
my $mtr_build_thread_option; # May be set by CL option
my $start_dir = cwd();
my $ext=".exe" if windows();

my $stress_duration = 300;  # Transactional load runs for 5 minutes
my $stress_threads = 2;     # Threads used in mysql-stress-test.pl
                            # Will be increased significantly when
                            # BUG#44568 has been fixed
                            # Can be set by CL option
my $stress_pid = 0;         # Used when killing mysql-stress-test.pl 

my $datadump_dir = "";
my $schema_file = "";
my $mysql_dir = "../";      # May be changed by CL option
my $client_dir;

my $server_host = "127.0.0.1";
my $server_port = 0;        # Port the main MySQL server listens to
my $slave_port = 0;         # Port the server that replays binlog listens to

# Binlog position of the backup image
my $backup_binlog_file = "";
my $backup_binlog_pos = 0;

my $tx_isolation_level;     # Transaction isolation level for BACKUP thread

# Store command line options
my $opt_result = GetOptions(
        'mysql-dir=s' => \$mysql_dir,
        'datadump-dir=s' => \$datadump_dir,
        'schema-file=s' => \$schema_file,
        'stress-threads=s' => \$stress_threads,
        'mtr-build-thread=s' => \$mtr_build_thread_option,
        'tx-isolation-level=s' => \$tx_isolation_level
);

# Expand all paths to absolute path
$datadump_dir= abs_path($datadump_dir)."/";
$mysql_dir= abs_path($mysql_dir)."/";
$start_dir= abs_path($start_dir)."/";

if (!$opt_result) {
    print_usage("");
    exit(1);
}

check_options();

my $test_dir=$mysql_dir."/mysql-test/";
my $master_dbdir=$test_dir."var/mysqld.1/data/";
my $slave_dbdir=$test_dir."var/mysqld.2/data/";
my $vptest_home="$test_dir/suite/backup_extra/vptest";

my $stress_suitedir= "$test_dir/suite/backup_extra/flightstats_stress/";
my $stress_basedir= "$test_dir/suite/backup_extra/stress_basedir";
my $stress_init_file = $stress_suitedir."init_flightstat_stress.txt";
my $stress_tests_file = $stress_suitedir."run_flightstat_stress.txt";
my $tmpfile_dir="$vptest_home/tmp/";
my $alter_file="$vptest_home/flightstats_alter.sql";

my $backup_file="flight.bak";
my $dump_file_orig=$tmpfile_dir."dump_orig_db.sql";
my $dump_file_restore=$tmpfile_dir."dump_restored_db.sql";
my $diff_file=$tmpfile_dir."dump.diff";
my $sql_outfile=$tmpfile_dir."sql_out.txt";
my $server_outfile=$tmpfile_dir."server_out.txt";

# Make sure version 2 of MTR is used
$ENV{MTR_VERSION} = 2;
# Setting the MTR_BUILD_THREAD env variable. 
if ($mtr_build_thread_option) {
    $ENV{MTR_BUILD_THREAD} = $mtr_build_thread_option;
} elsif (!$ENV{MTR_BUILD_THREAD}) {
    $ENV{MTR_BUILD_THREAD} = $mtr_build_thread_def;
}

$server_port=10*$ENV{MTR_BUILD_THREAD}+10000;
$slave_port=$server_port+1;

my $dbh;

# ***********
# Main script
# ***********

# ---------------------------
# SCRIPT PART I: Preparations
# ---------------------------

# Remove old contents in dump directories
# Note: Some Perl installations do not honor the keep_root flag and
# deletes the root directory as well. A workaround is to 
# recreate the directory instead
# rmtree(glob("$tmpfile_dir"), {keep_root => 1});
# rmtree(glob("$stress_basedir"), {keep_root => 1});
rmtree(glob("$tmpfile_dir"));
rmtree(glob("$stress_basedir"));
mkpath(glob("$tmpfile_dir"), {verbose => 1});
mkpath(glob("$stress_basedir"), {verbose => 1});

# Start two server instances. 
# * Main server listens to port $server_port
# * Slave server listens to port $slave_port 
# Note that replication is not started, but the servers are ready to
# go to replication mode (e.g., binlogging is enabled on main server)
start_server("");

# Create a database on the main server
create_database();

# Copy database to slave database location
create_slave_copy();
sleep(10); # Make sure servers are actually started.

# ---------------------------------------------------------------------
# SCRIPT PART II: Backup database on main server during concurrent load 
# ---------------------------------------------------------------------

# Start load on main server - runs for stress_duration seconds
start_stress_thread(); 
my $time_startstress=time();

# Backup database content on main server
sleep(30); # make sure stress test threads have started
backup();
my $time_stress_left=$stress_duration -
                     (time()-$time_startstress);

# Not platform independent - instead sleep until stress test threads exit
#Make sure some transactions execute after BACKUP has been performed.
#These should be gone after RESTORE 
#stop_stress_thread();

print("Now waiting ".$time_stress_left.
      " sec until the stress test threads exit\n");
sleep($time_stress_left+10); 

# ----------------------------------------------
# SCRIPT PART III: Replay binlog on slave server
# ----------------------------------------------

# Get binlog info from main server
get_binlog_info();
# Redo binlog on the slave server
redo_binlog_to_db_copy();

# -------------------------------------------------------
# SCRIPT PART IV: Compare images on main and slave server
# -------------------------------------------------------

# Drop existing database on main server and restore it
restore();

# Compare the tables on the two servers and dump tables that are not equal
if (!compare_checksums()) {
    print ("####################################\n");
    print ("# The two databases are NOT equal   \n");
    print ("# Dumping databases to file         \n");
    print ("####################################\n");
    dump_databases();
    cleanup();
    exit(1);
}

cleanup();

print ("####################################\n");
print ("#  TEST COMPLETED SUCCESSFULLY     #\n");
print ("####################################\n");

exit(0);

# ***********
# Subroutines
# ***********

sub dump_databases {

    # Make sure that AUTO_INC in both db versions are equal. Otherwise
    # diff of the dump files will fail
    set_autoincrement($slave_port, 100000);
    # Database on slave server can be dumped - binlog has been replayed
    dump_database($slave_port, $dump_file_orig);

    # Make sure that AUTO_INC in both db versions are equal. Otherwise
    # diff of the dump files will fail
    set_autoincrement($server_port, 100000);
    dump_database($server_port, $dump_file_restore);
}

sub compare_checksums {

    my $tables_are_equal=TRUE;

    my $checksum_restore_db="";
    my $checksum_redobinlog_db="";

    my @tables = ("aircraft",
                  "aircraft_engine_types",
                  "aircraft_engines",
                  "aircraft_models",
                  "aircraft_types",
                  "airport_remarks",
                  "airports",
                  "carriers",
                  "exceptions",
                  "states",
                  "zipcodes",
                  "ontime_1week");

    my $query;
    for $table (@tables) {
        $query = "CHECKSUM TABLE FlightStats_test.".$table;
        $checksum_restore_db = execute_sql($server_port, $query);
        $checksum_redobinlog_db = execute_sql($slave_port, $query);

        @restore_tokens = split("\t",$checksum_restore_db);
        @redobinlog_tokens = split("\t",$checksum_redobinlog_db);

        $checksum_restore_db = rtrim(@restore_tokens[-1]);
        $checksum_redobinlog_db = rtrim(@redobinlog_tokens[-1]);

        if ($checksum_restore_db == $checksum_redobinlog_db) {
            print ("Checksums for table '".$table."': OK\n");
        } else {
            print ("Checksums for table '".$table."': NOT OK\n");
            $tables_are_equal=FALSE;
        }
    }
    return $tables_are_equal;
}

sub print_usage {

    my ($message) =@_;
    print($message);
    print("Usage:\n");
    print("  perl run_test.pl --datadump-dir=PATH [OPTIONS]\n");
    print("Options:\n");
    print("  mtr-build-thread=INT (default: Value of env variable MTR_BUILD_THREAD, \n".
          "                        74 if env variable has not been set)\n");
    print("  mysql-dir=PATH       (default: '../')\n");
    print("  schema-file=FILE     (default: DATADUMP_DIR/flightstats_schema.sql)\n");
    print("  stress-threads=INT   (default: 2)\n");
    print("  tx-isolation-level=LVL (default: NONE)\n");
    print("Example:\n");
    print("  perl run_test.pl --datadump-dir=/mysql/flightstat/datadump/\n");
    print("  perl run_test.pl --datadump-dir=/mysql/flightstat/datadump/ \\ \n".
          "                   --mysql-dir=/mysql/mysql-5.4 \\ \n".
          "                   --mtr-build-thread=50 \\ \n".
          "                   --schema-file=/mysql/my_schema.sql \\ \n".
          "                   --stress-threads=20\n");
}

sub cleanup {
    shutdown_server();
}

sub diff_dumps {
    $ret_code= system("diff $dump_file_orig $dump_file_restore >$diff_file");
    if ($ret_code >> 8) {

        print ("####################################\n");
        print ("# The two dump files are NOT equal  \n");
        print ("# Writing diff to dump.diff         \n");
        print ("####################################\n");

        cleanup();
        exit(1);
    }
}

# Todo: Replace above function with below function when Text::Diff is
# available in PB. This will remove platform dependence (*nix)
# sub diff_dumps {
#     my $diff = diff($dump_file_orig, $dump_file_restore); 
#     if (!($diff eq "")) { 
#         print ("####################################\n");
#         print ("# The two dump files are NOT equal  \n");
#         print ("# Writing diff to dump.diff         \n");
#         print ("####################################\n");
#         open FILE, ">".$diff_file or die $!;
#         print FILE $diff;
#         close FILE;
#         cleanup();
#         exit(1);
#     }
# }

sub restore {
    chdir $test_dir;
    
    print ("Restoring database...");

    execute_sql($server_port, "DROP DATABASE FlightStats_test");
    execute_sql($server_port, "RESTORE FROM \'".$backup_file."\'");
    print ("done\n");
}

sub dump_database {
    my ($port,$dumpfile) =@_;

    print ("mysqldump from server on port ".$port." to file ".$dumpfile."...");

    unlink($dumpfile);

    my $dump_args= "--user=root --port=".$port." --protocol=tcp --compact ".
        "--order-by-primary --skip-extended-insert FlightStats_test > "
        .$dumpfile;

    $ret_code=system($client_dir."/mysqldump ".$dump_args);
    if ($ret_code >> 8) { 
        print("\n Failed to dump database - Aborting\n");
        cleanup();
        exit(1);
    }
    print ("done\n");
}

sub redo_binlog_to_db_copy {

    print ("Playing forward binlog to copy of original database...");

    my $cmd = 'CHANGE MASTER TO MASTER_PORT='.$server_port.
              ', MASTER_USER=\'root\', MASTER_HOST=\''.$server_host.'\'';
    execute_sql($slave_port, $cmd);

    $cmd ="START SLAVE UNTIL MASTER_LOG_FILE=\'".$backup_binlog_file.
          "\', MASTER_LOG_POS=".$backup_binlog_pos;

    execute_sql($slave_port, $cmd);

    # Wait until slave has reached requested binlog position
    $cmd ="select MASTER_POS_WAIT(\'".$backup_binlog_file.
          "\', \'".$backup_binlog_pos."\')";
    execute_sql($slave_port, $cmd);

    execute_sql($slave_port, "STOP SLAVE");

    # execute_sql($slave_port, "SHOW SLAVE STATUS");
    # execute_sql($slave_port, "SELECT COUNT(*) FROM FlightStats_test.aircraft");
    sleep(5);
    print("done\n");
}

sub get_binlog_info {

    print ("Getting binlog info...");
    my $binlog_info = execute_sql($server_port,
                                  "SELECT binlog_start_pos, binlog_file ".
                                  "FROM mysql.backup_history LIMIT 1");

    @tokens = split("\t",$binlog_info);

    ($backup_binlog_pos,$backup_binlog_file)=@tokens;

    $backup_binlog_file =rtrim($backup_binlog_file);

    print ("done\n");
}

# Right trim function to remove trailing whitespace
sub rtrim($)
{
	my $string = shift;
	$string =~ s/\s+$//;
	return $string;
}

sub backup {
    chdir $test_dir;
    
    print ("Backing up database...");

    unlink ($backup_file); #Remove file if already exists
    execute_sql($server_port,"PURGE BACKUP LOGS");
    execute_sql($server_port, 
                "SET TRANSACTION ISOLATION LEVEL $tx_isolation_level") 
                                                     if $tx_isolation_level;

    my $time_before=time();
    execute_sql($server_port,
                "BACKUP DATABASE FlightStats_test TO \'".$backup_file."\'");
    my $time_after=time();
    my $difference=$time_after-$time_before;
    print ("BACKUP took: ".$difference." sec...");
    print ("done\n");
}

# Note: 
# Since perl kill() does not work as expected, there is no platform
# independent way to kill the stress test threads. Consequence: must
# wait until the stress test threads exit.
sub stop_stress_thread {
    # Wait for stress test to complete
    print("Stopping stress test...");

    # TODO: Replace with kill(SIGNAL, PID) 
    #   Problem: kill(x, $stress_pid) only kills one thread (x=1 or 9 or 15)
    # system ("kill $stress_pid"); #only kills one of the stress test threads
    system("killall mysql-stress-test.pl"); # works on Linux
    #Wait until PID of mysql_stress_test no longer exists
    waitpid($stress_pid, 0); 
    print ("done\n");
}

sub start_stress_thread {

    my $pid = fork();
    if ($pid == 0) {
        # This is the stress test thread
        run_stress();
        exit(0);
    } else {
        # This is the main thread
        $stress_pid= $pid;
        # Make sure the stress test has started
        sleep(10);
    }
    return 0;
}

sub run_stress {
    chdir $test_dir;

    print("STARTING STRESS TEST\n");
    my $stress_param ="--stress-basedir=$stress_basedir ".
        "--stress-suite-basedir=$stress_suitedir ".
        "--server-logs-dir=$stress_basedir ".
        "--stress-init-file=$stress_init_file ".
        "--stress-tests-file=$stress_tests_file --server-user=root ".
        "--server-port=$server_port --server-host=$server_host ".
        "--mysqltest=$client_dir/mysqltest --test-duration=$stress_duration ".
        "--threads=$stress_threads > $stress_basedir/run.out 2>&1";

    ## If exec is replaced with system, kill must be changed because stress_pid is wrong
    exec("./mysql-stress-test.pl ".$stress_param); 
    print("STRESS TEST DONE\n"); # Not printed because 'exec' never returns

}    

# TODO: Replace dircopy() with fcopy if Perl::Copy::Recursive is made
# available in PB2: "fcopy($master_dbdir/,$slave_dbdir/) or die $!;"
sub create_slave_copy {
    shutdown_server();

    print("Delete existing slave data files\n");    
    # Note: Some Perl installations do not honor the keep_root flag and
    # deletes the root directory as well. A workaround is to 
    # recreate the directory instead
    #rmtree(glob("$slave_dbdir/"),  {keep_root => 1});
    rmtree(glob("$slave_dbdir/"));
    mkpath(glob("$slave_dbdir/"), {verbose => 1});
    
    print("Copying database to slave location\n");    

    dircopy($master_dbdir, $slave_dbdir);
    
    start_server("--start-dirty");
    return 0;
}

sub dircopy {
    my ($from, $to) =@_;
    opendir my($handle), $from or die "Could not open dir '$from': $!";
    for my $file (readdir $handle) {
        next if ($file eq '.') or ($file eq '..');
        my $source = "$from/$file";
        my $destination = "$to/$file";
        if (-d $source) {
            mkdir $destination or die "mkdir '$destination' failed: $!";
            dircopy($source, $destination);
        } else {
            copy($source, $destination) or die $!;
        }
    }
    closedir $handle;
    return 0;
}

sub shutdown_server {
    print("Shutdown server\n");
    $mysql_args="--user=root --protocol=tcp --port=".$server_port;
    system($client_dir."/mysqladmin ".$mysql_args." shutdown");

    $mysql_args="--user=root --protocol=tcp --port=".$slave_port;
    system($client_dir."/mysqladmin ".$mysql_args." shutdown");

    sleep(10);
    return 0;
}

sub create_database {

    # Creating flightstats_test dabase and insert data
    # ------------------------------------------------
    $mysql_args="--user=root --protocol=tcp --port=".$server_port;

    print("Creating database...");

    execute_sql($server_port, "DROP DATABASE IF EXISTS FlightStats_test");
    execute_sql($server_port, "CREATE DATABASE FlightStats_test");
    
    $ret_code=system($client_dir."/mysql ".$mysql_args." < ".$schema_file);
    if ($ret_code >> 8) { 
        print("\nFailed to create database - Aborting\n");
        cleanup();
        exit(1);
    }

    $ret_code=system($client_dir."/mysql ".$mysql_args." < ".$alter_file);
    if ($ret_code >> 8) { 
        print("\n Failed to alter database - Aborting\n");
        cleanup();
        exit(1);
    }
    print("done\n");

    # Load data from dump files
    # -------------------------

    my $pre_sql="load data infile '".$datadump_dir;
    my $post_sql=" fields terminated by '\\t' lines terminated by '\\n';\n";

    my $sql=$sql."use FlightStats_test;\n";
    $sql=$sql.$pre_sql."aircraft.dat' into table aircraft".$post_sql;
    $sql=$sql.$pre_sql."aircraft_engine_types.dat' ".
        "into table aircraft_engine_types".$post_sql;
    $sql=$sql.$pre_sql."aircraft_engines.dat' into table aircraft_engines".
        $post_sql;
    $sql=$sql.$pre_sql."aircraft_models.dat' into table aircraft_models".
        $post_sql;
    $sql=$sql.$pre_sql."aircraft_types.dat' into table aircraft_types".
        $post_sql;
    $sql=$sql.$pre_sql."airport_remarks.dat' into table airport_remarks".
        $post_sql;
    $sql=$sql.$pre_sql."airports.dat' into table airports".$post_sql;
    $sql=$sql.$pre_sql."carriers.dat' into table carriers".$post_sql;
    $sql=$sql.$pre_sql."exceptions.dat' into table exceptions".$post_sql;
    $sql=$sql.$pre_sql."states.dat' into table states".$post_sql;
    $sql=$sql.$pre_sql."zipcodes.dat' into table zipcodes".$post_sql;
    $sql=$sql.$pre_sql."ontime_1week.dat' into table ontime_1week".$post_sql;
    # ontime_1day can be used if fewer records are wanted
    # $sql=$sql.$pre_sql."ontime_1day.dat' into table ontime_1week".$post_sql;

    print("Loading data...");
    execute_sql($server_port,$sql);
    print("done\n");

    return 0;
}

sub execute_sql {
    my($port,$sql) =@_;

    my $do_sql = "END_OF_SQL\n".$sql."\nEND_OF_SQL\n";
    $mysql_args="--skip-column-names --user=root --protocol=tcp --port=".$port;
    my $command = $client_dir."/mysql ".$mysql_args.
        " << ".$do_sql;

    open(OUT, $command."|") or die "Failed to open pipe";
    my $output;
    while ($_ = <OUT>) {
        $output .= $_;
    }
    close(OUT);

    ## Print SQL and it's output to file so you can see what happens 
    open FILE, ">>".$sql_outfile;
    print FILE $command."\n";
    print FILE $output."\n";
    close FILE;
    
    return $output;
}

sub start_server {
    chdir($test_dir);

    # setting the command line arguments for mysql-test-run.pl
    my ($arg)=@_;
    print("Starting server with additional argument: ".$arg."\n");

    my $mysql_args="--start --mysqld=--secure-file-priv= ".
                   "--mysqld=--innodb --mysqld=--log-output=file ".
                   "--mysqld=--default-storage-engine=innodb ".
                   "--mysqld=--mysql-backup ".
                   $arg." start_masterslave --suite=backup_extra";

    my $server_start_script = File::Spec->catfile( abs_path( $mysql_dir ), 
                                                   'mysql-test', 
                                                   'mysql-test-run.pl' );

    my $pid = fork();
    if ($pid == 0) {
        # This is the thread that starts the servers;
        # start_masterslave starts both servers
        chdir($test_dir);
        $ret_code=system("perl $server_start_script ".$mysql_args.
                         ">> ".$server_outfile." 2>&1");
        exit(0);
    } else {
        # This is the main thread
        # Wait until the servers have started
        sleep(10);
    }

    # Let main thread check that the servers started without error
    if ($ret_code >> 8) { 
        print("\n Failed to start the Server - Aborting\n");
        cleanup();
        exit(1);
    }
}

sub check_options {
    # Check if value of base_dir and mysql_dir are valid
    if (-e $mysql_dir."/bin/mysql".$ext) {
        $client_dir = $mysql_dir."/bin";
    } elsif (-e $mysql_dir."/client/mysql".$ext) {
        $client_dir = $mysql_dir."/client";
    } else { 
        print_usage("Could not find MySQL in: \'".$mysql_dir."\'\n");
        exit(1);
    }

    if (! -e $datadump_dir."/aircraft.dat") {
        print_usage("Could not find test data in: \'".$datadump_dir.
                    "\'\nTEST WAS SKIPPED\n");
        exit(0);
    }

    if (!$schema_file) {
        if (! -e $datadump_dir."/flightstats_schema.sql") {    
            print_usage("Could not find schema file \'".$datadump_dir.
                        "/flightstats_schema.sql\'\n".
                        "Copy the file to this location or specify schema ".
                        "file using --schema-file\n");
            exit(1);
        } else {
            $schema_file=$datadump_dir."/flightstats_schema.sql";
        }
    } elsif (! -e $schema_file) {
        print_usage("Could not find schema file \'".$schema_file."\'\n");
        exit(1);
    }

    if (! -e $datadump_dir."/aircraft.dat") {
        print_usage("Could not find test data in: \'".$datadump_dir."\'\n");
        exit(1);
    }

    # Set correct isolation level for BACKUP thread
    
    if ($tx_isolation_level =~ /uncom/i) {
      	$tx_isolation_level= 'READ UNCOMMITTED';
    } elsif ($tx_isolation_level =~ /comm/i) {
      	$tx_isolation_level= 'READ COMMITTED';
    } elsif ($tx_isolation_level =~ /repeat/i) {
      	$tx_isolation_level= 'REPEATABLE READ';
    } elsif ($tx_isolation_level =~ /serial/i) {
      	$tx_isolation_level= 'SERIALIZABLE';
    } elsif ($tx_isolation_level) {
    	print("Unrecognized transaction isolation level '$tx_isolation_level'".
    	      " - ignoring\n");
      	$tx_isolation_level= undef;
    }  
}

sub set_autoincrement {
    my ($port,$autoinc_value)=@_;
    execute_sql($port, "ALTER TABLE FlightStats_test.ontime_all ".
                       "AUTO_INCREMENT=".$autoinc_value);
}

sub windows {
   if ( ($^O eq 'MSWin32') || 
        ($^O eq 'MSWin64') ||
        ($^O eq 'cygwin' )
      ) { return 1; } else { return 0; }
}

sub exit_if_not_nix {
    if (($^O eq 'solaris') || ($^O eq 'linux' )) {
        print ("Operating system \'".$^O."\' found. ".
              "Test can run on this platform\n");
       return 1; 
    } else  {
        print ("Operating system \'".$^O."\' found. ".
               "Test can NOT run on this platform\n");
        exit(0);
    }
}

