/* Copyright (C) 2000 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "mysys_priv.h"
#include <m_string.h>

/*
  Formats a filename with possible replace of directory of extension
  Function can handle the case where 'to' == 'name'
  For a description of the flag values, consult my_sys.h
  The arguments should be in unix format.

  Warning: The flag MY_UNPACK_FILENAME is evil. It makes a nice, clean
  path with no ./ and ../. But if a path element is a symbolic link and
  sufficient ../ are present to "step backwards" over it, the resulting
  path will be based on the symlink itself, not on its target:
    Assume symlink: /path/with/symlink -> /target/place
    fn_format(..., /path/with/symlink/../myfile, ...MY_UNPACK_FILENAME)
    results in /path/with/myfile
    OS functions would affect /target/myfile with the original path.
  MY_RETURN_REAL_PATH does a better job, but returns all symlinks
  resolved, which is normally not wanted. Also, it does not resolve
  the path, if the final target does not exist. Also, it cuts the
  resulting path name to FN_REFLEN if it is longer than that. Also,
  on Windows it does not clean the path.

  A correct algorithm has been tried with safe_cleanup_cat_path() in
  mf_pack.c. It calls my_is_symlink() on each path element from the
  beginning. If the element is not a symlink, it can be cleaned out. If
  a symlink is encountered, the part of the path behind it is treated as
  a relative path in itself. It is cleaned in itself like the above, but
  my_is_symlink() takes the concatenated previous path segments
  prepended. If another symlink is found, the algorithm repeats. That
  way we might end up with no ../ except possibly after symlinks:
    Input:  /path/with/symlink/up/../../yet/another/symlink/up/../end
    Output: /path/with/symlink/../yet/another/symlink/end
*/

char * fn_format(char * to, const char *name, const char *dir,
		    const char *extension, uint flag)
{
  char dev[FN_REFLEN], buff[FN_REFLEN], *pos, *startpos;
  const char *ext;
  reg1 size_t length;
  size_t dev_length;
  DBUG_ENTER("fn_format");
  DBUG_PRINT("enter",("name: %s  dir: %s  extension: %s  flag: %d",
		       name,dir,extension,flag));

  /* Copy and skip directory */
  name+=(length=dirname_part(dev, (startpos=(char *) name), &dev_length));
  if (length == 0 || (flag & MY_REPLACE_DIR))
  {
    /* Use given directory */
    convert_dirname(dev,dir,NullS);		/* Fix to this OS */
  }
  else if ((flag & MY_RELATIVE_PATH) && !test_if_hard_path(dev))
  {
    /* Put 'dir' before the given path */
    strmake(buff,dev,sizeof(buff)-1);
    pos=convert_dirname(dev,dir,NullS);
    strmake(pos,buff,sizeof(buff)-1- (int) (pos-dev));
  }

  if (flag & MY_PACK_FILENAME)
    pack_dirname(dev,dev);			/* Put in ./.. and ~/.. */
  if (flag & MY_UNPACK_FILENAME)
    (void) unpack_dirname(dev,dev);		/* Replace ~/.. with dir */

  if (!(flag & MY_APPEND_EXT) &&
      (pos= (char*) strchr(name,FN_EXTCHAR)) != NullS)
  {
    if ((flag & MY_REPLACE_EXT) == 0)		/* If we should keep old ext */
    {
      length=strlength(name);			/* Use old extension */
      ext = "";
    }
    else
    {
      length= (size_t) (pos-(char*) name);	/* Change extension */
      ext= extension;
    }
  }
  else
  {
    length=strlength(name);			/* No ext, use the now one */
    ext=extension;
  }

  if (strlen(dev)+length+strlen(ext) >= FN_REFLEN || length >= FN_LEN )
  {
    /* To long path, return original or NULL */
    size_t tmp_length;
    if (flag & MY_SAFE_PATH)
      DBUG_RETURN(NullS);
    tmp_length= strlength(startpos);
    DBUG_PRINT("error",("dev: '%s'  ext: '%s'  length: %u",dev,ext,
                        (uint) length));
    (void) strmake(to,startpos,min(tmp_length,FN_REFLEN-1));
  }
  else
  {
    if (to == startpos)
    {
      bmove(buff,(uchar*) name,length);		/* Save name for last copy */
      name=buff;
    }
    pos=strmake(strmov(to,dev),name,length);
    (void) strmov(pos,ext);			/* Don't convert extension */
  }
  /*
    If MY_RETURN_REAL_PATH and MY_RESOLVE_SYMLINK is given, only do
    realpath if the file is a symbolic link
  */
  if (flag & MY_RETURN_REAL_PATH)
    (void) my_realpath(to, to, MYF(flag & MY_RESOLVE_SYMLINKS ?
				   MY_RESOLVE_LINK: 0));
  else if (flag & MY_RESOLVE_SYMLINKS)
  {
    strmov(buff,to);
    (void) my_readlink(to, buff, MYF(0));
  }
  DBUG_RETURN(to);
} /* fn_format */


/*
  strlength(const string str)
  Return length of string with end-space:s not counted.
*/

size_t strlength(const char *str)
{
  reg1 const char * pos;
  reg2 const char * found;
  DBUG_ENTER("strlength");

  pos= found= str;

  while (*pos)
  {
    if (*pos != ' ')
    {
      while (*++pos && *pos != ' ') {};
      if (!*pos)
      {
	found=pos;			/* String ends here */
	break;
      }
    }
    found=pos;
    while (*++pos == ' ') {};
  }
  DBUG_RETURN((size_t) (found - str));
} /* strlength */


/**
  Cut path to a printable length.

  This can be used to put a path name into error/warning/info messages,
  which contain a conversion flag with length limit like %.64s.

  If 'path' is longer than 'length_limit', the end of the path is
  returned with prepended "... ". Often the end of a path is of
  more interest in error messages than the begin.

  @note A pointer to static storage is returned. It is assumed that
  only one path per error message needs to be cut.

  @param[in]    path            path name
  @param[in]    length_limit    length limit from format string

  @return       pointer to static string
*/

char *cut_print_path(const char* path, size_t length_limit)
{
  static char path_buff[FN_REFLEN];
  size_t plen= strlen(path);
  size_t offset= 0;

  if (length_limit >= FN_REFLEN)
    length_limit= FN_REFLEN - 1; /* Room for '\0'. */ /* purecov: inspected */

  if (plen > length_limit)
  {
    offset= plen - length_limit;
    plen= length_limit;
  }

  memcpy(path_buff, path + offset, plen + 1); /* Include '\0'. */

  if (offset)
  {
    path_buff[0]= '.';
    path_buff[1]= '.';
    path_buff[2]= '.';
    path_buff[3]= ' ';
  }

  return path_buff;
}

