#ifndef _BACKUP_AUX_H
#define _BACKUP_AUX_H

/**
  @file

  @brief Auxiliary declarations used in MySQL backup code.
*/

/// Definition for storage engine reference type.
typedef st_plugin_int* storage_engine_ref;

// Macro which transforms plugin_ref to storage_engine_ref.
#ifdef DBUG_OFF
/// Macro to map plugin_ref to se_ref.
#define plugin_ref_to_se_ref(A) (A)
/// Macro to map se_ref ro plugin_ref.
#define se_ref_to_plugin_ref(A) (A)
#else
/// Macro to map plugin_ref to se_ref.
#define plugin_ref_to_se_ref(A) ((A) ? *(A) : NULL)
/// Macro to map se_ref ro plugin_ref.
#define se_ref_to_plugin_ref(A) &(A)
#endif


/// Return the storage engine name.

inline
const char* se_name(storage_engine_ref se)
{ return se->name.str; }


/// Return the version of the plugin.

inline
uint se_version(storage_engine_ref se)
{ return se->plugin->version; }  // Q: Or, should it be A->plugin_dl->version?


/// Return the pointer to the handlerton.

inline
handlerton* se_hton(storage_engine_ref se)
{ 
  if (se)
    return (handlerton*)(se->data);
  else 
    return NULL;
}


/// Return a storage engine reference by name.

inline
storage_engine_ref get_se_by_name(const LEX_STRING name)
{
  plugin_ref plugin= ::ha_resolve_by_name(::current_thd, &name);
  return plugin_ref_to_se_ref(plugin);
}


namespace backup {

/**
  Local version of LEX_STRING structure.

  Defines various constructors for convenience.
*/

struct LEX_STRING: public ::LEX_STRING
{
  /// Base constructor for null string.
  LEX_STRING()
  {
    str= NULL;
    length= 0;
  }

  /**
    Constructor for LEX_STRING class.

    @param[in]  s  LEX_STRING string.
  */
  LEX_STRING(const ::LEX_STRING &s)
  {
    str= s.str;
    length= s.length;
  }

  /**
    Constructor for LEX_STRING class.

    @param[in]  s  Character class.
  */
  LEX_STRING(const char *s)
  {
    str= const_cast<char*>(s);
    length= strlen(s);
  }

  /**
    Constructor for LEX_STRING class.

    @param[in]  s  String class.
  */
  LEX_STRING(const String &s)
  {
    str= const_cast<char*>(s.ptr());
    length= s.length();
  }

  /**
     Constructor for the string class.

     This constructor takes a begging and ending pointer to construct the
     string.

     @param[in]  begin  Pointer to start of string.
     @param[in]  end    Pointer to end of string.
  */
  LEX_STRING(byte *begin, byte *end)
  {
    str= (char*)begin;
    if( begin && end > begin)
      length= end - begin;
    else
      length= 0;
  }
};


/**
  Local version of String class.

  Defines various constructors for convenience.
 */

class String: public ::String
{
public:

  /**
    Constructor for string class.

    @param[in]  s  String class.
  */
  String(const ::String &s) : ::String(s)
  {}

  /**
    Constructor for string class.

    @param[in]  s  LEX_STRING string.
  */
  String(const ::LEX_STRING &s)
    : ::String(s.str, (uint32)s.length, &::my_charset_bin)
  {
    // Check that string fits.
    DBUG_ASSERT(s.length <= ~((uint32)0));
  }

  /**
     Constructor.

     This constructor takes a begin and end pointer to construct the
     string. By default it creates a string in binary character set, that
     is, contents is not interpreted as characters. However, another 
     character set can be specified to treat the data between begin and end
     as a real string. 

     @param[in]  begin  Pointer to start of string.
     @param[in]  end    Pointer to end of string.
     @param[in]  charset  Optional character set of the string.
  */
  String(byte *begin, byte *end, ::CHARSET_INFO *charset= &::my_charset_bin)
    : ::String((char*)begin, (uint32)(end - begin), charset)
  {
    // Check that string length is correct.
    DBUG_ASSERT(begin <= end);
    /*
      This complex expression checks that the pointer difference fits into
      uint32 type reagardless of the size of pointer type and without generating
      compiler warnings (hopefully).

      The idea is to check that in the difference (Which is positive) no bits
      beyond the ones used by unit32 type are set.
    */
    DBUG_ASSERT(!((size_t)(end - begin) & ~((size_t)~((uint32)0))));

    // Note: explicit cast is needed to disambiguate.
    if (!begin)
     set((char*)NULL, 0, NULL);
  }

  /**
    Constructor for string class.

    @param[in]  s  Character string.
  */
  String(const char *s)
    : ::String(s, &::my_charset_bin)
  {}

  /// Base constructor for null string class.
  String() : ::String()
  {}
};


/**
  Create TABLE_LIST structure describing given table.

  @param[in]  tbl        A table for the list.
  @param[in]  lock_type  The lock type.
  @param[in]  mem        Pointer to base memory.

  @retval  TABLE_LIST
*/

inline
TABLE_LIST* mk_table_list_element(const Table_ref &tbl, thr_lock_type lock_type,
                                  MEM_ROOT *mem)
{
  DBUG_ASSERT(mem);

  TABLE_LIST *ptr= (TABLE_LIST*)alloc_root(mem, sizeof(TABLE_LIST));

  if (!ptr)
     return NULL;

  const char *t_name=     tbl.name().ptr();
  size_t      t_name_len= tbl.name().length();

  const char *d_name=     tbl.db().name().ptr();
  size_t      d_name_len= tbl.db().name().length();

  /*
    On lctn=2 system, even though original case of table names is preserved,
    low-level functions such as open_and_lock_tables() expect that names are
    converted to lower case. Since returned TABLE_LIST is used as parameter to
    such low-level functions, we do the transformation here (see BUG#48923).
  */ 

  if (lower_case_table_names == 2)
  {
    String *t_name_lc = new (mem) String(tbl.name());
    String *d_name_lc = new (mem) String(tbl.db().name());

    // Copy names to the newly allocated strings.
    t_name_lc->copy();
    d_name_lc->copy();

    // Convert them to lower case.
    my_casedn_str(t_name_lc->charset(), t_name_lc->c_ptr());
    my_casedn_str(d_name_lc->charset(), d_name_lc->c_ptr());

    t_name=     t_name_lc->ptr();
    t_name_len= t_name_lc->length();

    d_name=     d_name_lc->ptr();
    d_name_len= d_name_lc->length();
  }

  ptr->init_one_table(d_name, d_name_len, t_name, t_name_len, 
                      t_name, lock_type);

  return ptr;
}


/**
  Link two table lists together.

  @param[in] tl   The list to append onto.
  @param[in] next The list to append.

  @returns New list pointer.
*/

inline
TABLE_LIST* link_table_list(TABLE_LIST &tl, TABLE_LIST *next)
{
  tl.next_global= tl.next_local= tl.next_name_resolution_table= next;
  return &tl;
}


TABLE_LIST *build_table_list(const Table_list &tables, thr_lock_type lock,
                             MEM_ROOT *mem);
void free_table_list(TABLE_LIST* tables);

} // backup namespace


/**
  Implements a dynamic map from A to B* (also known as hash array).

  An instance @c map of class @c Map<A,B> can store mappings from values of
  type @c A to pointers of type @c B*. Such mappings are added with
  @code
   A a;
   B *ptr;

   map.insert(a,ptr);
  @endcode

  Later, one can examine the pointer assigned to a given value using operator[]
  @code
   A a
   B *ptr= map[a]
  @endcode

  If no mapping for the value a was defined, then returned pointer is NULL.

  In case type @c A is @c int we obtain a dynamic array where pointers are
  stored at indicated positions.
  @code

  Map<uint,B> map;

  B x,y;

  map.insert(1,&x);
  map.insert(7,&y);

  B *p1= map[1];  // p1 points at x
  B *p2= map[2];  // p2 is NULL
  @endcode

  However, it is also possible to have the pointers indexed by more complex
  values.

  @note We assume that type A has (fast) copy constructor.
*/

template<class A, class B>
class Map
{
  HASH     m_hash;      ///< For implementing the map.
  MEM_ROOT mem_root;    ///< For storing hash nodes.

public:

  /// Constructor.
  Map(size_t);
  ~Map();

  int insert(const A&, B*);
  B* operator[](const A&) const;

private:

  struct Node;
};


/*****************************************************************

  Implementation of Map template using HASH type.

 *****************************************************************/

/**
  Nodes inserted into HASH table.

  @note Such nodes can be allocated within a Map instance using placement
  @c new() operator. In that case the memory will be freed when the instance
  is destroyed.
*/

template<class A, class B>
struct Map<A,B>::Node
{
  /*
    Note: In the HASH handling code we assume that first sizeof(A) bytes
    of this structure contain the key value. Therefore it is important that
    Map<A,B>::Node is a POD, without a base class and without virtual table
    which would change the alignment of the key member.
  */
  A key;   ///< Key or index in hash/map.
  B *ptr;  ///< Pointer to item in hash/map.

  /// Constructor.
  Node(const A &a, B *b) :key(a), ptr(b) {}

  static void *operator new(size_t size, Map<A,B> *map) throw ()
  {
    DBUG_ASSERT(map);
    return alloc_root(&(map->mem_root), size);
  }

  static void operator delete(void *ptr, Map<A,B> *map)
  { /* never called */ }
};


template<class A, class B>
inline
Map<A,B>::Map(size_t init_size)
{
  init_alloc_root(&mem_root, ALLOC_ROOT_MIN_BLOCK_SIZE + init_size, 0);

  /*
    My_hash_init() expects the size of the hash as ulong value which
    on some systems is smaller than size_t. We will assume that size
    never exceeds the maximal value of ulong and check this here
    with assertion.
  */

  DBUG_ASSERT(init_size <= (size_t)(ulong)(~(ulong)0));

  my_hash_init(&m_hash, &::my_charset_bin, (ulong)init_size,
               0, sizeof(A), NULL, NULL, MYF(0));
}


template<class A, class B>
inline
Map<A,B>::~Map()
{
  my_hash_free(&m_hash);
  free_root(&mem_root, MYF(0));
}


/// Insert new mapping.

template<class A, class B>
inline
int Map<A,B>::insert(const A &a, B *b)
{
  Node *n= new (this) Node(a, b);

  return my_hash_insert(&m_hash, (uchar*) n);
}


/// Get pointer corresponding to the given value.

template<class A, class B>
inline
B* Map<A,B>::operator[](const A &a) const
{
  Node *n= (Node*) my_hash_search(&m_hash, (uchar*) &a, sizeof(A));

  return n ? n->ptr : NULL;
}


/**
  Specialization of Map template with integer indexes implemented as a
  Dynamic_array.
*/

template<class T>
class Map<uint,T>: public ::Dynamic_array< T* >
{
  typedef Dynamic_array< T* > Base;

public:

  /// Constructor.
  Map(uint init_size, uint increment);

  /**
    The index operator.

    @param[in]  pos  Position to retrieve.

    @returns The item and position @c pos.
  */
  T* operator[](ulong pos) const;

  /**
    Insert an item in the map.

    @param[in]  pos  Desired position.
    @param[in]  ptr  Item to insert.

    @retval  Status of insert.
  */
  int insert(ulong pos, T *ptr);
  ulong count() const;

private:

   void clear_free_space();
};


template<class T>
inline
Map<uint,T>::Map(uint init_size, uint increment) :Base(init_size, increment)
{
  clear_free_space();
}


template<class T>
inline
void Map<uint,T>::clear_free_space()
{
   DYNAMIC_ARRAY *array_ptr= &(this->array);

   /*
     Note: array_ptr is needed because dynamic_array_ptr() is a macro
     which applies -> operator to its argument.
   */
   uchar *start= dynamic_array_ptr(array_ptr, array_ptr->elements);
   uchar *end= dynamic_array_ptr(array_ptr, array_ptr->max_element);
   if (end > start)
     bzero(start, end - start);
}


template<class T>
inline
int Map<uint,T>::insert(ulong pos, T *ptr)
{
  uchar *entry;
  DYNAMIC_ARRAY *array_ptr= &(this->array);

  while (pos >= Base::array.max_element)
  {
    entry= alloc_dynamic(array_ptr);
    if (!entry)
     break;
  }

  clear_free_space();

  if (pos >= Base::array.max_element)
    return 1;

  if (pos >= Base::array.elements)
    Base::array.elements= pos + 1;

  entry= dynamic_array_ptr(array_ptr, pos);
  *(T**)entry= ptr;

  return 0;
}


template<class T>
inline
T* Map<uint,T>::operator[](ulong pos) const
{
  if (pos >= Base::elements())
    return NULL;

  return Base::at(pos);
}


/**
  Return number of entries in the dynamic array.

  @note Some of the entries can store NULL pointers as no mapping was
  defined for them.
 */

template<class T>
inline
ulong Map<uint,T>::count() const
{ return Base::array.elements; }


#endif
