/* Copyright (C) 2004-2007 MySQL AB, 2008-2009 Sun Microsystems, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/

/**
  @file

  @brief Contains the thread methods for MYSQL backup.

  The methods in this class are used to initialize the mutexes
  for the backup threads. Helper methods are included to make thread
  calls easier for the driver code.
*/

#include "../mysql_priv.h"
#include "be_thread.h"

/**
  @brief Creates a new THD object.

  Creates a new THD object for use in running as a separate thread.

  @returns Pointer to new THD object or 0 if error.

  @todo Move this method to a location where ha_ndbcluster_binlog.cc can
        use it and replace code in ndb_binlog_thread_func(void *arg) to
        call this function.

  @note my_net_init() this should be paired with my_net_end() on
        close/kill of thread.
*/

THD *create_new_thd()
{
  THD *thd;
  DBUG_ENTER("Create new THD object");

  thd= new THD;
  if (unlikely(!thd))
    DBUG_RETURN(0);
  THD_CHECK_SENTRY(thd);

  thd->thread_stack = (char*)&thd; // Remember where our stack is.
  mysql_mutex_lock(&LOCK_thread_count);
  thd->thread_id= thread_id++;
  mysql_mutex_unlock(&LOCK_thread_count);
  if (unlikely(thd->store_globals())) // For a proper MEM_ROOT.
  {
    thd->cleanup();
    delete thd;
    DBUG_RETURN(0);
  }

  thd->init_for_queries(); // Opening tables needs a proper LEX.
  thd->command= COM_DAEMON;
  thd->system_thread= SYSTEM_THREAD_BACKUP;
  thd->version= refresh_version;
  thd->set_time();
  thd->main_security_ctx.host_or_ip= "";
  thd->client_capabilities= 0;
  my_net_init(&thd->net, 0);
  thd->main_security_ctx.master_access= ~0;
  thd->main_security_ctx.priv_user= 0;
  thd->real_id= pthread_self();

  lex_start(thd);
  mysql_reset_thd_for_next_command(thd);

  /*
    Making this thread visible to SHOW PROCESSLIST is useful for
    troubleshooting a backup job (why does it stall etc).
  */
  mysql_mutex_lock(&LOCK_thread_count);
  threads.append(thd);
  thread_count++;
  thread_running++;
  mysql_mutex_unlock(&LOCK_thread_count);

  DBUG_RETURN(thd);
}


/**
  @brief Lock tables in driver.

  This method creates a new THD for use in the new thread. It calls
  the method to open and lock the tables.

  @note my_thread_init() should be paired with my_thread_end() on
        close/kill of thread.
*/

pthread_handler_t backup_thread_for_locking(void *arg)
{
  Locking_thread_st *locking_thd= static_cast<Locking_thread_st *>(arg);

  DBUG_PRINT("info", ("Default_backup - lock_tables_in_separate_thread"));

  my_thread_init();

  // First, create a new THD object.
  DBUG_PRINT("info",("Online backup creating THD struct for thread"));
  THD *thd= create_new_thd();

  if (thd == 0 || thd->killed)
  {
    locking_thd->m_error= TRUE;
    goto end2;
  }

  THD_SET_PROC_INFO(thd, "Locking thread started");
  thd->set_query(locking_thd->thd_name.c_ptr(), locking_thd->thd_name.length());

  pthread_detach_this_thread();

  locking_thd->lock_thd= thd;
  locking_thd->lock_state= LOCK_IN_PROGRESS;

  // There should be some tables to be opened. We error if this is not the case.

  DBUG_PRINT("info",("Online backup open tables in thread"));
  if (!locking_thd->tables_in_backup)
  {
    DBUG_PRINT("info",("Online backup locking error no tables to lock"));
    THD_SET_PROC_INFO(thd, "lock error");
    locking_thd->m_error= TRUE;
    goto end2;
  }

  // Now open and lock the tables.

  /*
    The MYSQL_OPEN_SKIP_TEMPORARY flag is needed so that temporary tables are
    not opened which would occlude the regular tables selected for backup
    (BUG#33574).
    
    Note: open_and_lock_tables_derived() will signal error when the
    locking thread is killed.
  */
  if (open_and_lock_tables(thd, locking_thd->tables_in_backup,
                           FALSE, /* Do not process derived tables. */
                           MYSQL_OPEN_SKIP_TEMPORARY)
     )
  {
    DBUG_PRINT("info",("Online backup locking thread dying"));
    THD_SET_PROC_INFO(thd, "lock error");
    locking_thd->m_error= TRUE;
    goto end;
  }

  locking_thd->lock_state= LOCK_ACQUIRED;

  /*
    Part of work is done. Wait until we are killed (COND_killed condition), 
    then release table locks and terminate. 
  */
  THD_SET_PROC_INFO(thd, "waiting for signal");

  mysql_mutex_lock(&locking_thd->THR_LOCK_cond);
  thd->enter_cond(&locking_thd->COND_killed,
                  &locking_thd->THR_LOCK_cond,
                  "Locking thread: holding table locks");
  while (!thd->killed)
    mysql_cond_wait(&locking_thd->COND_killed,
                    &locking_thd->THR_LOCK_cond);
  thd->exit_cond("Locking thread: terminating");
  THD_SET_PROC_INFO(thd, "terminating");

  DBUG_PRINT("info",("Locking thread locking thread terminating"));

  // Cleanup and return.

end:
  close_thread_tables(thd);

end2:
  THD_SET_PROC_INFO(thd, "lock done");

  // Final synchronization with driver's thread.

  mysql_mutex_lock(&locking_thd->THR_LOCK_cond);

  /*
    We must wait if do_not_terminate flag is TRUE. This is because driver's 
    thread is about to use lock_thd to kill us. Therefore we can't destroy THD
    instance yet, but must wait for COND_can_terminate.
  */ 
  while (locking_thd->do_not_terminate)
    mysql_cond_wait(&locking_thd->COND_can_terminate,
                    &locking_thd->THR_LOCK_cond);

  locking_thd->lock_state= LOCK_DONE;
  locking_thd->lock_thd= NULL;

  // Signal COND_done condition in case driver's thread is waiting for us.
  mysql_cond_signal(&locking_thd->COND_done);

  mysql_mutex_unlock(&locking_thd->THR_LOCK_cond);

  // Unregister this thread.
  
  net_end(&thd->net);
  mysql_mutex_lock(&LOCK_thread_count);
  delete thd;
  thread_count--;
  thread_running--;
  mysql_mutex_unlock(&LOCK_thread_count);

  // Final cleanup.

  my_thread_end(); /* always last, after all mutex usage */
  pthread_exit(0);
  return (0);
}


/// Constructor for Locking_thread_st structure.

Locking_thread_st::Locking_thread_st()
 : tables_in_backup(NULL), lock_thd(NULL), lock_state(LOCK_NOT_STARTED),
   m_thd(NULL), m_thread_started(FALSE), do_not_terminate(FALSE), 
   m_error(FALSE)
{
  // Initialize the thread mutex and cond variable.
  mysql_mutex_init(INSTRUMENT_ME, &THR_LOCK_cond, MY_MUTEX_INIT_FAST);
  mysql_cond_init(INSTRUMENT_ME, &COND_killed, NULL);
  mysql_cond_init(INSTRUMENT_ME, &COND_can_terminate, NULL);
  mysql_cond_init(INSTRUMENT_ME, &COND_done, NULL);
  thd_name.length(0);
};


/// Destructor for Locking_thread_st structure.

Locking_thread_st::~Locking_thread_st()
{
  // Kill the locking thread (if it is running)
  if (m_thread_started)
    kill_locking_thread();

  // Destroy the thread mutexes and cond variables.
  mysql_mutex_destroy(&THR_LOCK_cond);
  mysql_cond_destroy(&COND_killed);
  mysql_cond_destroy(&COND_can_terminate);
  mysql_cond_destroy(&COND_done);
}


/**
   Start the driver's lock thread.

   Launches a separate thread ("locking thread") which will lock
   tables.

   @param[in] tname The name of the thread which will show in
              process list.
*/

result_t Locking_thread_st::start_locking_thread(const char *tname)
{
  DBUG_ENTER("Locking_thread_st::start_locking_thread");
  pthread_t th;
  thd_name.append(tname);
  if (mysql_thread_create(key_thread_backup,
                          &th, &connection_attrib,
                          backup_thread_for_locking, this))
    SET_STATE_TO_ERROR_AND_DBUG_RETURN;
  m_thread_started= TRUE;
  DBUG_RETURN(backup::OK);
}


/**
   Kill the driver's lock thread and wait for it to terminate.

   This method issues the awake and broadcast to kill the locking thread. It
   also waits for COND_done so that we know that the locking thread will 
   terminate shortly. A care is taken to ensure that the lock_thd instance is
   not deleted when it is used to kill the locking thread. 
*/

void Locking_thread_st::kill_locking_thread()
{
  DBUG_ENTER("Locking_thread_st::kill_locking_thread");

  // Nothing to do if the locking thread has not been started.
  if (!m_thread_started)
    DBUG_VOID_RETURN;

 retry:

  mysql_mutex_lock(&THR_LOCK_cond);

  /*
    We hold THR_LOCK_cond. If lock_state == LOCK_DONE done then it means that 
    the locking thread is in its final section (after end2 label) and will
    terminate without any further action. 
  */ 
  if (lock_state == LOCK_DONE)
  {
    mysql_mutex_unlock(&THR_LOCK_cond);  
    DBUG_VOID_RETURN;
  }

  /*
    Otherwise (lock_state != LOCK_DONE), the locking thread must be still 
    before the final synchronization phase (i.e., before waiting for 
    COND_can_terminate). We want to ensure that lock_thd is set before we use
    it for killing the locking thread. It could happen that the locking thread
    have not yet created THD instance - in that case we must wait for it.
    Since this situation is unlikely, we do not bother with introducing yet 
    another condition, but we do active wait.
  */    
  if (unlikely(lock_thd == NULL))
  {
    mysql_mutex_unlock(&THR_LOCK_cond);  
    my_sleep(1);
    goto retry;
  }

  /*  
    Now we know that lock_thd is set. Since we are holding THR_LOCK_cond mutex,
    the THD instance will not be deleted by the locking thread. However, we 
    must release the mutex because THD::awake() can not be used while we are 
    holding one. To protect locking thread from terminating and deleting 
    lock_thd in the meanwhile, we set do_not_terminate flag to TRUE.
  */
  do_not_terminate= TRUE;
  mysql_mutex_unlock(&THR_LOCK_cond);  

  // Kill locking thread.
  DBUG_ASSERT(lock_thd);
  mysql_mutex_lock(&lock_thd->LOCK_thd_data);
  lock_thd->awake(THD::KILL_CONNECTION);
  mysql_mutex_unlock(&lock_thd->LOCK_thd_data);

  /*
    Now we will wait for COND_done to be sure that the locking thread goes 
    away. But first we will signal both conditions on which the locking thread
    might be waiting at the moment. 
  */ 

  mysql_mutex_lock(&THR_LOCK_cond);
  do_not_terminate= FALSE;
  mysql_cond_signal(&COND_killed);
  mysql_cond_signal(&COND_can_terminate);
  
  // Now Wait for COND_done.

  m_thd->enter_cond(&COND_done, &THR_LOCK_cond,
                    "Locking thread: waiting until locking thread is done");
  while (lock_state != LOCK_DONE)
    mysql_cond_wait(&COND_done, &THR_LOCK_cond);
  m_thd->exit_cond("Locking thread: terminating");

  m_thread_started= FALSE;

  if (m_error)
    THD_SET_PROC_INFO(m_thd, "error in the locking thread");

  DBUG_VOID_RETURN;
}
