#ifndef _BACKUP_THREAD_H
#define _BACKUP_THREAD_H

/* Copyright (c) 2008, 2010 Oracle and/or its affiliates. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <backup_engine.h>
#include <backup/image_info.h>

/// Macro for error handling.
#define SET_STATE_TO_ERROR_AND_DBUG_RETURN {                                \
    DBUG_PRINT("error",("driver got an error at %s:%d",__FILE__,__LINE__)); \
    DBUG_RETURN(backup::ERROR); }

/// Locking of tables goes through several states.
typedef enum {
  LOCK_NOT_STARTED,
  LOCK_IN_PROGRESS,
  LOCK_ACQUIRED,
  LOCK_DONE
} LOCK_STATE;

using backup::result_t;
using backup::Table_list;

/**
   backup_thread_for_locking

   This method creates a new thread and opens and locks the tables.

   @param[in] arg  List of arguments.
*/
pthread_handler_t backup_thread_for_locking(void *arg);

#ifdef HAVE_PSI_INTERFACE
extern PSI_thread_key key_thread_backup;
#endif

/**
  Locking_thread

  @brief Adds variables for using a locking thread for opening tables.

  The Backup_thread structure contains a mutex and condition variable
  for using a thread to open and lock the tables. This is meant to be a
  generic class that can be used elsewhere for opening and locking tables.

  @note This class correctly handles the separate locking thread. However,
  the class itself is *not* multi-thread safe. An instance of Locking_thread_st
  should be used by one thread only. In particular, calling
  @c start_locking_thread() from one thread and @c kill_locking_thread() from
  another thread running in parallel is not guaranteed to work.
*/

struct Locking_thread_st
{
public:
  Locking_thread_st();
  ~Locking_thread_st();

  /** 
    Signalled when the locking thread has been killed and should release
    table locks.
  */
  mysql_cond_t  COND_killed;

  /** 
    Signalled when it is OK for the locking thread to delete the THD instance
    and terminate.
  */
  mysql_cond_t  COND_can_terminate;

  /// Signalled by locking thread when it has finished and is about to exit.
  mysql_cond_t  COND_done;
  mysql_mutex_t THR_LOCK_cond; ///< Mutex for thread variables.

  TABLE_LIST *tables_in_backup;    ///< List of tables used in backup.
  THD *lock_thd;                   ///< Locking thread pointer.
  LOCK_STATE lock_state;           ///< Current state of the lock call.
  THD *m_thd;                      ///< Pointer to current thread struct.
  String thd_name;                 ///< Name of locking thread.
  /// Indicates if the locking thread has been started.
  my_bool m_thread_started;

  /** 
    Set to TRUE if the locking thread should not yet terminate. In particular
    the THD instance should not be deleted if this flag is TRUE.
  */
  my_bool  do_not_terminate;

  /** 
    TRUE if error was encountered by the locking thread.
  */
  my_bool  m_error;

  result_t start_locking_thread(const char *tname);
  void kill_locking_thread();

}; // Locking_thread_st


/**
  @class Backup_thread_driver

  @brief Adds variables for using a locking thread for opening tables.

  The Backup_thread_driver class extends the Backup_driver class by adding
  a mutex and condition variable for using a thread to open and lock the
  tables.

  @see Backup_driver
*/

class Backup_thread_driver : public Backup_driver
{
public:

  /// Constructor.
  Backup_thread_driver(const backup::Table_list &tables)
    :Backup_driver(tables) { locking_thd = new Locking_thread_st(); }
  /// Destructor.
  ~Backup_thread_driver() { delete locking_thd; }

  Locking_thread_st *locking_thd;  ///< Pointer to locking thread data.
}; // Backup_thread_driver class


#endif

