#ifndef RESTORE_INFO_H_
#define RESTORE_INFO_H_

/* Copyright 2008-2009 Sun Microsystems, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <backup_kernel.h>
#include <backup/image_info.h>
#include <backup/stream_services.h>


class Restore_info;


namespace backup {

class Logger;
class Input_stream;

int restore_table_data(THD*, Restore_info&,
                       backup::Input_stream&);
} // backup namespace


/**
  Specialization of @c Image_info which is in restore operation.

  An instance of this class is created by
  @c Backup_restore_ctx::prepare_for_restore() method, which reads the
  catalogue from a backup image.

  Currently it is not possible to select which objects will be restored. This
  class can only be used to examine what is going to be restored.
*/

class Restore_info: public backup::Image_info
{
public:

  backup::Logger &m_log;  ///< Pointer to logger class.

  ~Restore_info();

  /// Determine of information class is valid.
  bool is_valid();

  Image_info::Ts*    add_ts(obs::Obj*, uint);
  Image_info::Db*    add_db(obs::Obj*, uint);
  Image_info::Table* add_table(Image_info::Db&, obs::Obj*,
                               backup::Snapshot_info&, ulong);
  Image_info::Dbobj* add_db_object(Image_info::Db &db, const obj_type type, 
                                   obs::Obj*, ulong pos);
                               
  bool check_restore_privileges(struct st_bstream_item_info *item);
  
  /// Return whether it is safe to elevate restore.
  bool restore_elevation() { return m_restore_elevation; }
  
private:

  /*
    Note: constructor is private because instances of this class are supposed
    to be created only with Backup_restore_ctx::prepare_for_restore() method.
  */
  Restore_info(backup::Logger&, THD*);

  // Prevent copying/assignments.
  Restore_info(const Restore_info&);
  Restore_info& operator=(const Restore_info&);

  THD *m_thd;

  /**
    Determines if it is safe to implement restore elevation. Specifically,
    that the user has SUPER_ACL and restore_elevation is turned on.
  */
  bool m_restore_elevation;
  
  /// A flag to indicate that data has been modified during restore operation.
  bool m_data_changed;

  friend int backup::restore_table_data(THD*, Restore_info&,
                                        backup::Input_stream&);
  friend int ::bcat_add_item(st_bstream_image_header*,
                             struct st_bstream_item_info*);
  friend int ::bcat_create_item(st_bstream_image_header *catalogue,
                                struct st_bstream_item_info *item,
                                bstream_blob create_stmt,
                                bstream_blob other_meta_data);
  friend class Backup_restore_ctx;    // Needs access to the constructor.
};


inline
Restore_info::Restore_info(backup::Logger &log, THD *thd)
  :m_log(log), m_thd(thd), m_data_changed(FALSE)
{
  m_restore_elevation= (opt_restore_elevation &&
                        (m_thd->security_ctx->master_access & SUPER_ACL));
}


inline
Restore_info::~Restore_info()
{
  //  Delete Snapshot_info instances - they are created in bcat_reset().
  for (ushort n=0; n < snap_count(); ++n)
    delete m_snap[n];
}


inline
bool Restore_info::is_valid()
{
  return TRUE;
}


/// Wrapper around Image_info method which reports errors.

inline
backup::Image_info::Ts*
Restore_info::add_ts(obs::Obj *ts_obj, uint pos)
{
  const String *name= ts_obj->get_name();

  if (Ts *other= find_ts(*name))
  {
    m_log.report_error(ER_RESTORE_NONUNIQUE_OBJECT, other->name().ptr());
    return NULL;
  }

  Ts *ts= Image_info::add_ts(ts_obj, pos);

  if (!ts)
    m_log.report_error(ER_BACKUP_CATALOG_ADD_TS, name->ptr());

  return ts;
}


/// Wrapper around Image_info method which reports errors and checks privileges.

inline
backup::Image_info::Db*
Restore_info::add_db(obs::Obj *db_obj, uint pos)
{
  const String *name= db_obj->get_name();

  if (Db *other= find_db(*name))
  {
    m_log.report_error(ER_RESTORE_NONUNIQUE_OBJECT, other->name().ptr());
    return NULL;
  }

  Db *db= Image_info::add_db(db_obj, pos);

  if (!db)
    m_log.report_error(ER_BACKUP_CATALOG_ADD_DB, name->ptr());

  return db;
}


/// Wrapper around Image_info method which reports errors.

inline
backup::Image_info::Table*
Restore_info::add_table(Image_info::Db &db, obs::Obj *table_obj,
                        backup::Snapshot_info &snap, ulong pos)
{
  const String *name= table_obj->get_name();

  if (Table *other= db.find_table(*name))
  {
    m_log.report_error(ER_RESTORE_NONUNIQUE_DBOBJECT, other->name().ptr(), 
                                                      db.name().ptr());
    return NULL;
  }

  Table *t= Image_info::add_table(db, table_obj, snap, pos);

  if (!t)
    m_log.report_error(ER_BACKUP_CATALOG_ADD_TABLE, db.name().ptr(), 
                       name->ptr());

  return t;
}


inline
backup::Image_info::Dbobj*
Restore_info::add_db_object(Image_info::Db &db, const obj_type type, 
                            obs::Obj *obj, ulong pos)
{
  if (Dbobj *other= db.find_object(type, *obj->get_name()))
  {
    m_log.report_error(ER_RESTORE_NONUNIQUE_DBOBJECT, other->name().ptr(), 
                                                      db.name().ptr());
    return NULL;
  }

  Dbobj *ob= Image_info::add_db_object(db, type, obj, pos);

  return ob;
}

#endif /*RESTORE_INFO_H_*/
