#include <string.h>
#include <stdlib.h>

#include "stream_v1.h"
#include "stream_v1_services.h"

/**
  @file

  @brief
  Implementation of the high-level functions for writing and reading backup
  image using version 1 of backup image format (see @ref image_format1).

  @todo handle errors when creating iterators in functions like 
        bstream_wr_catalogue()
  @todo use data chunk sequence numbers to detect discontinuities in backup 
        stream.
*/

/**
  @page image_format1  Backup Image Format (v1)

  The backup image structure can be thought of at two levels:
  - The transport layer consists of blocks, all of which have the same size
    (except possibly the last). The size is given at the beginning of the first
    block.
  - The image layer consists of chunks, which have varying size.

  At the transport layer, blocks are read and decomposed to obtain fragments
  (one or more per block). Fragments are assembled to obtain chunks.
  A chunk might be smaller than a block, or it might contain information from
  multiple blocks.

  At the higher level, backup image is seen as a sequence of data chunks. How to
  interpret contents of these chunks is explained on @ref image_layer page.

  Page @ref transport_layer describes the low level structure of a backup image
  -- how fixed size blocks are decomposed into fragments from which data chunks
  are assembled.
*/

/**
  @page image_layer Image Layer Format

  On the top level, backup image consists of a premable, followed by several
  chunks with table data and closed with a summary chunk.

@verbatim
  backup_image: preamble table_data ... summary
@endverbatim

  @c Preamble describes the image and objects that it contains. @c Table_data
  chunks contain table contents.
  @c Summary contains information which is known only when backup image has
  been created. Currently this is mainly information about the validity point of
  the image and data needed for point-in-time recovery operations.
*/

/**
 @defgroup streamlib Backup Stream Library

  This library defines version 1 of backup image format described
  @ref image_format1 "here".
  It provides functions for writing and reading streams using this format.
  The functions are declared in the stream_v1.h header.
*/


#ifdef DBUG_OFF
# define ASSERT(X)
#else
# include <assert.h>
/** Macro to map assertion call when debug is off. */
# define ASSERT(X) assert(X)
#endif


/* local types */

typedef unsigned char bool; /**< definition of bool */
#define TRUE    1           /**< definition of true */
#define FALSE   0           /**< definition of false */

typedef bstream_byte byte;  /**< type definition of bstream byte */
typedef bstream_blob blob;  /**< type definition of bstream blob */

/**
  Macro mapping bzro() to memset().

  @note This is needed for seamless compilation on windows.
*/
#define bzero(A,B)  memset((A),0,(B))


/*
  Macros used to test results of functions writing/reading backup stream.

  To use these macros, function should define ret variable of type int and
  also wr_error/rd_error label to which these macros would jump in case of
  error.
*/

/** Macro to check write result. */
#define CHECK_WR_RES(X) \
  do{\
   if ((ret= X) != BSTREAM_OK) goto wr_error;\
  } while(0)

/** Macro to check read ok return. */
#define CHECK_RD_OK(X) \
 do{\
   if ((ret= X) != BSTREAM_OK)\
    { ret=BSTREAM_ERROR; goto rd_error; }\
 } while(0)

/** Macro to check read result. */
#define CHECK_RD_RES(X) \
 do{\
   if ((ret= X) == BSTREAM_ERROR) goto rd_error;\
 } while(0)


/* Functions for writing basic types. */

int bstream_wr_byte(backup_stream*, unsigned short int);
int bstream_wr_int2(backup_stream*, unsigned int);
int bstream_wr_int4(backup_stream*, unsigned long int);
int bstream_wr_num(backup_stream*, unsigned long int);
int bstream_wr_string(backup_stream*, bstream_blob);
int bstream_wr_time(backup_stream*, time_t);

/* 
  Low level i/o operations on backup stream (defined in 
  stream_v1_carrier.c). 
*/

int bstream_write(backup_stream*, bstream_blob*);
int bstream_write_part(backup_stream*, bstream_blob*, bstream_blob);
int bstream_write_blob(backup_stream*, bstream_blob);
int bstream_end_chunk(backup_stream*);

/**
 Flush backup stream`s output buffer to the output stream.

 This empties the output buffer.

 @param[in]  s  The backup stream to flush.

 @returns Status of operation.
*/
int bstream_flush(backup_stream*);

int bstream_read(backup_stream*, bstream_blob*);
int bstream_read_part(backup_stream*, bstream_blob*, bstream_blob);
int bstream_read_blob(backup_stream*, bstream_blob);
int bstream_skip(backup_stream*, unsigned long int);


/* Helper functions to safely convert values to byte type. */
byte get_byte_ulong(unsigned long int value)
{
  ASSERT(value < 256);
  return (byte)value;
}


byte get_byte_uint(unsigned int value)
{
  ASSERT(value < 256);
  return (byte)value;
}


byte get_byte_short(short value)
{
  ASSERT(value < 256);
  return (byte)value;
}


/**
  Return byte from unsigned short.

  @param[in]  value  Item to be changed to byte type.
*/

byte get_byte_ushort(unsigned short value)
{
  ASSERT(value < 256);
  return (byte)value;
}


byte get_byte_size_t(size_t value)
{
  ASSERT(value < 256);
  return (byte)value;
}


/*************************************************************************
 *
 *   IMAGE PREAMBLE AND SUMMARY
 *
 *************************************************************************/

/**
  @page image_layer
  @section preamble Preamble Format

  The preamble contains global information about the image and describes
  objects it contains.

@verbatim
  preamble: header snapshot_description ... catalog metadata
@endverbatim

  @c Preamble consists of several chunks, which contain the following
  information:
  - Header information, such as global image flags, image creation time,
    and server version number
  - Information about table data snapshots stored in the image
  - A catalog listing all objects stored in the image
  - Metadata for the objects

  Backup image can include several table data snapshots storing data from the
  tables. The snapshots correspond to backup drivers which were used to create
  them. Snapshots are numbered, starting from 1, according to the order in
  which their descriptions appear in the preamble. The number of snapshots
  present in the image is given by the @c snapshot_count field in the header.
*/

int bstream_wr_header(backup_stream*, struct st_bstream_image_header*);
int bstream_wr_catalogue(backup_stream*, struct st_bstream_image_header*);
int bstream_wr_meta_data(backup_stream *, struct st_bstream_image_header*);

/** Write backup image preamble */

int bstream_wr_preamble(backup_stream *s, struct st_bstream_image_header *hdr)
{
  int ret= BSTREAM_OK;

  CHECK_WR_RES(bstream_wr_header(s, hdr));
  CHECK_WR_RES(bstream_end_chunk(s));

  CHECK_WR_RES(bstream_wr_catalogue(s, hdr));
  CHECK_WR_RES(bstream_end_chunk(s));

  CHECK_WR_RES(bstream_wr_meta_data(s, hdr));
  CHECK_WR_RES(bstream_end_chunk(s));

wr_error:

  return ret;
}


/**
  Read backup image preamble creating all objects stored in it

  @retval BSTREAM_ERROR  Error while reading preamble
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOS    Preamble has been read and there are no more chunks in
                         the stream.
  @see @ref preamble.
*/

int bstream_rd_preamble(backup_stream *s, struct st_bstream_image_header *hdr)
{
  int ret= BSTREAM_OK;

  CHECK_RD_RES(bstream_rd_header(s, hdr));
  if (ret == BSTREAM_EOS)
    return BSTREAM_ERROR;
  CHECK_RD_OK(bstream_next_chunk(s));

  CHECK_RD_RES(bstream_rd_catalogue(s, hdr));
  if (ret == BSTREAM_EOS)
    return BSTREAM_ERROR;
  CHECK_RD_OK(bstream_next_chunk(s));

  CHECK_RD_RES(bstream_rd_meta_data(s, hdr));
  CHECK_RD_RES(bstream_next_chunk(s));

rd_error:

  return ret;
}


/*************************************************************************
 *
 *   IMAGE HEADER
 *
 *************************************************************************/

int bstream_wr_snapshot_info(backup_stream*, struct st_bstream_snapshot_info*);
int bstream_rd_snapshot_info(backup_stream*, struct st_bstream_snapshot_info*);

/**
  @page image_layer
  @subsection header Header Format

@verbatim
  header: flags creation_time snapshot_count server_version extra_data
@endverbatim

  @c Header chunk contains the following fields:
  - @c flags (2 bytes). Integer image flags.
  - @c creation_time (6 bytes). Time of image creation.
  - @c snapshot_count (1 byte). Number of snapshots included in the image.
  - @c server_version (variable length). Version of the server which created the
    image.
  - @c extra_data (variable length). The remaining bytes to the end of the chunk
    are reserved for storing additional information - currently they are
    ignored.

  Bits in @c flags field have the following meanings:
  - Bit 0: Reserved.
  - Bit 1 (BSTREAM_FLAG_BIG_ENDIAN)
   - If set, the server that created the backup has big-endian architecture.
   - If clear, the server has little-endian architecture.
  - Bit 2 (BSTREAM_FLAG_BINLOG)
   - If set, the binlog_coords and binlog_group_coords fields in the summary
     contain valid values.
   - If clear, binlog_coords and binlog_group_coords should be ignored (they are
     present in the summary but the values are not useful).
  - Bits 3-15: Reserved.

  @c Server_version is encoded as follows:
  - 1 byte. Integer major number
  - 1 byte. Integer minor number
  - 1 byte. Integer release number
  - Variable length. String representation of the version
  .
  For example, a server version of "6.0.8-alpha" is stored using this
  byte sequence:
  0x06 0x00 0x08 0x0b 0x36 0x2e 0x30 0x2e 0x38 0x2d 0x61 0x6c 0x70 0x68 0x61

  @note
  The big-endian flag has no bearing on storage of values in the backup image
  itself. It might be of use in the case where an image was written by some
  native driver but fails to restore on a machine other than the one on which
  it was created. If one checks the big-endian flag and finds that the backup
  host byte order differs from the restore host byte order, that might indicate
  that the native backup and restore drivers did not correctly deal with
  different host byte orders. A possible workaround would be to attempt the
  restore on a host with the same byte order as the backup host.
*/

/**
  Write header of backup image

  @see @ref header.
*/

int bstream_wr_header(backup_stream *s, struct st_bstream_image_header *hdr)
{
  int ret= BSTREAM_OK;
  unsigned int i;

  CHECK_WR_RES(bstream_wr_int2(s, hdr->flags));
  CHECK_WR_RES(bstream_wr_time(s, hdr->start_time));
  CHECK_WR_RES(bstream_wr_byte(s, hdr->snap_count));

  CHECK_WR_RES(bstream_wr_byte(s, hdr->server_version.major));
  CHECK_WR_RES(bstream_wr_byte(s, hdr->server_version.minor));
  CHECK_WR_RES(bstream_wr_byte(s, hdr->server_version.release));
  CHECK_WR_RES(bstream_wr_string(s, hdr->server_version.extra));

  for (i=0; i < hdr->snap_count; ++i)
  {
    CHECK_WR_RES(bstream_end_chunk(s));
    CHECK_WR_RES(bstream_wr_snapshot_info(s, &hdr->snapshot[i]));
  }

wr_error:

  return ret;
}


/**
  Read backup image header and fill @c st_bstream_image_header structure
  with the data read.

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached

  @see @ref header.
*/

int bstream_rd_header(backup_stream *s, struct st_bstream_image_header *hdr)
{
  int          ret= BSTREAM_OK;
  unsigned int i;

  CHECK_RD_OK(bstream_rd_int2(s, &hdr->flags));
  CHECK_RD_OK(bstream_rd_time(s, &hdr->start_time));
  CHECK_RD_OK(bstream_rd_byte(s, &hdr->snap_count));

  CHECK_RD_OK(bstream_rd_byte(s, &hdr->server_version.major));
  CHECK_RD_OK(bstream_rd_byte(s, &hdr->server_version.minor));
  CHECK_RD_OK(bstream_rd_byte(s, &hdr->server_version.release));
  CHECK_RD_RES(bstream_rd_string(s, &hdr->server_version.extra));

  for (i=0; i < hdr->snap_count; ++i)
  {
    if (ret != BSTREAM_EOC)
      return BSTREAM_ERROR;

    CHECK_RD_OK(bstream_next_chunk(s));
    CHECK_RD_RES(bstream_rd_snapshot_info(s, &hdr->snapshot[i]));
  }

rd_error:

  return ret;
}


/**
  @page image_layer
  @subsection snapshot Snapshot Description Format

  Table data snapshots are created by backup drivers and store the data from
  tables handled by these drivers. Each snapshot present in the image is
  described by a @c snapshot_description chunk in the preamble.

@verbatim
  snapshot_description: snapshot_type format_version global_options
                        table_count [storage_engine_info] extra_data

  backup_engine_info: engine_name major_version minor_version
@endverbatim

  Each @c snapshot_description chunk contains the following information:
  - @c snapshot_type (1 byte). Integer snapshot type, encoded as follows:
   - 0 (BI_NATIVE): Snapshot was created by native backup driver.
   - 1 (BI_DEFAULT): Snapshot was created by the built-in blocking driver.
   - 2 (BI_CS): Snapshot was created by the built-in driver that uses
     a consistent read transaction.
   - 3 (BI_NODATA): Phony snapshot used for tables whose data is not stored
     in the image.
  - @c format_version (2 bytes). Integer snapshot format version.
  - @c global_options (2 bytes). Integer global options. Reserved for future 
    use.
  - @c table_count (variable length). Integer count of the tables stored
    in the snapshot.
  - @c storage_engine_info (variable length). Information about storage engine
    which provided the native backup driver (present only for native snapshots).
  - @c extra_data (variable length). Extra data. Reserved for future use;
    currently empty.

  The @c backup_engine_info field is present only for native snapshots, that is,
  snapshots of type 0 (BI_NATIVE). It has this format:

  - @c engine_name (variable-length). String engine name.
  - @c major_version (1 byte). Integer major version of the engine.
  - @c minor_version (1 byte). Integer minor version of the engine.
*/

/**
  Save description of table data snapshot

  @see @ref snapshot
*/

int bstream_wr_snapshot_info(backup_stream *s,
                             struct st_bstream_snapshot_info *info)
{
  int ret= BSTREAM_OK;

  switch (info->type) {
  case BI_NATIVE:  ret= bstream_wr_byte(s, 0); break;
  case BI_DEFAULT: ret= bstream_wr_byte(s, 1); break;
  case BI_CS:      ret= bstream_wr_byte(s, 2); break;
  case BI_NODATA:  ret= bstream_wr_byte(s, 3); break;
  default:         ret= bstream_wr_byte(s, 4); break;
  }

  CHECK_WR_RES(bstream_wr_int2(s, info->version));

  if (ret != BSTREAM_OK)
    goto wr_error;

  CHECK_WR_RES(bstream_wr_int2(s, info->options));
  CHECK_WR_RES(bstream_wr_num(s, info->table_count));

  if (info->type == BI_NATIVE )
  {
    CHECK_WR_RES(bstream_wr_string(s, info->engine.name));
    CHECK_WR_RES(bstream_wr_byte(s, info->engine.major));
    CHECK_WR_RES(bstream_wr_byte(s, info->engine.minor));
  }

wr_error:

  return ret;
}


/**
  Read description of table data snapshot and save it in
  @c st_bstream_snapshot_info structure.

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached

  @note
  This function allocates memory to store snapshot info. The caller is
  responsible for freeing this memory.

  @see @ref snapshot
*/

int bstream_rd_snapshot_info(backup_stream *s,
                             struct st_bstream_snapshot_info *info)
{
  unsigned short int type;
  int                ret= BSTREAM_OK;

  bzero(info, sizeof(struct st_bstream_snapshot_info));

  CHECK_RD_OK(bstream_rd_byte(s, &type));

  switch (type) {
  case 0: type= BI_NATIVE; break;
  case 1: type= BI_DEFAULT; break;
  case 2: type= BI_CS; break;
  case 3: type= BI_NODATA; break;
  default: return BSTREAM_ERROR;
  }

  info->type= type;

  CHECK_RD_OK(bstream_rd_int2(s, &info->version));
  CHECK_RD_OK(bstream_rd_int2(s, &info->options));
  CHECK_RD_RES(bstream_rd_num(s, &info->table_count));

  if (type == BI_NATIVE )
  {
    if (ret != BSTREAM_OK)
      return BSTREAM_ERROR;

    CHECK_RD_OK(bstream_rd_string(s, &info->engine.name));
    CHECK_RD_OK(bstream_rd_byte(s, &info->engine.major));
    CHECK_RD_RES(bstream_rd_byte(s, &info->engine.minor));
  }

rd_error:

  return ret;
}


/*************************************************************************
 *
 *   CATALOGUE
 *
 *************************************************************************/

/**
  @page image_layer

  @section catalog Catalog Format

  The catalog describes what objects are stored in the backup image. It contains
  no metadata because that is stored in a separate section of the image.
  The catalog only lists the objects and provides the information needed to
  identify and select them. It consists of a header chunk followed by zero or
  more chunks that each describe objects in a single database:

@verbatim
  catalog: catalog_header [db_catalog ...]
@endverbatim

  @subsection catalog_header Catalog Header Format

@verbatim
  catalog_header: charsets 0x00 [users] 0x00 [tablespaces] 0x00 databases
@endverbatim

  The charsets, users, and tablespaces sections of the @c catalog_header chunk
  each contain a list of strings and are terminated by a 0x00 byte
  (an empty string). The databases section contains database-information items
  and extends to the end of the @c catalog_header chunk.

  The @c catalog_header starts with a list of character sets identified by name:

@verbatim
  charsets: charset_name ...
@endverbatim

  Character set names are string values represented using ASCII characters.
  The first two @c charset_name entries have a special meaning and should
  always be present:
  -# The character set used to encode all strings stored in
    the preamble following the charsets list. This should be a universal
    character set capable of representing any string, such as UTF8.
  -# The default character set of the server on which the backup image was
    created. It can be the same as the first character set.

  @c charset_name values following the first two, if present, are any character
  sets used by the objects stored in the image and thus needed to restore those
  objects.

  References to character sets at other locations in the backup image are as
  0-based positions within the character set list. The number of character sets
  is limited to 256 so that one byte is sufficient to identify a character set
  by its number.

  @note Currently, even if some objects (such as tables or databases) use
  character sets in their definition, these character sets will not be listed in
  the image. On restore, all entries in this list are ignored.
  Collation information is not stored currently, but might be in the future.

  @note Character set and collation numbers used within the image catalog are
  internal to the backup image and have nothing to do with the IDs used in the
  server. Backup image-processing code must translate between image IDs and
  internal server IDs if necessary.

  After the character set names, a list of users follows:

@verbatim
  users: user_name ...
@endverbatim

  The users list is currently always empty. It is reserved for future use when
  there will be a need to refer to users inside backup image (e.g., when user
  definitions are included).

  After users, a list of names of all tablespaces stored in the image follows:

@verbatim
  tablespace: tablespace_name ...
@endverbatim

  Each @c tablespace_name value is a string.

  After tablespaces, a list of all databases follows. If the list is empty, it
  consists of a single empty string. Otherwise, it is a list of one or more
  @c db_info entries:

@verbatim
  databases: 0x00 | db_info ...
@endverbatim

  @c db_info entries each contain fields that describe a single database:

@verbatim
  db_info: db_name db_flags [extra_data]
@endverbatim
  - @c db_name (variable length). String database name.
  - @c db_flags (1 byte). Integer database flags.
  - @c extra_data (variable length). Extra data. Optional; present only if
    indicated in the flags. It consists of:
    - Integer data length (2 bytes).
    - Data bytes, as many as specified by the length.

  Bits in the @c db_flags field are used as follows:
  - Bits 0-6: Reserved
  - Bit 7 (BSTREAM_FLAG_HAS_EXTRA_DATA): Set if the extra_data field is present
    in the db_info entry.
*/

/** Definition of extra data flag bits. */
#define BSTREAM_FLAG_HAS_EXTRA_DATA   0x80

int bstream_wr_db_catalogue(backup_stream*, struct st_bstream_image_header*,
                            struct st_bstream_db_info*);
int bstream_rd_db_catalogue(backup_stream*, struct st_bstream_image_header*,
                            struct st_bstream_db_info*);

/**
  Save catalogue of backup image.

  The contents of the image is read from the @c cat object using iterators
  and @c bcat_*() functions defined by the program using this library.

  @see @ref catalog
*/

int bstream_wr_catalogue(backup_stream *s, struct st_bstream_image_header *cat)
{
  void                      *it= NULL;
  blob                      *name;
  struct st_bstream_db_info *db_info;
  struct st_bstream_ts_info *ts_info;
  int                       ret;

  /* Charset list. */

  it= bcat_iterator_get(cat,BSTREAM_IT_CHARSET);

  if (!it)
    return BSTREAM_ERROR;

  while ((name= (blob*) bcat_iterator_next(cat, it)))
  {
    CHECK_WR_RES(bstream_wr_string(s, *name));
  }

  CHECK_WR_RES(bstream_wr_byte(s, 0x00));

  bcat_iterator_free(cat, it);
  it= NULL;

  /* List of users. */

  it= bcat_iterator_get(cat, BSTREAM_IT_USER);

  if (!it)
    return BSTREAM_ERROR;

  while ((name= (blob*) bcat_iterator_next(cat, it)))
  {
    CHECK_WR_RES(bstream_wr_string(s, *name));
  }

  CHECK_WR_RES(bstream_wr_byte(s, 0x00));

  bcat_iterator_free(cat, it);
  it= NULL;

  /* List of table spaces. */

  it= bcat_iterator_get(cat, BSTREAM_IT_TABLESPACE);

  if (!it)
    return BSTREAM_ERROR;

  while ((ts_info= (struct st_bstream_ts_info*) bcat_iterator_next(cat, it)))
  {
    CHECK_WR_RES(bstream_wr_string(s, ts_info->base.name));
  }

  CHECK_WR_RES(bstream_wr_byte(s, 0x00));

  bcat_iterator_free(cat, it);
  it= NULL;

  /* List of databases. */

  it= bcat_iterator_get(cat, BSTREAM_IT_DB);

  if (!it)
    return BSTREAM_ERROR;

  while ((db_info= (struct st_bstream_db_info*) bcat_iterator_next(cat, it)))
  {
    CHECK_WR_RES(bstream_wr_string(s, db_info->base.name));
    CHECK_WR_RES(bstream_wr_byte(s, 0x00)); /* flags */
  }

  bcat_iterator_free(cat, it);
  it= NULL;

  /* Database catalogues. */

  it= bcat_iterator_get(cat, BSTREAM_IT_DB);

  if (!it)
    return BSTREAM_ERROR;

  while ((db_info= (struct st_bstream_db_info*) bcat_iterator_next(cat, it)))
  {
    CHECK_WR_RES(bstream_end_chunk(s));
    CHECK_WR_RES(bstream_wr_db_catalogue(s, cat, db_info));
  }

  bcat_iterator_free(cat, it);
  it= NULL;

wr_error:

  /* Free iterator if not already done */
  if (it)
    bcat_iterator_free(cat, it);         /* purecov: inspected */

  return ret;
}


/**
  Read backup image catalogue.

  The @c cat object is populated with items read from the stream using
  @c bcat_add_item() function defined by the program using this library.

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached

  @see @ref catalog
*/

int bstream_rd_catalogue(backup_stream *s, struct st_bstream_image_header *cat)
{
  int                ret= BSTREAM_OK;
  unsigned short int flags;
  unsigned int       len;
  void               *iter= NULL;
  struct st_bstream_item_info item;
  struct st_bstream_db_info   *db_info;

  ret= bcat_reset(cat);
  if (ret != BSTREAM_OK)
    return BSTREAM_ERROR;

  /* Charset list. */

  item.type= BSTREAM_IT_CHARSET;
  item.pos= 0;

  do{

    CHECK_RD_OK(bstream_rd_string(s, &item.name));

    /* Empty string signals end of the list. */
    if (item.name.begin == NULL)
      break;

    if (bcat_add_item(cat, &item) != BSTREAM_OK)
      return BSTREAM_ERROR;

    item.pos++;

  } while (ret == BSTREAM_OK);

  /* List of users. */

  item.type= BSTREAM_IT_USER;
  item.pos= 0;

  do{

    CHECK_RD_RES(bstream_rd_string(s, &item.name));

    /* Empty string signals end of the list. */
    if (item.name.begin == NULL)
      break;

    if (bcat_add_item(cat, &item) != BSTREAM_OK)
      return BSTREAM_ERROR;

    item.pos++;

  } while (ret == BSTREAM_OK);

  /* List of table spaces. */

  item.type= BSTREAM_IT_TABLESPACE;
  item.pos= 0;

  do{

    CHECK_RD_RES(bstream_rd_string(s, &item.name));

    /* Empty string signals end of the list. */
    if (item.name.begin == NULL)
      break;

    if (bcat_add_item(cat, &item) != BSTREAM_OK)
      return BSTREAM_ERROR;

    item.pos++;

  } while (ret == BSTREAM_OK);

  /*
    If ret != BSTREAM_OK here, we hit end of chunk or stream. This means
    that there are no databases in the image and we are done reading the
    catalogue.
  */
  if (ret != BSTREAM_OK)
    return ret;

  /* List of databases. */

  item.type= BSTREAM_IT_DB;
  item.pos= 0;

  do {

    CHECK_RD_OK(bstream_rd_string(s, &item.name));

    if (bcat_add_item(cat, &item) != BSTREAM_OK)
      return BSTREAM_ERROR;

    item.pos++;

    CHECK_RD_RES(bstream_rd_byte(s, &flags));

    /*
      In current version we never store extra data in database catalogue entry.
      Flags field is always 0x00 (see bstream_wr_catalogue(), code which
      writes list of databases).
    */

    /* purecov: begin inspected */
    if (flags & BSTREAM_FLAG_HAS_EXTRA_DATA)
    {
      if (ret != BSTREAM_OK)
        return BSTREAM_ERROR;

      CHECK_RD_OK(bstream_rd_int2(s, &len));
      CHECK_RD_RES(bstream_skip(s, len));
    }
    /* purecov: end inspected */

  } while (ret == BSTREAM_OK);


  /* Database catalogues. */

  iter= bcat_iterator_get(cat, BSTREAM_IT_DB);

  if (!iter)
    return BSTREAM_ERROR;

  while ((db_info= (struct st_bstream_db_info*) bcat_iterator_next(cat, iter)))
  {
    if (ret != BSTREAM_EOC)
    {
      bcat_iterator_free(cat, iter);
      return BSTREAM_ERROR;
    }

    CHECK_RD_OK(bstream_next_chunk(s));
    CHECK_RD_RES(bstream_rd_db_catalogue(s, cat, db_info));
  }

  bcat_iterator_free(cat, iter);
  iter= NULL;

  if (bcat_close(cat) != BSTREAM_OK)
    return BSTREAM_ERROR;

rd_error:

  /* Free iterator if not already done */
  if (iter)
    bcat_iterator_free(cat, iter);       /* purecov: inspected */

  return ret;
}


/**
  @page image_layer

  @subsection db_catalog Database Catalog Format

  Each @c db_catalog chunk lists all tables and other per-database objects
  belonging to a single database. If there are no databases in the image, the
  databases list in the @c catalog_header chunk is empty and the catalog
  contains no @c db_catalog chunks following the @c catalog_header chunk.

  If a database is empty, its @c db_catalog chunk consists of two 0x00 bytes.
  Otherwise, @c db_catalog contains lists of @c table_info and @c db_item_info
  entries:

@verbatim
  db_catalog: 0x00 0x00 | db_tables db_other_items

  db_tables: table_info ...
  db_other_items: db_item_info ...
@endverbatim

  Each item entry in @c db_catalog starts with 2 byte integer indicating type of
  the object. All tables are stored before other per-database objects. This
  is important for addressing the per-database objects because the coordinates
  for such objects refer to the position in the @c db_other_items list.
  (For example, per-database object 0 is the object described by the first
   item in @c db_other_items.)

@verbatim
  table_info: type table_name flags snapshot_num table_pos
@endverbatim

  - @c type (2 bytes): Integer object type (always 0x05 0x00 for a table).
  - @c table_name (variable length): String table name.
  - @c flags (1 byte): Integer flags.
  - @c snapshot_num (1 byte): Integer indicating which snapshot contains
    table's data (0-based).
  - @c table_pos (variable length): Integer position of the table within
    its snapshot (0-based).

  The @c flags field is reserved for future use.

@verbatim
  db_item_info: type name [snapshot_num table_pos]
@endverbatim

  Per-database objects other than tables are described by db_item_info entries.
  For most objects such entry contains just object type and name. For triggers, 
  additionally coordinates of the table on which the trigger fires are stored.

  - @c type (2 bytes): Integer object type.
  - @c name (variable length): String object name.
  - @c snapshot_num (1 byte): Integer indicating which snapshot contains
    table's data (0-based). Only for triggers.
  - @c table_pos (variable length): Integer position of the table within
    its snapshot (0-based). Only for triggers.
 
  Allowable object type values are given in @ref basic_data_itype.
*/

int bstream_wr_item_type(backup_stream *s, enum enum_bstream_item_type type);
int bstream_rd_item_type(backup_stream *s, enum enum_bstream_item_type *type);

/**
  Save catalogue of objects belonging to given database.

  @see @ref db_catalog
*/

int bstream_wr_db_catalogue(backup_stream *s,
                            struct st_bstream_image_header *cat,
                            struct st_bstream_db_info *db_info)
{
  void                          *iter= NULL;
  struct st_bstream_dbitem_info *item;
  int                           ret= BSTREAM_OK;
  bool                          catalogue_empty= TRUE;

  iter= bcat_db_iterator_get(cat, db_info);

  if (!iter)
    return BSTREAM_ERROR;

  while ((item= bcat_db_iterator_next(cat, db_info, iter)))
  {
    catalogue_empty= FALSE;

    CHECK_WR_RES(bstream_wr_item_type(s, item->base.type));
    CHECK_WR_RES(bstream_wr_string(s, item->base.name));

    if (item->base.type == BSTREAM_IT_TABLE)
    {
      CHECK_WR_RES(bstream_wr_byte(s, 0x00)); /* flags */
      CHECK_WR_RES(bstream_wr_byte(s,
                              ((struct st_bstream_table_info*)item)->snap_num));
      CHECK_WR_RES(bstream_wr_num(s, item->base.pos));
    }

    /*
      If this is a trigger, save its table reference to the stream.
    */
    if (item->base.type == BSTREAM_IT_TRIGGER)
    {
      CHECK_WR_RES(bstream_wr_byte(s, item->snap_num));
      CHECK_WR_RES(bstream_wr_num(s, item->pos));
    }
  }

  bcat_db_iterator_free(cat, db_info, iter);
  iter= NULL;

  if (catalogue_empty)
    CHECK_WR_RES(bstream_wr_item_type(s, BSTREAM_IT_LAST));

wr_error:

  /* Free iterator if not already done. */
  if (iter)
    bcat_db_iterator_free(cat, db_info, iter);

  return ret;
}


/**
  Read catalogue of given database.

  Object @c cat is populated with the items read using @c bcat_add_item()
  function defined by the program using this library.


  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached

  @see @ref db_catalog
*/

int bstream_rd_db_catalogue(backup_stream *s,
                            struct st_bstream_image_header *cat,
                            struct st_bstream_db_info *db_info)
{
  unsigned short int           flags;
  struct st_bstream_table_info ti;
  unsigned long int            pos= 0;
  int                          ret;

  bzero(&ti, sizeof(ti));
  ti.base.db= db_info;

  /* We read the first byte to see if the catalogue is empty. */

  CHECK_RD_RES(bstream_rd_item_type(s, &ti.base.base.type));

  if (ti.base.base.type == BSTREAM_IT_LAST)
    return ret;

  /*
    We have some entries in the catalogue, we read them in the following
    loop.
  */

  do {

    CHECK_RD_RES(bstream_rd_string(s, &ti.base.base.name));

    if (ti.base.base.type == BSTREAM_IT_TABLE)
    {
      if (ret != BSTREAM_OK)
        return BSTREAM_ERROR;

      /* Flags are currently ignored. */
      CHECK_RD_OK(bstream_rd_byte(s, &flags));
      CHECK_RD_OK(bstream_rd_byte(s, &ti.snap_num));
      CHECK_RD_RES(bstream_rd_num(s, &ti.base.base.pos));
    }
    else
      ti.base.base.pos= pos++;

    /*
      If this is a trigger, read the table reference information.
    */
    if (ti.base.base.type == BSTREAM_IT_TRIGGER)
    {
      /* Read table reference. */
      CHECK_RD_RES(bstream_rd_byte(s, &ti.base.snap_num));
      CHECK_RD_RES(bstream_rd_num(s, &ti.base.pos));
    }
    
    if (bcat_add_item(cat, &ti.base.base) != BSTREAM_OK)
      return BSTREAM_ERROR;

    /* Read type of next item, if we haven't hit end of chunk/stream. */

    if (ret == BSTREAM_OK)
      CHECK_RD_OK(bstream_rd_item_type(s, &ti.base.base.type));

  } while (ret == BSTREAM_OK);

rd_error:

  return ret;
}


/*************************************************************************
 *
 *   METADATA
 *
 *************************************************************************/

/**
  @page image_layer

  @section metadata Metadata Format

  The metadata section contains information for objects that need to be created
  when restoring data. It has three main sections:
  - Metadata for global objects (single chunk)
  - Metadata for tables (one chunk per a database)
  - Metadata for other per-database objects (single chunk)

@verbatim
  metadata: global_items [tables ... other_items]
@endverbatim

  If there are no databases in the image, @c metadata consists of the
  @c global_items chunk only. Otherwise, for each database stored in the image
  one @c tables chunk with table metadata follows. Finally, there is
  @c other_items chunk containing metadata for per-database objects other than
  tables.

  Each of these chunks contains a list of metadata entries, each entry 
  containing information required for restoring a single object.
  The order of metadata entries is relevant. They are stored in such an order
  that objects can be created while reading these entries without breaking any
  dependencies. An exception is metadata for tables which are grouped by
  database, which can break dependencies. However, this is OK since only
  foreign key constraints can introduce dependencies between tables and they are
  disabled during restore. If the list stored in a chunk is empty, then the
  chunk consists of two 0x00 bytes.

  Currently the @c global_items chunk contains metadata only for tablespaces and
  databases, but it might be used to store other global objects in the future.
  Tablespace definitions precede database definitions.

  The @c other_items chunk stores metadata for per-database objects other than
  tables. Currently these are stored routines, views, triggers, events and
  grants. They are not grouped by a database but rather stored in a dependency
  preserving order. For historical reasons, there are always two 0x00 bytes at
  the end of @c other_items chunk.
*/

/** Different formats in which object positions are stored. */
enum enum_bstream_meta_item_kind {
  GLOBAL_ITEM,   /**< Only object position is stored. */
  TABLE_ITEM,    /**< Only table position is stored (database is implicit). */
  PER_DB_ITEM    /**< Object position followed by its database position. */
};


int bstream_wr_meta_item(backup_stream*, enum enum_bstream_meta_item_kind,
                         unsigned short int, struct st_bstream_item_info*);

int bstream_rd_meta_item(backup_stream *s,
                         enum enum_bstream_meta_item_kind kind,
                         unsigned short int *flags,
                         struct st_bstream_item_info **item);

int bstream_wr_item_def(backup_stream*, struct st_bstream_image_header*,
                        enum enum_bstream_meta_item_kind,
                        struct st_bstream_item_info*);

int read_and_create_items(backup_stream *s, struct st_bstream_image_header *cat,
                          enum enum_bstream_meta_item_kind kind);

/**
  Write metadata section of a backup image

  @see @ref metadata
*/

int bstream_wr_meta_data(backup_stream *s, struct st_bstream_image_header *cat)
{
  void                        *iter= NULL, *titer= NULL;
  struct st_bstream_item_info *item;
  struct st_bstream_db_info   *db_info;
  int                         ret=          BSTREAM_OK;
  bool                        item_written= FALSE;
  bool                        has_db=       FALSE;

  /* Global objects (this includes databases). */

  iter= bcat_iterator_get(cat, BSTREAM_IT_GLOBAL);

  if (!iter)
    return BSTREAM_ERROR;

  while ((item= bcat_iterator_next(cat, iter)))
  {
    item_written= TRUE;
    CHECK_WR_RES(bstream_wr_item_def(s, cat, GLOBAL_ITEM,item));
  }

  /* Mark empty list if no items were written. */
  if (!item_written)
    CHECK_WR_RES(bstream_wr_item_type(s, BSTREAM_IT_LAST));

  bcat_iterator_free(cat, iter);
  iter= NULL;

  /* Tables. */

  iter= bcat_iterator_get(cat, BSTREAM_IT_DB);

  if (!iter)
    return BSTREAM_ERROR;

  while ((db_info= (struct st_bstream_db_info*)bcat_iterator_next(cat, iter)))
  {
    has_db= TRUE;
    CHECK_WR_RES(bstream_end_chunk(s));

    titer= bcat_db_iterator_get(cat, db_info);

    if (!titer)
    {
      bcat_iterator_free(cat, iter);
      return BSTREAM_ERROR;
    }

    item_written= FALSE;
    while ((item= (struct st_bstream_item_info*)
                  bcat_db_iterator_next(cat, db_info, titer)))
    {
      if (item->type != BSTREAM_IT_TABLE)
        continue;

      CHECK_WR_RES(bstream_wr_item_def(s, cat, TABLE_ITEM, item));
      item_written= TRUE;
    }

    /* Mark empty list. */
    if (!item_written)
      CHECK_WR_RES(bstream_wr_item_type(s, BSTREAM_IT_LAST));

    bcat_db_iterator_free(cat, db_info, titer);
    titer= NULL;
  }

  bcat_iterator_free(cat, iter);
  iter= NULL;

  /* If we found no databases in the catalogue, we are done. */
  if (!has_db)
    return BSTREAM_OK;

  /* Other per-db objects. */

  CHECK_WR_RES(bstream_end_chunk(s));

  iter= bcat_iterator_get(cat, BSTREAM_IT_PERDB);

  if (!iter)
    return BSTREAM_ERROR;

  while ((item= bcat_iterator_next(cat, iter)))
  {
    if (item->type == BSTREAM_IT_TABLE)
      continue;

    CHECK_WR_RES(bstream_wr_item_def(s, cat, PER_DB_ITEM,item));
  }

  bcat_iterator_free(cat, iter);
  iter= NULL;

  /* The final 0x00 bytes. */

  CHECK_WR_RES(bstream_wr_item_type(s, BSTREAM_IT_LAST));

wr_error:

  /* Free iterators if not already done. */
  if (titer)
    bcat_iterator_free(cat, titer);      /* purecov: inspected */

  if (iter)
    bcat_iterator_free(cat, iter);

  return ret;
}


/**
  Read backup image metadata section.

  All objects read are created using @c bstream_create_item() function.

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached

  @see @ref metadata
*/

int bstream_rd_meta_data(backup_stream *s, struct st_bstream_image_header *cat)
{
  void                      *iter=  NULL;
  struct st_bstream_db_info *db_info;
  int                       ret=    BSTREAM_OK;
  bool                      has_db= FALSE;

  /* Global objects. */

  CHECK_RD_RES(read_and_create_items(s, cat, GLOBAL_ITEM));

  /* Tables. */

  iter= bcat_iterator_get(cat, BSTREAM_IT_DB);

  if (!iter)
    return BSTREAM_ERROR;

  while ((db_info= (struct st_bstream_db_info*)bcat_iterator_next(cat, iter)))
  {
    has_db= TRUE;

    if (ret != BSTREAM_EOC)
    {
      bcat_iterator_free(cat, iter);
      return BSTREAM_ERROR;
    }

    CHECK_RD_OK(bstream_next_chunk(s));
    CHECK_RD_RES(read_and_create_items(s, cat, TABLE_ITEM));
  }

  bcat_iterator_free(cat, iter);
  iter= NULL;

  /* If image has no databases, there is nothing more to read. */

  if (!has_db)
    return ret;

  /* Other per-db objects. */

  if (ret != BSTREAM_EOC)
    return BSTREAM_ERROR;

  CHECK_RD_OK(bstream_next_chunk(s));
  CHECK_RD_RES(read_and_create_items(s, cat, PER_DB_ITEM));

rd_error:

  /* Free iterator if not already done. */
  if (iter)
    bcat_iterator_free(cat, iter);       /* purecov: inspected */

  return ret;
}


/**
  @page image_layer

  @subsection metadata_entry Single metadata entry format

  A single metadata entry has the following format:

@verbatim
  item_entry: type flags catalog_pos [extra_data] [object_metadata]
@endverbatim

  - @c type (2 bytes): Integer object type.
  - @c flags (1 byte): Integer flags.
  - @c catalog_pos (variable length): Catalog coordinates of the object - exact
    format depends on the type of the object.
  - @c extra_data  (variable length): Optional; present only if
    BSTREAM_FLAG_HAS_EXTRA_DATA is set in the flags. (Currently not used.) If
    present it consists of
    - Integer data length (2 bytes).
    - Data bytes, as many as specified by the length.
  - @c object_metadata (variable length): String with object's serialization.
    Optional; present only if BSTREAM_FLAG_HAS_CREATE_STMT is set in the flags.

  Allowable object type values are given in @ref basic_data_itype. The format of
  the catalog coordinates depends on the object type, as described below.

  Object metadata is a serialization of the object as a string, as
  provided by server's object services API. Usually, it is an SQL CREATE
  statement for the object, but it can also contain additional SQL statements
  and other information. There is also a possibility of storing arbitrary
  additional metadata in the optional @c extra_data field, however this field
  is not used currently.

  The @c flags field indicates which metadata elements are present in
  the entry:
  - Bits 0-5: Reserved
  - Bit 6 (BSTREAM_FLAG_HAS_CREATE_STMT): The object_metadata field is present.
  - Bit 7 (BSTREAM_FLAG_HAS_EXTRA_DATA): The extra_data field is present.

  @subsubsection catalog_pos_global Catalog coordinates for global objects

  These are used to identify global objects whose metadata is stored in
  @c global_items chunk of metadata section. Coordinate of a global object
  consists of a single variable length integer:

@verbatim
  catalog_pos: pos_in_the_list
@endverbatim

  It is the 0-based position of the object in the corresponding list
  in the @c catalog_header chunk of the catalog. The catalog is capable of
  storing four types of global objects: databases, tablespaces, users, and
  character sets. For all these global objects, @c catalog_pos is a single
  number pointing at the entry on the appropriate list in the @c catalog_header
  chunk:
  - Database: The @c databases list.
  - Tablespace: The @c tablespaces list.
  - User: The @c users list.
  - Character set: The @c charsets list.

  @note Only tablespace and database metadata is stored in the @c global_items
  chunk of metadata section thus only tablespace and database coordinates are
  currently used.

  @subsubsection catalog_pos_table Catalog coordinates for tables

  These are used for identifying tables inside table metadata entries in @c
  tables chunks of metadata section.

@verbatim
  catalog_pos: table_pos snapshot_num
@endverbatim
  - @c table_pos (variable length): Integer position of the table within the
  snapshot storing this table's data (0-based).
  - @c snapshot_num (1 byte). Integer snapshot number (0-based).

  Table coordinates (K,N) indicate the table described by the @c table_info
  entry with @c snapshot_num = N and @c table_pos = K. This @c table_info entry
  can be found in the @c db_catalog chunk corresponding to the database
  containing the table.

  @subsubsection catalog_pos_perdb Catalog coordinates for other per-database 
                                   objects

  They are used for identifying per-database objects for which metadata is
  stored in the @c other_items chunk of the metadata section.

@verbatim
  catalog_pos: item_pos db_num
@endverbatim
  - @c item_pos (variable length): 0-based position in the @c db_other_items
  list from the @c db_catalog chunk corresponding to the given database.
  - @c db_num (variable length): 0-based position of the database in the
  @c databases list in @c catalog_header.

  The other per-database objects represent views, stored procedures,
  stored functions, events, triggers, and grants. Coordinates (K,N) indicate
  the object described by the K-th @c db_item_info entry in the
  @c db_other_items list inside the N-th @c db_catalog chunk.
*/

/** Definition of create statement flag bits. */
#define BSTREAM_FLAG_HAS_CREATE_STMT 0x40

/**
  Write entry describing single object but without CREATE statement or
  other meta data.

  @see @ref metadata_entry
*/

int bstream_wr_meta_item(backup_stream *s,
                         enum enum_bstream_meta_item_kind kind,
                         unsigned short int flags,
                         struct st_bstream_item_info *item)
{
  int ret= BSTREAM_OK;

  /* Save type and flags. */

  CHECK_WR_RES(bstream_wr_item_type(s, item->type));
  CHECK_WR_RES(bstream_wr_byte(s, flags));

  /* Save object's position in the catalogue. */

  CHECK_WR_RES(bstream_wr_num(s, item->pos));

  if (kind == TABLE_ITEM)
  {
    CHECK_WR_RES(bstream_wr_byte(s,
                              ((struct st_bstream_table_info*)item)->snap_num));
    return ret;
  }

  if (kind == PER_DB_ITEM)
    CHECK_WR_RES(bstream_wr_num(s,
                         ((struct st_bstream_dbitem_info*)item)->db->base.pos));

wr_error:

  return ret;
}


/**
  Read initial part of object entry and locate that object in the catalogue.

  Pointer to an appropriate structure describing the located object is stored in
  @c (*item). This description is not persistent - next call to this function
  can overwrite it with description of another object.

  @param[in] s    the backup stream
  @param[in] kind  format in which object coordinates are stored
  @param[in] flags the flags saved in the entry are stored in that location
  @param[in] item  pointer to a structure describing object found is stored 
                   here.

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached

  @note If function returns BSTREAM_OK and @c (*item) is set to NULL, it means
        that we are looking at an empty object list.

  @see @ref metadata_entry
*/

int bstream_rd_meta_item(backup_stream *s,
                         enum enum_bstream_meta_item_kind kind,
                         unsigned short int *flags,
                         struct st_bstream_item_info **item)
{
  static struct st_bstream_db_info db;

  static union
  {
    struct st_bstream_item_info   any;
    struct st_bstream_db_info     db;
    struct st_bstream_table_info  table;
    struct st_bstream_dbitem_info per_db;
  } item_buf;

  int ret= BSTREAM_OK;

  CHECK_RD_RES(bstream_rd_item_type(s, &item_buf.any.type));

  /* type == BSTREAM_IT_LAST means that we hit a no-item marker (0x00) */
  if (item_buf.any.type == BSTREAM_IT_LAST)
  {
    *item= NULL;
    return ret;
  }

  if (ret != BSTREAM_OK)
    return BSTREAM_ERROR;

  ASSERT(item);
  *item= &item_buf.any;

  CHECK_RD_OK(bstream_rd_byte(s, flags));

  /* Read object's position. */

  CHECK_RD_RES(bstream_rd_num(s, &item_buf.any.pos));

  if (kind == TABLE_ITEM)
  {
    if (ret != BSTREAM_OK)
      return BSTREAM_ERROR;

    CHECK_RD_RES(bstream_rd_byte(s, &item_buf.table.snap_num));
      return ret;
  }

  /* Read db pos if present. */

  if (kind == PER_DB_ITEM)
  {
    if (ret != BSTREAM_OK)
      return BSTREAM_ERROR;

    db.base.type= BSTREAM_IT_DB;
    CHECK_RD_RES(bstream_rd_num(s, &db.base.pos));

    item_buf.per_db.db= &db;
  }

rd_error:

  return ret;
}


/**
  Write entry with given object's metadata.

  @param[in]  s     backup stream.
  @param[in]  cat   image catalogue.
  @param[in]  kind  determines format in which object's coordinates are saved.
  @param[in]  item  stream item to write.

  @returns Status of operation.

  @see @ref metadata_entry
*/

int bstream_wr_item_def(backup_stream *s,
                        struct st_bstream_image_header *cat,
                        enum enum_bstream_meta_item_kind kind,
                        struct st_bstream_item_info *item)
{
  unsigned short int flags= 0x00;
  blob query;
  blob data;
  int  ret= BSTREAM_OK;

  /* Initialize variables. */
  data.begin=  0;
  data.end=    0;
  query.begin= 0;
  query.end=   0;

  /*
    Fetch object's create query and/or extra metadata data. Note that
    the BSTREAM_EOS reply from bcat_get_item_create_*() functions
    indicates lack of the corresponding piece of metadata.
  */

  ret= bcat_get_item_create_query(cat, item, &query);
  if (ret == BSTREAM_OK)
    flags |= BSTREAM_FLAG_HAS_CREATE_STMT;
  else if (ret == BSTREAM_ERROR)
    goto wr_error;

  ret= bcat_get_item_create_data(cat, item, &data);
  if (ret == BSTREAM_OK)
    flags |= BSTREAM_FLAG_HAS_EXTRA_DATA;
  else if (ret == BSTREAM_ERROR)
    goto wr_error;

  /* Save the header of metadata entry, containing object coordinates. */

  ret= bstream_wr_meta_item(s, kind, flags, item);
  if (ret == BSTREAM_ERROR)
    goto wr_error;

  /* Save create query and/or create data. */

  if (flags & BSTREAM_FLAG_HAS_EXTRA_DATA)
    CHECK_WR_RES(bstream_wr_string(s, data));

  if (flags & BSTREAM_FLAG_HAS_CREATE_STMT)
    CHECK_WR_RES(bstream_wr_string(s, query));

wr_error:

  return ret;
}


/**
  Read list of metadata entries and create the corresponding objects.

  The entries are read until the end of chunk or 0x00 marker is hit.
  After reading metadata for each object (CREATE statement and/or extra 
  metadata) the object is created using @c bstream_create_item() function which
  should be implemented by the program using this library.

  @retval BSTREAM_ERROR error while reading
  @retval BSTREAM_EOC   either list was empty or all objects were read and
                        created successfully
  @retval BSTREAM_EOS   all objects read and created successfully and end of
                        stream has been reached

  @see @ref metadata_entry
*/

int read_and_create_items(backup_stream *s, struct st_bstream_image_header *cat,
                          enum enum_bstream_meta_item_kind kind)
{
  unsigned short int          flags;
  unsigned int                ret;
  struct st_bstream_item_info *item;
  blob                        query,data;

  do {

    CHECK_RD_RES(bstream_rd_meta_item(s, kind, &flags, &item));

    /* If 0x00 marker was read, item == NULL. */
    if (item == NULL)
      return ret;

    query.begin= query.end= NULL;
    data.begin=  data.end=  NULL;

    if (flags & BSTREAM_FLAG_HAS_EXTRA_DATA)
    {
      if (ret != BSTREAM_OK)
        return BSTREAM_ERROR;
      CHECK_RD_RES(bstream_rd_string(s, &data));
    }

    if (flags & BSTREAM_FLAG_HAS_CREATE_STMT)
    {
      if (ret != BSTREAM_OK)
        return BSTREAM_ERROR;
      CHECK_RD_RES(bstream_rd_string(s, &query));
    }

    if (bcat_create_item(cat, item, query, data) != BSTREAM_OK)
      return BSTREAM_ERROR;

    bstream_free(query.begin);
    bstream_free(data.begin);

  } while (ret == BSTREAM_OK);

rd_error:

  return ret;
}



/*************************************************************************
 *
 *   TABLE DATA
 *
 *************************************************************************/

/**
  @page image_layer

  @section table_data Table Data Format

  A snapshot of data from several tables is created by a backup driver
  and used by restore driver to restore these table data. Each table is handled
  by one of the drivers and each driver has a list of tables which it handles.
  The number of snapshots contained in a backup image is given in the
  @c header chunk (see @ref header). Each snapshot is described by a single
  @c snapshot_description chunk in the preamble (see @ref snapshot). Snapshots
  are numbered by consecutive numbers starting with 1.

  Table data snapshot consists of several data chunks. Usually, each data chunk
  is assigned to one of the tables within the snapshot, but there can also
  be "common data" chunks which are not tied to any particular table.
  Data chunks from different snapshots can be mixed in the image.

  A single @c table_data chunk within backup image contains a data chunk from
  one of the snapshots. It has the following format:

@verbatim
  table_data: snapshot_num sequence_num flags table_num data
@endverbatim
  - @c snapshot_num (1 byte): Integer snapshot number (1-based).
  - @c sequence_num (2 bytes): Integer sequence number. Currently not used and
    is always 0.
  - @c flags (1 byte): Integer flags.
  - @c table_num (variable length): Integer table number (1-based).
  - @c data (variable length): Table data, to end of chunk.

  The snapshot to which this data chunk belongs is identified in the first byte
  of the chunk (@c snapshot_num field). Since snapshot numbers begin with 1,
  this byte must be non-zero. This allows detecting the end of the @c table_data
  chunks sequence in the stream as the final @c summary chunk starts with 0x00
  byte (see @ref summary).

  In the future, sequence numbers will be used to detect discontinuities in a
  backup stream. However, this is not currently implemented and a sequence
  number is always 0 (two 0x00 bytes). Upon restore, sequence numbers are
  currently ignored.

  Bits in the flags field are used as follows:
  - Bit 0 (BSTREAM_FLAG_LAST_CHUNK): Set if the chunk is the last data chunk
    for the table.
  - Bits 1-7. Reserved.

  Field @c table_num identifies the table to which this data chunk belongs. If
  @c table_num is 0 then this is a common data chunk not associated with any
  particular table. Otherwise (@c table_num-1, @c snapshot_num-1) are the
  catalog coordinates of the table as described in @ref catalog_pos_table.

  The format of @c data field is determined by the backup driver which created
  given snapshot, as specified in the corresponding @c snapshot_description
  chunk in the preamble.
*/

/**
  Write chunk with data from backup driver.

  @see @ref table_data
*/

int bstream_wr_data_chunk(backup_stream *s,
                          struct st_bstream_data_chunk *chunk)
{
  int ret= BSTREAM_OK;

  ASSERT(chunk);

  CHECK_WR_RES(bstream_wr_byte(s, chunk->snap_num + 1));
  CHECK_WR_RES(bstream_wr_int2(s, 0)); /* Sequence number - not used now. */
  CHECK_WR_RES(bstream_wr_byte(s, chunk->flags));
  CHECK_WR_RES(bstream_wr_num(s, chunk->table_num));
  CHECK_WR_RES(bstream_write_blob(s, chunk->data));
  CHECK_WR_RES(bstream_end_chunk(s));

wr_error:

  return ret;
}


/**
  The amount by which input buffer is increased if whole data chunk can't fit
  into it.
*/
#define DATA_BUF_STEP   (1024*1024)


/**
  Read chunk of data for restore driver.

  Blob @c chunk->data describes memory area where data from the chunk is saved.
  If this blob is non-empty when function is called, the data will be stored
  in the given area if it fits. If chunk data doesn't fit, or the blob was
  empty, the data is read into an internal buffer and the blob is updated to
  indicate where to find it. In the latter case the data will be overwritten
  upon next call to this function.

  @retval BSTREAM_OK   data chunk has been read, stream moved to next chunk
  @retval BSTREAM_EOS  data chunk has been read, no more chunks in the stream.
  @retval BSTREAM_EOC  not a data chunk (0x00 read). Note: stream doesn't have
                       to be at end of chunk!

  Return value @c BSTREAM_EOC indicates that all table data chunks have been
  read. The rest of the backup stream can contain image summary block.

  @see @ref table_data
*/

int bstream_rd_data_chunk(backup_stream *s,
                          struct st_bstream_data_chunk *chunk)
{
  blob              *buf= &s->data_buf;
  byte              *new_buf;
  blob              *envelope;
  blob              to_read;
  unsigned long int howmuch;
  unsigned int      seq_num;
  int               ret= BSTREAM_OK;

  ASSERT(chunk);

  CHECK_RD_RES(bstream_rd_byte(s, &chunk->snap_num));

  /*
    Saved snapshot numbers start from 1 - if we read 0 it means that this is not
    a table data chunk
  */
  if (chunk->snap_num == 0)
    return BSTREAM_EOC;
  else if (ret != BSTREAM_OK)
    return BSTREAM_ERROR;

  (chunk->snap_num)--;

  /* FIxME: handle sequence numbers. */

  CHECK_RD_OK(bstream_rd_int2(s, &seq_num));
  CHECK_RD_OK(bstream_rd_byte(s, &chunk->flags));
  CHECK_RD_OK(bstream_rd_num(s, &chunk->table_num));

  /*
    Read rest of the chunk data into provided buffer or the internal buffer
    @c s->data_buf if there is not enough space.

    envelope indicates which buffer we are using: provided or internal
    (initially provided).
  */

  envelope= &chunk->data;
  to_read= *envelope;      /* How much space is available for reading bytes. */

  /*
    In the following loop we call bstream_read_part() until we hit end of the
    chunk. If there is no more space to fit the data, buffer is enlarged
    (reallocated) and the bytes which were read before are copied into the
    new buffer.
  */

  while (ret == BSTREAM_OK)
  {
    /* Read bytes until current buffer is full or end of chunk is reached. */
    while (ret == BSTREAM_OK && (to_read.end > to_read.begin))
      ret= bstream_read_part(s, &to_read, *envelope);

    if (ret == BSTREAM_OK)
    {
      /*
        We have filled-up the buffer - we need to enlarge it. We first compute
        how much data has been read so far.
      */

      ASSERT(to_read.begin >= envelope->begin);
      howmuch= (unsigned long)(to_read.begin - envelope->begin);

      /* If there is not enough space in the internal buffer, enlarge it. */
      if ( (buf->begin + howmuch) >= buf->end )
      {
        new_buf= bstream_alloc(howmuch + DATA_BUF_STEP);

        if (!new_buf)
          return BSTREAM_ERROR;

        /* Copy data from old buffer to the new one. */
        if (buf->begin && (buf->end > buf->begin))
          memmove(new_buf, buf->begin, buf->end - buf->begin);
        bstream_free(buf->begin);

        buf->begin= new_buf;
        buf->end= buf->begin + howmuch + DATA_BUF_STEP;
      }

      /* If we were using the provided buffer, switch to internal one. */
      if (envelope == &chunk->data)
      {
        memmove(buf->begin, chunk->data.begin, howmuch);
        envelope= buf;
      }

      /* Update chunk->data to point to the new buffer. */
      chunk->data= *buf;

      /* Update to_read blob to indicate free space left. */
      to_read.begin= buf->begin + howmuch;
      to_read.end= buf->end;
    }
  }

  if (ret == BSTREAM_ERROR)
    return BSTREAM_ERROR;

  /* We have read all data from the chunk - record where it ends. */

  chunk->data.end= to_read.begin;

  /* Move to next chunk. */

  CHECK_RD_RES(bstream_next_chunk(s));

rd_error:

  return ret;
}


/**
  @page image_layer

  @section summary Summary Format

  The summary chunk stores information which are known only at the end of backup
  process. Specifically the binary log coordinates of image's validity point,
  which enable point-in-time recovery operations that combine use of the backup
  image and a log of modifications made after the validity point.

  The summary chunk appears at the end of the backup image, after all table data
  chunks. It always starts with 0x00 byte, to distinguish it from table data
  chunks which never begin with 0x00 (see @ref table_data).

@verbatim
  summary: 0x00 vp_time end_time binlog_coords binlog_group_coords
@endverbatim
  - 0x00 (1 byte): To distinguishing the summary from preceding table data
                   chunks.
  - @c vp_time (8 bytes): Time of validity point.
  - @c end_time (8 bytes): Time when backup ended.
  - @c binlog_coords (variable length): Binary log coordinates, stored as:
      - @c pos (4 bytes): Integer binary log position.
      - @c binlog_file (variable length): String binary log filename.
  - @c binlog_group_coords (variable length): Binary log group coordinates.
    The format is same as for @c binlog_coords. This field is reserved for
    future use.
*/

/**
  Save binlog position.

  @see @ref summary.
*/

int bstream_wr_binlog_pos(backup_stream *s, struct st_bstream_binlog_info pos)
{
  blob name;
  int  ret= BSTREAM_OK;

  name.begin= (byte *)pos.file;
  name.end= name.begin + (pos.file ? strlen(pos.file) : 0);
  CHECK_WR_RES(bstream_wr_int4(s, pos.pos));
  CHECK_WR_RES(bstream_wr_string(s, name));

wr_error:

  return ret;
}


/**
  Read binlog position.

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached

  @see @ref summary.
*/

int bstream_rd_binlog_pos(backup_stream *s, struct st_bstream_binlog_info *pos)
{
  blob name= {NULL, NULL};
  int  ret= BSTREAM_OK;

  CHECK_RD_OK(bstream_rd_int4(s, &pos->pos));
  CHECK_RD_RES(bstream_rd_string(s, &name));

  if (ret != BSTREAM_ERROR)
    pos->file= (char*)name.begin;

rd_error:

  return ret;
}


/**
  Write backup image summary.

  This function assumes that all members of the header (such as @c vp_time) are
  already filled. It also stores the 0x00 byte separating summary from the
  preceding table data chunks.

  @see @ref summary
*/

int bstream_wr_summary(backup_stream *s, struct st_bstream_image_header *hdr)
{
  int ret= BSTREAM_OK;

  CHECK_WR_RES(bstream_wr_byte(s, 0x00));
  CHECK_WR_RES(bstream_wr_time(s, hdr->vp_time));
  CHECK_WR_RES(bstream_wr_time(s, hdr->end_time));
  CHECK_WR_RES(bstream_wr_binlog_pos(s, hdr->binlog_info));
  CHECK_WR_RES(bstream_wr_binlog_pos(s, hdr->binlog_group));

wr_error:

  return ret;
}


/**
  Read backup image summary

  The information stored in the summary section is read and stored in the image
  header structure.

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached

  @see @ref summary
*/

int bstream_rd_summary(backup_stream *s, struct st_bstream_image_header *hdr)
{
  int ret= BSTREAM_OK;

  CHECK_RD_OK(bstream_rd_time(s, &hdr->vp_time));
  CHECK_RD_OK(bstream_rd_time(s, &hdr->end_time));
  CHECK_RD_OK(bstream_rd_binlog_pos(s, &hdr->binlog_info));
  CHECK_RD_RES(bstream_rd_binlog_pos(s, &hdr->binlog_group));

rd_error:

  return ret;
}


/*********************************************************************
 *
 *   WRITING/READING BASIC TYPES
 *
 *********************************************************************/

/**
  @page image_layer

  @section basic_data Basic Data Type Storage

  This section discusses storage for basic data types:
  - Fixed-length and variable-length integers
  - Strings
  - Times

  @subsection basic_data_int Integer Storage

  Only unsigned integers are stored.

  Fixed-length integers are stored as 1-byte, 2-byte, or 4-byte values, least
  significant byte first.

  Variable-length integers are stored as a sequence of bytes, each byte
  storing 7 bits (least significant bits first). The most significant bit in
  a byte is set if more bytes follow. There is no maximum number of bytes that
  a variable-length integer can contain. A program that reads them must be
  prepared to deal with arbitrarily large values, or to throw an error if
  it encounters a value larger than what it can handle.

  Example: The binary number 100101011101011010110100011110 will be split
  into 7-bit groups 10 0101011 1010110 1011010 0011110 and then stored in
  5 bytes (least significant bytes first): 0x9E 0xDA 0xD6 0xAB 0x02
*/

/** Write single byte to backup stream */

int bstream_wr_byte(backup_stream *s, unsigned short int x)
{
  byte buf= x & 0xFF;
  blob b;

  b.begin= &buf;
  b.end=   &buf + 1;

  return bstream_write_part(s, &b, b);
}


/**
  Read one byte from backup stream.

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached
*/

int bstream_rd_byte(backup_stream *s, unsigned short int *x)
{
  byte buf;
  blob b;
  int  ret;

  b.begin= &buf;
  b.end=   &buf + 1;

  ret= bstream_read_part(s, &b, b);

  if (b.begin == b.end)
  {
    *x= buf;
    return ret;
  }
  else return BSTREAM_ERROR;
}


/**
  Write 2 byte unsigned number. Least significant byte
  comes first.
*/

int bstream_wr_int2(backup_stream *s, unsigned int x)
{
  byte buf[2];
  blob b;

  buf[0]= x & 0xFF;
  buf[1]= (x >> 8) & 0xFF;

  b.begin= buf;
  b.end= buf+2;

  return bstream_write_blob(s, b);
}


/**
  Read 2 byte unsigned number.

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached
*/

int bstream_rd_int2(backup_stream *s, unsigned int *x)
{
  byte buf[2];
  blob b;
  int  ret;

  b.begin= buf;
  b.end=   buf+2;

  ret= bstream_read_blob(s, b);
  if (ret == BSTREAM_ERROR)
    return BSTREAM_ERROR;

  *x = buf[0] + (buf[1] << 8);

  return ret;
}


/** Write 4 byte unsigned number. Least significant bytes come first. */

int bstream_wr_int4(backup_stream *s, unsigned long int x)
{
  byte buf[4];
  blob b;
  int  i;

  b.begin= buf;
  b.end=   buf+4;

  for (i= 0; i < 4; i++)
  {
    buf[i]= get_byte_ulong((x & 0xFF));
    x >>= 8;
  }

  return bstream_write_blob(s, b);
}


/**
  Read 4 byte unsigned number.

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached
*/

int bstream_rd_int4(backup_stream *s, unsigned long int *x)
{
  byte buf[4];
  blob b;
  int  ret;

  b.begin= buf;
  b.end=   buf+4;

  ret= bstream_read_blob(s, b);
  if (ret == BSTREAM_ERROR)
    return BSTREAM_ERROR;

  *x = (unsigned long int)buf[0];
  *x += ((unsigned long int)buf[1] << 8);
  *x += ((unsigned long int)buf[2] << 2*8);
  *x += ((unsigned long int)buf[3] << 3*8);

  return ret;
}


/** Write 8 byte unsigned number. Least significant bytes come first. */

int bstream_wr_int8(backup_stream *s, unsigned long long x)
{
  byte buf[8];
  blob b;
  int  i;

  b.begin= buf;
  b.end=   buf+8;

  for (i= 0; i < 8; i++)
  {
    buf[i]= get_byte_ulong((x & 0xFF));
    x >>= 8;
  }

  return bstream_write_blob(s, b);
}


/**
  Read 8 byte unsigned number.

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached
*/

int bstream_rd_int8(backup_stream *s, unsigned long long *x)
{
  byte buf[8];
  blob b;
  int  ret;

  b.begin= buf;
  b.end=   buf+8;

  ret= bstream_read_blob(s, b);
  if (ret == BSTREAM_ERROR)
    return BSTREAM_ERROR;

  *x = (unsigned long long)buf[0];
  *x += ((unsigned long long)buf[1] << 8);
  *x += ((unsigned long long)buf[2] << 2*8);
  *x += ((unsigned long long)buf[3] << 3*8);
  *x += ((unsigned long long)buf[4] << 4*8);
  *x += ((unsigned long long)buf[5] << 5*8);
  *x += ((unsigned long long)buf[6] << 6*8);
  *x += ((unsigned long long)buf[7] << 7*8);

  return ret;
}


/**
  Write number using variable length format.

  Number is stored as a sequence of bytes, each byte storing 7 bits. The most
  significant bit in a byte tells if there are more bytes to follow
  (if it is set) or if current byte is the last one (if it is not set).
  The bits are saved starting with least significant ones.

  @see @ref basic_data_int
*/

int bstream_wr_num(backup_stream *s, unsigned long int x)
{
  int  ret= BSTREAM_OK;
  byte b;

  do {
    b= get_byte_ulong((x & 0x7F));
    CHECK_WR_RES(bstream_wr_byte(s, b | ( x>0x7F ? 0x80: 0)));
    x >>= 7;
  } while (x);

wr_error:

  return ret;
}


/**
  Read number saved using variable length format.

  @retval BSTREAM_ERROR  Error while reading or number doesn't fit in unsigned
                         long variable
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached

  @see @ref basic_data_int
*/

int bstream_rd_num(backup_stream *s, unsigned long int *x)
{
  unsigned short int b;
  unsigned int       i=0;
  int                ret= BSTREAM_OK;

  *x= 0;

  do {

    ret= bstream_rd_byte(s, &b);

    if (ret == BSTREAM_ERROR)
      return BSTREAM_ERROR;

    *x += (b & 0x7F) << (7*(i++));

  } while ((b & 0x80) && (i < sizeof(unsigned long int)));

  return (b & 0x80) ? BSTREAM_ERROR : ret;
}


/**
  @page image_layer

  @subsection basic_data_string String Storage

  Strings are stored as counted strings: A byte count indicating the length of
  the string, followed by the bytes in the string. The length is stored as a
  variable-length integer using the encoding just described.

  Strings are encoded using a universal encoding listed as the first item in
  the @c charsets list in the @c catalog_header chunk.

  An empty string (or "null string") is represented as a single 0x00 byte.

   Example: "abcd" is encoded as 0x04 0x61 0x62 0x63 0x64
   Example: "" (the empty string) is encoded as 0x00
*/

/**
  Write a string.

  @see @ref basic_data_string
*/

int bstream_wr_string(backup_stream *s, bstream_blob str)
{
  int ret= BSTREAM_OK;

  ASSERT(str.end >= str.begin);
  CHECK_WR_RES(bstream_wr_num(s, (unsigned long)(str.end - str.begin)));
  CHECK_WR_RES(bstream_write_blob(s, str));

wr_error:

  return ret;
}


/**
  Read a string.

  New memory is allocated (with @c bstream_alloc() function) to accommodate the
  string. Blob @c str is updated to point at the memory where string is stored.
  The byte after last byte of the string is set to 0.

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached

  @note
  Caller of this function is responsible for freeing memory allocated to store
  the string.

  @see @ref basic_data_string
*/

int bstream_rd_string(backup_stream *s, bstream_blob *str)
{
  int               ret= BSTREAM_OK;
  unsigned long int len;

  ret= bstream_rd_num(s, &len);

  if (len == 0)
  {
    str->begin= str->end= NULL;
    return ret;
  }

  if (ret != BSTREAM_OK)
    return BSTREAM_ERROR;

  str->begin= bstream_alloc(len+1);
  if (!str->begin)
    return BSTREAM_ERROR;
  str->end= str->begin + len;
  *str->end= '\0';

  return bstream_read_blob(s, *str);
}


/**
  @page image_layer

  @subsection basic_data_time Time Storage

 Times are stored in UTC. A time value is a 64 bit unsigned int with
 LSB endianness which will work on all platforms.

  A time value may consist entirely of 0x00 bytes, which means "no date."
*/

/**
  Write time entry
*/

int bstream_wr_time(backup_stream *s, time_t time)
{
  return bstream_wr_int8(s, (unsigned long long)time);
}


/**
  Read time entry

  @retval BSTREAM_ERROR  Error while reading
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached
*/

int bstream_rd_time(backup_stream *s, time_t *time)
{
  unsigned long long time_8;

  int ret= bstream_rd_int8(s, &time_8);
  *time= (time_t) time_8;
  return ret;
}


/**
  @page image_layer

   @subsection basic_data_itype Object Type Encoding

   Object type encoding is used for @c db_item values in @c db_catalog chunks,
   and for object types in metadata item lists.

   Allowable object types are encoded as follows:
   - 1 = character set
   - 2 = user
   - 3 = privilege
   - 4 = database
   - 5 = table
   - 6 = view
   - 7 = stored procedure
   - 8 = stored function
   - 9 = event
   - 10 = trigger
   - 11 = tablespace

   Object types are encoded using two bytes. A value of 0 is not a valid object
   type, so a 0x00 0x00 sequence can be used to separate one object list from
   the next.
*/

/**
  Save object type.

  See @ref basic_data_itype for encoding of object types.

  @retval BSTREAM_OK   type was saved successfully
  @retval BSTREAM_ERROR error writing or attempt to save unknown type.
*/

int bstream_wr_item_type(backup_stream *s, enum enum_bstream_item_type type)
{
  switch (type) {
  case BSTREAM_IT_CHARSET:    return bstream_wr_int2(s, 1);
  case BSTREAM_IT_USER:       return bstream_wr_int2(s, 2);
  case BSTREAM_IT_PRIVILEGE:  return bstream_wr_int2(s, 3);
  case BSTREAM_IT_DB:         return bstream_wr_int2(s, 4);
  case BSTREAM_IT_TABLE:      return bstream_wr_int2(s, 5);
  case BSTREAM_IT_VIEW:       return bstream_wr_int2(s, 6);
  case BSTREAM_IT_SPROC:      return bstream_wr_int2(s, 7);
  case BSTREAM_IT_SFUNC:      return bstream_wr_int2(s, 8);
  case BSTREAM_IT_EVENT:      return bstream_wr_int2(s, 9);
  case BSTREAM_IT_TRIGGER:    return bstream_wr_int2(s, 10);
  case BSTREAM_IT_TABLESPACE: return bstream_wr_int2(s, 11);
  case BSTREAM_IT_LAST:       return bstream_wr_int2(s, 0);
  default: return BSTREAM_ERROR;
  }
}


/**
  Read object type.

  See @ref basic_data_itype for encoding of object types.

  @retval BSTREAM_ERROR  Error while reading or non-recognized type found.
  @retval BSTREAM_OK     Read successful
  @retval BSTREAM_EOC    Read successful and end of chunk has been reached
  @retval BSTREAM_EOS    Read successful and end of stream has been reached
*/

int bstream_rd_item_type(backup_stream *s, enum enum_bstream_item_type *type)
{
  int          ret;
  unsigned int x;

  ret= bstream_rd_int2(s, &x);

  if (ret == BSTREAM_ERROR)
    return BSTREAM_ERROR;

  switch (x) {
  case 0:  *type= BSTREAM_IT_LAST; break;
  case 1:  *type= BSTREAM_IT_CHARSET; break;
  case 2:  *type= BSTREAM_IT_USER; break;
  case 3:  *type= BSTREAM_IT_PRIVILEGE; break;
  case 4:  *type= BSTREAM_IT_DB; break;
  case 5:  *type= BSTREAM_IT_TABLE; break;
  case 6:  *type= BSTREAM_IT_VIEW; break;
  case 7:  *type= BSTREAM_IT_SPROC; break;
  case 8:  *type= BSTREAM_IT_SFUNC; break;
  case 9:  *type= BSTREAM_IT_EVENT; break;
  case 10: *type= BSTREAM_IT_TRIGGER; break;
  case 11: *type= BSTREAM_IT_TABLESPACE; break;
  default: return BSTREAM_ERROR;
  }

  return ret;
}

