#ifndef BACKUP_STREAM_V1_
#define BACKUP_STREAM_V1_

#include <time.h>

/**
  @file

  @brief
  Backup stream library API for version 1 of stream format.

  This file declares functions and data types used to read and write backup
  stream using version 1 of backup stream format.
*/

/*********************************************************************
 *
 *   BASIC TYPES
 *
 *********************************************************************/

/** Definition for backup stream byte type. */
typedef unsigned char bstream_byte;

/**
  Describes continuous area of memory.

  The @c begin member points at the first byte in the area, @c at one byte
  after the last byte of the area. Thus, blob @c b contains exactly
  <code>b.end - b.begin</code> bytes and is empty if
  <code>b.begin == b.end</code>. A null blob is a blob @c b with
  <code>b.begin == NULL</code>.
*/

struct st_blob
{
  bstream_byte *begin; /**< First byte of the blob. */
  bstream_byte *end;   /**< One byte after the last byte of the blob. */
};

/** Definition for backup stream blob structure. */
typedef struct st_blob bstream_blob;

/**
  Describes position of an event in MySQL server's binary log.

  The event is identified by the name of the file in which it is stored and
  the position within that file.
*/

struct st_bstream_binlog_info
{
  char              *file;  /**< Binlog file storing the event. */
  unsigned long int pos;    /**< Position (offset) within the file. */
};

/** Definition for backup stream structure. */
/** @note The struct st_backup_stream is defined below. */
typedef struct st_backup_stream backup_stream;

/** Codes returned by backup stream functions. */
enum enum_bstream_ret_codes {
  BSTREAM_OK=0,  /**< Success. */
  BSTREAM_EOC,   /**< End of chunk detected. */
  BSTREAM_EOS,   /**< End of stream detected. */
  BSTREAM_ERROR  /**< Error. */
};

/*********************************************************************
 *
 *   TYPES FOR IMAGE HEADER
 *
 *********************************************************************/

/**
  Describes version of MySQL server.

  For example, if server has version "5.2.32-online-backup" then:
  major   = 5
  minor   = 2
  release = 32
  extra   = "5.2.32-online-backup"
*/

struct st_server_version
{
  unsigned short int  major;   /**< Major version number. */
  unsigned short int  minor;   /**< Minor version number. */
  unsigned short int  release; /**< Release number. */
  bstream_blob        extra;   /**< Extra version suffix. */
};

/**
  Describes version of storage engine.
*/

struct st_bstream_engine_info
{
  bstream_blob        name;   /**< Name of the storage engine. */
  unsigned short int  major;  /**< Version number (major). */
  unsigned short int  minor;  /**< Version number (minor). */
};

/** Types of table data snapshots. */
enum enum_bstream_snapshot_type {
  BI_NATIVE,  /**< Created by native backup driver of a storage engine. */
  BI_DEFAULT, /**< Created by built-in blocking backup driver. */
  /** Created by built-in driver using consistent read transaction. */
  BI_CS,
  BI_NODATA,  /**< Created by built-in nodata backup driver. */
  BI_UNKNOWN  /**< Error condition used for testing error handling. */
};

/** Describes table data snapshot. */

struct st_bstream_snapshot_info
{
  unsigned int            version;      /**< Snapshot format version. */
  enum enum_bstream_snapshot_type type; /**< Type of the snapshot. */
  /** Snapshot options (not used currently). */
  unsigned int            options;
  unsigned long int       table_count; /**< Number of tables in the snapshot. */
  /**
    In case of native snapshot, information about storage engine
    which created it.
  */
  struct st_bstream_engine_info  engine;
};

/**
  Extension of st_bstream_snapshot_info describing snapshot created by a native
  backup driver.
*/

struct st_native_snapshot_info
{
  struct st_bstream_snapshot_info  base;   /**< Standard snapshot data. */
};

/** Information about backup image. */

struct st_bstream_image_header
{
  unsigned int      version;    /**< Image's format version number. */
  /** version of the server which created the image. */
  struct st_server_version  server_version;
  unsigned int      flags;      /**< Image options. */
  time_t            start_time; /**< Time when backup operation started. */
  time_t            end_time;   /**< Time when backup operation completed. */
  time_t            vp_time;    /**< Time of the image's validity point. */

  /*
    If server which created backup image had binary log enabled, the image
    contains information about the position inside the log corresponding to
    the validity point time.
   */

  /** Position of the last binlog entry at the VP time. */
  struct st_bstream_binlog_info  binlog_info;
  /** Start of the last binlog event group at the VP time. */
  struct st_bstream_binlog_info  binlog_group;

  /** Number of table data snapshots in the image. */
  unsigned short int        snap_count;

  /**
    Descriptions of table data snapshots.

    We have at most 256 snapshots because their number is stored using one byte.
   */
  struct st_bstream_snapshot_info snapshot[256];
};

/* Flags */

/**
  Byte order of the server which created backup image.

  If set, informs that backup image was created on big-endian server. This might
  be useful to detect problems, if backup engines are not endian-agnostic.
*/
#define BSTREAM_FLAG_BIG_ENDIAN     (1U<<1)

/**
  Informs if image stores binlog position.

  If this flag is not set, the @c binlog_pos and @c binlog_group entries in
  the image should be ignored.
 */
#define BSTREAM_FLAG_BINLOG         (1U<<2)



/*********************************************************************
 *
 *   TYPES FOR DESCRIBING BACKUP ITEMS
 *
 *********************************************************************/

/**
  Types of items stored in a backup image.

  @note Not all of these types are supported currently.
*/
enum enum_bstream_item_type {
   BSTREAM_IT_CHARSET,
   BSTREAM_IT_USER,
   BSTREAM_IT_PRIVILEGE,
   BSTREAM_IT_DB,
   BSTREAM_IT_TABLE,
   BSTREAM_IT_VIEW,
   BSTREAM_IT_SPROC,
   BSTREAM_IT_SFUNC,
   BSTREAM_IT_EVENT,
   BSTREAM_IT_TRIGGER,
   BSTREAM_IT_TABLESPACE,
   BSTREAM_IT_LAST
};

/** Common data about backup image item. */

struct st_bstream_item_info
{
  enum enum_bstream_item_type type;  /**< Type of the item. */
  bstream_blob name;                 /**< Name of the item. */
  unsigned long int pos;  /**< Position of the item in image's catalogue. */
};

/**
  Describes tablespace item.

  Currently no data specific to tablespace items is used.
*/

struct st_bstream_ts_info
{
  struct st_bstream_item_info  base;  /**< The base of the info class. */
};

/**
  Describes database item.

  Currently no data specific to database items are used.
*/

struct st_bstream_db_info
{
  struct st_bstream_item_info  base;  /**< The base of the info class. */
};

/** 
  Describes item which sits inside a database. 
 
  @note The fields snap_num, pos are used for triggers only and are NULL
  otherwise. The fields contain the coordinates of the table on which the
  trigger fires.
*/

struct st_bstream_dbitem_info
{
  struct st_bstream_item_info  base;/**< Data common to all items. */
  struct st_bstream_db_info    *db; /**< Database to which this item belongs. */
  unsigned short int  snap_num;     /**< Snapshot where table data is stored. */
  unsigned long int   pos;          /**< Table position within the snapshot. */
};

/**
  Describes a table item.

  Table is a per-database item. Additionally we store information about the
  snapshot which contains its data.
*/

struct st_bstream_table_info
{
  struct st_bstream_dbitem_info  base;  /**< Data common to all per-db items. */
  unsigned short int  snap_num;   /**< Snapshot where table's data is stored. */
};


/*
  The following constants denote additional backup item categories. Items of
  different types can fall into one of these categories. Thus the categories
  are something different than item types and therefore are not listed
  inside enum_bstream_item_type but defined separately.
*/

/** Definition of backup stream global item index. */
#define BSTREAM_IT_GLOBAL    BSTREAM_IT_LAST
/** Definition of backup stream per database item index. */
#define BSTREAM_IT_PERDB     (BSTREAM_IT_LAST+1)


/*************************************************************************
 *
 *   STRUCTURE FOR WRITING/READING TABLE DATA
 *
 *************************************************************************/

/** Describe chunk of data from backup driver or for restore driver. */

struct st_bstream_data_chunk
{
  unsigned long int  table_num; /**< Table to which this data belongs. */
  bstream_blob       data;      /**< The data. */
  /** Flags to be saved together with the chunk. */
  unsigned short int flags;
  unsigned short int snap_num;  /**< Which snapshot this chunk belongs to. */
};

/** Indicates that given chunk is the last chunk of data for a given table. */
#define BSTREAM_FLAG_LAST_CHUNK  1


/*************************************************************************
 *
 *   FUNCTIONS FOR WRITING BACKUP IMAGE
 *
 *************************************************************************/

int bstream_wr_preamble(backup_stream*, struct st_bstream_image_header*);
int bstream_wr_data_chunk(backup_stream*,
                          struct st_bstream_data_chunk*);
int bstream_wr_summary(backup_stream *s, struct st_bstream_image_header *hdr);

/**
 Flush backup stream`s output buffer to the output stream.

 This empties the output buffer.

 @param[in]  s  The backup stream to flush.

 @returns Status of operation.
*/
int bstream_flush(backup_stream* s);

/*********************************************************************
 *
 *   FUNCTIONS FOR READING BACKUP IMAGE
 *
 *********************************************************************/

int bstream_rd_preamble(backup_stream*, struct st_bstream_image_header*);
int bstream_rd_data_chunk(backup_stream*,
                          struct st_bstream_data_chunk*);
int bstream_rd_summary(backup_stream *s, struct st_bstream_image_header *hdr);

/* Parts of preamble. */

int bstream_rd_header(backup_stream*, struct st_bstream_image_header*);
int bstream_rd_catalogue(backup_stream*, struct st_bstream_image_header*);
int bstream_rd_meta_data(backup_stream *, struct st_bstream_image_header*);

/* Basic types. */

int bstream_rd_time(backup_stream*, time_t*);
int bstream_rd_string(backup_stream*, bstream_blob*);
int bstream_rd_num(backup_stream*, unsigned long int*);
int bstream_rd_int8(backup_stream*, unsigned long long*);
int bstream_rd_int4(backup_stream*, unsigned long int*);
int bstream_rd_int2(backup_stream*, unsigned int*);
int bstream_rd_byte(backup_stream*, unsigned short int*);

int bstream_next_chunk(backup_stream*);


/*********************************************************************
 *
 *   DEFINITION OF BACKUP STREAM STRUCTURE
 *
 *********************************************************************/

/**
  Structure defining base I/O operations.

  Abstract stream is a way of defining the final destination/source of
  the bytes being written to/read from a backup stream. This is done by
  storing pointers to functions performing basic I/O operations inside
  this structure.

  Low-level @c write(), @c read() and @c forward() functions should return
  BSTREAM_OK on success and BSTREAM_ERROR on error. Additionally, @c read()
  should return BSTREAM_EOS if there is no more data in the input stream.

  Functions @c write() and @c read() should behave like functions
  @c bstream_write_part() and @c bstream_read_part(), respectively. The latter
  are documented in stream_v1_transport.c. The only difference is that low-level
  @c read() doesn't deal with chunk boundaries and never returns BSTREAM_EOC.
  However, it should detect end of stream and return BSTREAM_EOS in that case.
 */

struct st_abstract_stream
{
  /** Pointer to write method. */
  int (*write)(void*, bstream_blob*, bstream_blob);
  /** Pointer to read method. */
  int (*read)(void*, bstream_blob*, bstream_blob);
  /** Pointer to forward method. */
  int (*forward)(void*, unsigned long int*);
};

/**
  Structure describing backup stream`s input or output buffer.
  @see stream_v1_carrier.c
*/

struct st_bstream_buffer
{
  bstream_byte  *begin;  /**< Pointer to start of stream buffer. */
  bstream_byte  *pos;    /**< Current position in buffer. */
  bstream_byte  *header; /**< Pointer to the header of the next fragment. */
  bstream_byte  *end;    /**< Pointer to the end of the current stream block. */
};

/**
  Structure describing state of a backup stream.
*/

struct st_backup_stream
{
  struct st_abstract_stream stream; /**< Stream metadata. */
  unsigned long int block_size;     /**< Block size. */
  short int first_block;        /**< Writing/reading first block of a stream. */
  short int init_block_count;       /**< Initial block count. */
  /** Enumeration for state of the stream. */
  enum { CLOSED,         /**< Stream has been closed. */
         NORMAL,         /**< Normal operation. */
         NEW_CHUNK,      /**< New chunk was opened but no bytes written yet. */
         LAST_FRAGMENT,  /**< Reading last fragment of a chunk. */
         EOS,            /**< End of stream detected. */
         ERROR } state;  /**< State of the stream. */
  /** Enumeration for mode of stream (read, write). */
  enum { READING, WRITING } mode;   /**< Current mode. */
  struct st_bstream_buffer buf;     /**< Stream buffer. */
  bstream_blob mem;                 /**< Pointer to location in buffer. */
  bstream_blob data_buf;            /**< Pointer to data buffer. */
};

int bstream_open_wr(backup_stream*, unsigned long int, unsigned long int);
int bstream_open_rd(backup_stream*, unsigned long int);
int bstream_close(backup_stream*);

#endif /*BACKUP_V1_*/
