/* Copyright (C) 2004-2007 MySQL AB, 2008-2009 2008 Sun Microsystems, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/

/**
   @file

   @brief Contains methods to implement the Backup Shared/Exclusive
   Locks service.

   This file contains methods that allow getting and releasing Shared
   and Exclusive locks. The service follows the standard lock
   compatibility matrix for Shared/Exclusive locking:

     | S  X
   --+-----
   S | Y  N
   X | N  N

   Where S=Shared lock and X=Exclusive lock

   The following properties are enforced:

   * A thread requesting an S lock gets the lock immediately unless 
     - another thread already got the X lock, or 
     - another thread has requested the X lock and is waiting for the release
       of S locks to get it

     In both these, the request is blocked until the X lock has been
     released or a timeout occurs

   * A thread requesting an X lock gets the lock immediately unless 
     - one or more threads has S locks, in which case the request is
       blocked until all existing S locks have been released, or
     - another thread has (requested) the X lock. Since only one
       BACKUP/RESTORE process can exist at any time, X lock requests
       return with an error if the X lock has already been set.


   The Backup Shared/Exclusive Lock service is used by BACKUP for two things:
    - Backup Metadata Locking
    - Backup Commit Blocker

   Hence, two Backup_sx_lock objects are defined at MySQL server startup:

   Backup_sx_lock *BML_instance
   Backup_sx_lock *BCB_instance

   BACKUP METADATA LOCKING
   -----------------------
   For the duration of BACKUP/RESTORE, most DDL statements are
   disallowed. This is implemented by making BACKUP acquire an exclusive
   lock for the duration that must be protected, and making DDL
   statements aquire Shared locks.

   The list of statements that should obey BML is as follows:

    DROP   DATABASE/TABLE/VIEW/FUNCTION/PROCEDURE/EVENT/TRIGGER/INDEX
    DROP   USER/TABLESPACE
    CREATE DATABASE/TABLE/VIEW/FUNCTION/PROCEDURE/EVENT/TRIGGER/INDEX
    ALTER  DATABASE/TABLE/VIEW/FUNCTION/PROCEDURE/EVENT/TABLESPACE
    RENAME TABLE/USER
    GRANT/REVOKE
    TRUNCATE/OPTIMIZE/REPAIR TABLE

   The parser (mysql_execute_command() in sql_parse.cc) arranges for calls to
   get_shared_lock() and release_shared_lock() for these statements.

   BACKUP COMMIT BLOCKER
   ---------------------
   To get a transaction consistent image of all tables in all
   databases being backed up, BACKUP needs to disallow commits for a
   short period of time to synchronize data from all storage
   engines. This is implemented by making BACKUP acquire an exclusive lock
   for the duration of the synchronization phase, and by making all
   commits of transactions that have modified data acquire Shared locks.

   ha_commit_trans in handler.cc arranges for calls to get_shared_lock()
   and release_shared_lock().
*/

#include "bml.h"
#include "debug_sync.h"

Backup_sx_lock::Backup_sx_lock()
{
  mysql_mutex_init(INSTRUMENT_ME, &LOCK_mutex, MY_MUTEX_INIT_FAST);
  mysql_cond_init(INSTRUMENT_ME, &COND_no_slock, NULL);
  mysql_cond_init(INSTRUMENT_ME, &COND_no_xlock, NULL);
  xlock_set= NULL;
  slock_count= 0;
}

Backup_sx_lock::~Backup_sx_lock()
{
  mysql_mutex_destroy(&LOCK_mutex);
  mysql_cond_destroy(&COND_no_slock);
  mysql_cond_destroy(&COND_no_xlock);
}

/**
   Check if the exclusive lock has been set. If so, wait until it has
   been released. When the exclusive lock has been released, acquire
   a Shared lock by increasing the Shared Lock Counter (slock_count).

   @param[in] thd        The THD object from the caller.

   @note: A successful call to get_shared_lock() must be matched by
   release_shared_lock().

   @return Operation status
     @retval TRUE if a Shared Lock was successfully acquired
     @retval FALSE if timeout occurred while waiting for the exclusive
             lock to be released
*/
my_bool Backup_sx_lock::get_shared_lock(THD *thd)
{
  int ret = 0;
  struct timespec ddl_timeout;
  DBUG_ENTER("Backup_sx_lock::get_shared_lock()");

  set_timespec(ddl_timeout, thd->variables.backup_wait_timeout);

  /*
    Check whether an exclusive lock is set. If yes, wait for release
    which is signalled with COND_no_xlock.
  */
  mysql_mutex_lock(&LOCK_mutex);
  thd->enter_cond(&COND_no_xlock, &LOCK_mutex,
                  "Backup_sx_lock: waiting until exclusive lock is released");
  while (xlock_set && xlock_set!=thd && !thd->killed && (ret == 0))
  {
    if (thd->variables.backup_wait_timeout == 0)
    {
      ret = -1;
    }
    else
    {
      /*
        The below sync point fires if the exclusive lock is set.

        WARNING: Be careful when using WAIT_FOR with this sync point.
        We hold LOCK_mutex here.
      */
      DEBUG_SYNC(thd, "wait_if_backup_exclusive_locked");
      ret= mysql_cond_timedwait(&COND_no_xlock, &LOCK_mutex,
                                &ddl_timeout);
    }
  }

  if (thd->killed)
  {
    /* Releases LOCK_mutex */
    thd->exit_cond("Backup_sx_lock: Thread was killed");
    DBUG_RETURN(FALSE);
  }

  if (ret == 0)
    slock_count++;
  else
    my_error(ER_DDL_TIMEOUT, MYF(0), thd->query());

  thd->exit_cond("Backup_sx_lock: entered");

  DEBUG_SYNC(thd, "backup_got_shared_lock");
  DBUG_RETURN(ret == 0);
}

/**
   Release the Shared lock by decreasing slock_count
*/
void Backup_sx_lock::release_shared_lock()
{
  DBUG_ENTER("Backup_sx_lock::release_shared_lock()");
  mysql_mutex_lock(&LOCK_mutex);

  DBUG_ASSERT(slock_count > 0);
  slock_count--;

  if (slock_count == 0 && xlock_set) 
    mysql_cond_signal(&COND_no_slock);

  mysql_mutex_unlock(&LOCK_mutex);
  DBUG_VOID_RETURN;
}

/**
   Block new shared lock requests by setting the exclusive lock. Wait
   until all currently active shared locks have been released and then
   return.

   @param[in] thd        The THD object from the caller.

   @note: A successful call to get_exclusive_lock() must be matched by
   release_exclusive_lock().

   @return Operation status
     @retval TRUE if an exclusive lock was successfully acquired
     @retval FALSE otherwise
*/
my_bool Backup_sx_lock::get_exclusive_lock(THD *thd)
{
  DBUG_ENTER("Backup_sx_lock::get_exclusive_lock()");

  //part 1 - get the exclusive lock
  mysql_mutex_lock(&LOCK_mutex);
  if (unlikely(xlock_set))
  {
    /*
       Should never happen since only one backup/restore operation is
       allowed at any time
    */
    thd->exit_cond("Backup_sx_lock: An exclusive lock has already been set");
    DBUG_RETURN(FALSE);
  }

  xlock_set= thd; // Blocks new attempts at getting a shared lock

  //part 2 - wait for all shared locks to be released
  thd->enter_cond(&COND_no_slock, &LOCK_mutex,
                  "Backup_sx_lock: waiting for release of all Shared locks");
  while (slock_count != 0 && !thd->killed)
  {
    /*
      The below sync point fires if a shared lock is set.

      WARNING: Be careful when using WAIT_FOR with this sync point.
      We hold LOCK_mutex here.
    */
    DEBUG_SYNC(thd, "wait_backup_exclusive_lock");
    /* wait in queue that is activated when slock_count drops to 0 */
    mysql_cond_wait(&COND_no_slock, &LOCK_mutex);
  }
  thd->exit_cond("Backup_sx_lock: Exclusive lock set (or thread killed)");

  if (thd->killed)
  {
    /* This thread alread got X lock; release it */
    release_exclusive_lock(thd);
    DBUG_RETURN(FALSE);
  }

  DEBUG_SYNC(thd, "after_backup_exclusivelock_set");
  DBUG_RETURN(TRUE);
}

/**
   Release previously obtained Exclusive lock.

   @param[in] thd        The THD object from the caller.

   Exclusive lock can be released only by the thread who has acquired it, as
   determined by @c thd parameter. If @c xlock_set equals @c thd then 
   @c xlock_set is set to NULL indicating that the lock is released.
*/
void Backup_sx_lock::release_exclusive_lock(THD *thd)
{
  DBUG_ENTER("Backup_sx_lock::release_exclusive_lock()");
  mysql_mutex_lock(&LOCK_mutex);
  if (xlock_set == thd)
  {
    xlock_set= NULL;
    mysql_cond_broadcast(&COND_no_xlock);
  }
  mysql_mutex_unlock(&LOCK_mutex);
  DBUG_VOID_RETURN;
}

Backup_sx_lock *Backup_sx_lock::bml_instance= NULL;
Backup_sx_lock *Backup_sx_lock::bcb_instance= NULL;

/**
  Singleton method to get the Backup Metadata Lock instance
*/
Backup_sx_lock *Backup_sx_lock::get_BML_instance()
{
  if (bml_instance == NULL)
    bml_instance = new Backup_sx_lock();
  return bml_instance;
}

/**
  Singleton method to destroy the Backup Metadata Lock instance
*/
void Backup_sx_lock::destroy_BML_instance()
{
  delete bml_instance;
  bml_instance= NULL;
}


/**
  Singleton method to get the Backup Commit Blocker instance
*/
Backup_sx_lock *Backup_sx_lock::get_BCB_instance()
{
  if (bcb_instance == NULL)
    bcb_instance = new Backup_sx_lock();
  return bcb_instance;
}

/**
  Singleton method to destroy the Backup Commit Blocker instance
*/
void Backup_sx_lock::destroy_BCB_instance()
{
  delete bcb_instance;
  bcb_instance= NULL;
}

