#ifndef BML_INCLUDED
#define BML_INCLUDED

/* Copyright 2008-2009 Sun Microsystems, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/**
  @file

  Header file for Backup Metadata Lock.
 */

#include "mysql_priv.h"

/**
   @class Backup_sx_lock
 
   @brief Implements a simple Backup Shared/Exclusive Lock service
   used by BML (backup metadata locking) and Backup Commit Blocker.

   Two singleton instances are created of this class:
     * BML_instance, which is used by Backup to block concurrent 
       DDL statements
     * bcb_instance, which is used by Backup to block concurrent
       transactions from committing while backup performs the
       synchronization stage.

   Backup Metadata Locking works as follows:
     * A transaction that wants to modify the database metada by
       performing a DDL operation first needs to acquire a shared lock
       in the BML_instance. It does so by calling get_shared_lock(). A
       transaction that has taken a shared lock has to release it
       by calling release_shared_lock() before committing.
     * When backup want's to block DDL statements so that the database
       metadata remains unchanged for the duration of backup, it has
       to acquire the exclusive lock in the BML_instance. It does so
       by calling get_exclusive_lock(). Once metadata changes no
       longer need to be blocked, backup will release the exlusive
       lock by calling release_exclusive_lock().

   Backup Commit Blocking works as follows:
     * A transaction that wants to commit has to acquire the shared
       lock in the bcb_instance first. Once it has committed, the
       shared lock must be released.
     * When backup reaches the synchronization phase and needs to
       block commits, it acquires the exclusive lock. Once
       synchronization is completed, it releases the lock.


   Singleton Methods
   The creation of the locking instances is accomplished using the
   get_BML_instance() and get_BCB_instance() methods. These methods
   are called from mysqld.cc and creates and initializes all of the
   private mutex, condition, and controlling variables. The methods
   destroy_BML_instance() and destroy_BCB_instance() destroys the
   mutex and condition variables.

   Calling the Singletons

   To call the BML or BCB singletons, you must declare external
   variables to the global variables BML_instance and BCB_instance as
   shown below.

   @c extern Backup_sx_lock *BML_instance;
   @c extern Backup_sx_lock *BCB_instance;


   Calling methods on the singleton is accomplished using the
   BML_instance or BCB_instance variable such as: 

   @c BML_instance->get_shared_lock().

   @note: This class is currently only used in MySQL backup. If you would
   like to use it elsewhere and have questions, please contact
   Chuck Bell (cbell@mysql.com) for more details and how to setup
   a test case to test the Backup_sx_lock mechanism for your use.
*/
class Backup_sx_lock
{
 public:
  
  static Backup_sx_lock *get_BML_instance();
  static void destroy_BML_instance();
  static Backup_sx_lock *get_BCB_instance();
  static void destroy_BCB_instance();
  
  my_bool get_shared_lock(THD *thd);
  my_bool get_exclusive_lock(THD *thd);
  
  void release_shared_lock();
  void release_exclusive_lock(THD *thd);
  
 private:
  Backup_sx_lock();
  ~Backup_sx_lock();
  
  /// Mutex for protecting xlock_set and slock_count
  mysql_mutex_t LOCK_mutex;
  /// Signals THR_LOCK_lock mutex was released and slock_count dropped to zero
  mysql_cond_t COND_no_slock;
  /// Signals wlock was released
  mysql_cond_t COND_no_xlock;
  
  /// THD that has the X lock, or NULL if X lock is not set
  THD *xlock_set;
  /// Number of shared locks currently set
  int slock_count; 

  
  static Backup_sx_lock *bml_instance; ///< instance var for singleton 
  static Backup_sx_lock *bcb_instance; ///< instance var for singleton 
  
};
#endif /* BML_INCLUDED */
