/** @file

  Trigger errors in callback functions which are called from bstream
  library when writing/reading backup image.
*/

#include <tap.h>

#include "stream_test.h"
#include "catalog.h"

image_header  img_header;

typedef enum {WR,RD} op_type;

/*
  These helper functions look at callback counters and arrange for a specified
  callback function to report error when called for the 1-st, 2-nd, ... time.
  This happens during writing or reading of backup image.
*/

static int fail_iterator(op_type, item_type);
static int fail_db_iterator(op_type, uint);
static int fail_get_meta(item_type type);
static int fail_item(item_type type);
static int fail_catalog();
static int fail_alloc();

void run_test()
{
  uint i;
  uint planned;
  uint read_planned, write_planned;
  
  write_planned= 1 + global_count + 4 + db_count
                   + 1 + global_count + perdb_count;
 
  read_planned= 3 + global_count + perdb_count + 2;

  planned = read_planned+write_planned;

  plan(planned);

  init_catalog(&img_header);

  diag("Testing callback failures during writing of backup image");

  /*
    Do a clean write of backup image to fill callback_counts structure holding
    counters for callback functions.
  */
  reset_counts();
  ok(!(write_image(&img_header, "test_image.bst") || check_iterators()),
     "Counting callbacks");

  /* Failures in get_iterator */

  /* iterate over all global object types present in the catalogue */
  for (i=0; i < global_count; ++i)
    ok(fail_iterator(WR, objects_global[i].type),
       "Errors in bcat_iterator_get() callback (for global objects of type %d)",
       objects_global[i].type);

  /* cover remaining iterator types */
  ok(fail_iterator(WR, BSTREAM_IT_DB),
     "Errors in bcat_iterator_get() callback (for databases)");
  ok(fail_iterator(WR, BSTREAM_IT_GLOBAL),
     "Errors in bcat_iterator_get() callback (for all global objects)");
  ok(fail_iterator(WR, BSTREAM_IT_PERDB),
     "Errors in bcat_iterator_get() callback (for per-database objects)");

  /* test failures in database iterators (for each database in the catalogue) */
  for (i=0; i < db_count; ++i)
    ok(fail_db_iterator(WR, i),
       "Errors in bcat_db_iterator_get() callback (for database #%d)",i);

  /* Failures in get_item_create_query/data */

  ok(fail_get_meta(BSTREAM_IT_DB),
     "Errors in bcat_get_item_create_query/data() callback (for databases)");

  /* iterate over all global object types present in the catalogue */
  for(i=0; i < global_count; ++i)
    ok(fail_get_meta(objects_global[i].type),
       "Errors in bcat_get_item_create_query/data() callback"
       " (for objects of type %d)", objects_global[i].type);

  /* iterate over all per-database object types present in the catalogue */
  for(i=0; i < perdb_count; ++i)
    ok(fail_get_meta(objects_perdb[i].type),
       "Errors in bcat_get_item_create_query/data() callback"
       " (for objects of type %d)", objects_perdb[i].type);

  /* test errors in memory allocation */
  ok(fail_alloc(WR), "Errors in memory allocation");

  diag("Testing callback failures during reading of backup image");


  /* Do a clean read of the backup image to set the callback counters. */
  reset_counts();
  ok(!(read_image(&img_header, "test_image.bst")
       || check_iterators()),
     "Counting callbacks");

  /*
    The only iterator used when reading backup image is the database iterator.
  */
  ok(fail_iterator(RD, BSTREAM_IT_DB),
     "Errors in bcat_iterator_get() callback (for databases)");

  /* Failures in bcat_add/create_item */

  ok(fail_item(BSTREAM_IT_DB),
     "Errors in bcat_add/create_item() callback (for databases)");

  for(i=0; i < global_count; ++i)
    ok(fail_item(objects_global[i].type),
       "Errors in bcat_add/create_item() callbacks"
       " (for objects of type %d)", objects_global[i].type);

  for(i=0; i < perdb_count; ++i)
    ok(fail_item(objects_perdb[i].type),
       "Errors in bcat_add/create_item() callbacks"
       " (for objects of type %d)", objects_perdb[i].type);

  /* test failures in misc catalogue managing callbacks */
  ok(fail_catalog(), "Errors in bcat_reset/close() callbacks");

  /* test errors in memory allocation */
  ok(fail_alloc(RD), "Errors in memory allocation");

}


/**************************************************************************

  Implementation of helper functions.

**************************************************************************/

int fail_callback(op_type, int*, uint);

/**
  Read or write image triggering errors in bcat_iterator_get().

  @param[IN] rd_or_wr   tells whether image should be written or read; if
                        reading is specified, then "test_image.bst" backup
                        image should be created earlier.
  @param[IN] type       specifies which iterator should fail.

  @returns True if test went as expected, that is, read/write operations
  were aborted with error due to error in the callback call.

  @see @c fail_callback().
*/

static
int fail_iterator(op_type rd_or_wr, item_type type)
{
  int *ctr;

  DBUG_ASSERT(type < 128);

  ctr= &(callback_counts.iterator_get[type]);
  return fail_callback(rd_or_wr, ctr, *ctr);
}


/**
  Read or write image triggering errors in bcat_db_itearator_get().

  @param[IN] rd_or_wr   tells whether image should be written or read; if
                        reading is specified, then "test_image.bst" backup
                        image should be created earlier.
  @param[IN] db_num     specifies which databse's iterator should fail.

  @returns True if test went as expected, that is, read/write operations
  were aborted with error due to error in the callback call.

  @see @c fail_callback().
*/

static
int fail_db_iterator(op_type rd_or_wr, uint db_num)
{
  int *ctr;

  DBUG_ASSERT(db_num < db_count);

  ctr= &(callback_counts.db_iterator_get[db_num]);
  return fail_callback(rd_or_wr, ctr, *ctr);
}


/**
  Write image triggering errors in bcat_get_item_create_query/data().

  @param[IN] type  The type of item for which callback should fail.

  @returns True if test went as expected, that is, write operations
  were aborted with error due to error in the callback call.

  @see @c fail_callback().
*/

static
int fail_get_meta(item_type type)
{
  DBUG_ASSERT(type < 128);
  int ret1, ret2;
  int *ctr;

  ctr= &(callback_counts.get_serialization[type]);
  ret1= fail_callback(WR, ctr, *ctr);

  ctr= &(callback_counts.get_metadata[type]);
  ret2= fail_callback(WR, ctr, *ctr);

  return ret1 || ret2;
}


/**
  Read image triggering errors in bcat_add/create_item().

  @param[IN] type  The type of item for which callback should fail.

  @returns True if test went as expected, that is, read operations
  were aborted with error due to error in the callback call.

  @see @c fail_callback().
*/

static
int fail_item(item_type type)
{
  DBUG_ASSERT(type < 128);
  int ret1, ret2;
  int *ctr;

  ctr= &(callback_counts.add_item[type]);
  ret1= fail_callback(RD, ctr, *ctr);

  ctr= &(callback_counts.create_item[type]);
  ret2= fail_callback(RD, ctr, *ctr);

  return ret1 || ret2;
}


/**
  Read image triggering errors in bcat_reset/close().

  @returns True if test went as expected, that is, read operations
  were aborted with error due to error in the callback call.

  @see @c fail_callback().
*/

static
int fail_catalog()
{
  int ret1, ret2;
  int *ctr;

  ctr= &(callback_counts.cat_reset);
  ret1= fail_callback(RD, ctr, *ctr);

  ctr= &(callback_counts.cat_close);
  ret2= fail_callback(RD, ctr, *ctr);

  return ret1 || ret2;
}

/**
  Read/write image triggering memory allocation errors.

  @returns True if test went as expected, that is, read/write operations
  were aborted with error due to error in the callback call.

  @see @c fail_callback().
*/

static
int fail_alloc(op_type rd_or_wr)
{
  int *ctr= &(callback_counts.alloc);

  return fail_callback(rd_or_wr, ctr, *ctr);
}


/*
  Read or write backup image triggering errors in specified callback function.

  This function writes or reads backup image several times, each time arranging
  for a specified callback to report error. The error will be reported at n-th
  call to the callback function, where n ranges from @a howmuch to @a 1.

  Triggering of error is arranged by assigning negative value to the 
  corresponding callback counter (see documentation of @c callback_counts in
  catalog.h). 

  @param[IN] rd_or_wr   tells whether image should be written or read; if
                        reading is specified, then "test_image.bst" backup
                        image should be created earlier
  @param[IN] cntr       pointer to the counter in callback_counts structure,
                        corresponding to the callback which should be tested
  @param[IN] howmuch    specifies how many calls to the callback function will
                        be tested

  @returns True if the test went as expected, that is, write/read operation
  returned error each time it was performed with failing callback function.
*/

int fail_callback(op_type rd_or_wr, int *cntr, uint howmuch)
{
  struct st_callback_counts saved_counts= callback_counts;
  int ret= 1, UNINIT_VAR(ret1);

  if (!howmuch)
  {
    diag("No callbacks recorded");
    return 0;
  }

  for(; howmuch; --howmuch)
  {
    reset_counts();
    *cntr= -howmuch;

    switch (rd_or_wr) {
    case WR: ret1= write_image(&img_header, "fail_callback.bst"); break;
    case RD: ret1= read_image(&img_header, "test_image.bst"); break;
    }
    ret= ret && ret1 && !check_iterators();
  }

  callback_counts= saved_counts;
  return ret;
}
