/** @file

  Write to stream table data chunks of various sizes relative to the
  stream block size. Repeat it for different block sizes.
*/

#include <tap.h>

#include "stream_test.h"

/**
  @brief How many stream blocks to write.

  When writing to stream, enough data chunks will be written to cover this
  many stream blocks. However, no more than MAX_STREAM_LEN bytes will be
  written.
*/

#define BLOCK_COUNT     20

/**
  @brief Limit on the size of backup stream created in this test.
*/

#define MAX_STREAM_LEN  (5*1024*1024)

static int chunks(size_t block_size, size_t chunk_size);
static int write_data_chunk(backup_stream *stream, size_t chunk_size);
static int read_data_chunk(backup_stream *stream, size_t *data_size);

/**
  Helper function which runs @c chunks() test via @c ok() call with
  appropriate message.
*/

void ok_chunks(size_t block_size, size_t chunk_size)
{
  ok(!chunks(block_size, chunk_size),
     "Chunks of size %lu over blocks of size %lu",
     (long unsigned)chunk_size, (long unsigned)block_size);
}


void run_test()
{
  size_t block_size[] = { 20, 1024, 4*1024, 16*1024, 1024*1024, 0 };
  size_t bsize;
  uint i;

  plan(5*6);

  for(i=0; block_size[i]; ++i)
  {
    bsize= block_size[i];

    ok_chunks(bsize, bsize/10 + 1);
    ok_chunks(bsize, (5*bsize)/7 + 1);
    ok_chunks(bsize, bsize );
    ok_chunks(bsize, (7*bsize)/5 + 1);
    ok_chunks(bsize, 3*bsize + bsize/3);
    ok_chunks(bsize, 10*bsize);
  }
}


/**
  Counter for table data chunks.

  The counter counts how many table data chunks have been
  written/read by write/read_data_chunk() functions. It should be
  set to 0 before writing/reading starts.
*/

unsigned long int chunk_count=0;
byte *data_buffer= NULL;        /**< buffer for storing chunk contents */


/**
  Write chunks of size @c chunk_size to stream using blocks of
  size @block_size. Then read them back and check data integrity.
*/

int chunks(size_t block_size, size_t chunk_size)
{
  int  ret= 1;
  int  alloc;
  test_backup_stream stream;
  size_t data_len;
  byte   *ptr;


  stream_block_size= block_size;

  /*
    Table data chunk header takes at least 5 bytes. We also need
    at least 1 byte of data. Thus chunk_size can't be less than 6.
   */

  if (chunk_size < 6)
    chunk_size= 6;

  /* allocate buffer for data */

  byte *mem= data_buffer= (byte*)malloc(chunk_size);
  if (!mem)
  {
    diag("Failed to allocate buffer for chunk data");
    alloc= 0;
    goto end;
  }

  alloc= 10;

  /* Fill buffer with data */

  memset(data_buffer, 0xF5, chunk_size);

  /* open stream */

  if (backup_stream_open_wr(&stream, "data_chunks.bst"))
  {
    diag("Failed to open stream for writing");
    goto end;
  }

  alloc= 20;

  /*
    Write enough chunks to output BLOCK_COUNT stream blocks, but no more
    than MAX_STREAM_LEN bytes.
  */
  for(chunk_count=0; chunk_count*chunk_size <= BLOCK_COUNT*block_size
                     && chunk_count*chunk_size < MAX_STREAM_LEN; )
    if (write_data_chunk((backup_stream*)&stream, chunk_size))
      goto end;

  backup_stream_close(&stream);

  alloc= 10;

  /* now start reading */

  if (backup_stream_open_rd(&stream, "data_chunks.bst"))
  {
    diag("Failed to open stream for reading");
    goto end;
  }

  alloc= 20;

  /*
    Check that block size, which is detected when the stream is opened for
    reading, is as we specified when writing the stream.
  */
  if (stream.base.block_size != block_size)
  {
    diag("Incorrect block size %lu instead of %lu.",
         stream.base.block_size, (long unsigned)block_size);
    goto end;
  }

  chunk_count= 0;

  /* read chunks from the stream */

  do 
  {
    ret= read_data_chunk((backup_stream*)&stream, &data_len);

    if (ret == BSTREAM_ERROR)
      goto end;

    /* check that all bytes read are correct */

    for (ptr= data_buffer; ptr < data_buffer + data_len; ++ptr)
      if (*ptr != 0xF5)
      {
        diag("Data of chunk #%lu of length %lu"
             " corrupted at byte %ld which is 0x%X.",
             chunk_count, (long unsigned)data_len, ptr-data_buffer, *ptr);
        ret= 1;
        goto end;
      };

  } while (ret == BSTREAM_OK);

  ret= 0;

 end:

  /* 
    In a moment we are going to free the buffer memory - avoid the dangling 
    pointer. 
  */
  data_buffer= NULL;

  if (alloc >= 20) backup_stream_close(&stream);
  if (alloc >= 10) free(mem);

  return ret;
}


/**
  Write a table data chunk of the given size.

  Data to be written is taken from the global buffer @c data_buffer which
  should be allocated and of appropriate size. Size of the data
  payload is calculated so that the complete chunk (data payload
  + headers) has size @c chunk_size.

  Apart from the data payload, the following information will be
  stored in the table data chunk header, for debugging purposes.

  table number     - size of the data payload,
  snapshot number  - consecutive chunk number modulo 128,
  flags            - 0x0 (not used)

  @returns 0 on success.
*/

int write_data_chunk(backup_stream *stream, size_t chunk_size)
{
  struct st_bstream_data_chunk dchunk;
  size_t data_size= chunk_size;

  memset(&dchunk, 0, sizeof(dchunk));

  ++chunk_count;

  dchunk.table_num=  data_size;
  dchunk.snap_num=   chunk_count & 0x7F;
  dchunk.data.begin= data_buffer;
  dchunk.data.end=   data_buffer+data_size;

  if (BSTREAM_OK != bstream_wr_data_chunk(stream, &dchunk))
  {
    diag("Error when writing data chunk #%lu", chunk_count);
    return 1;
  }

  return 0;
}


int read_data_chunk1(backup_stream *stream, size_t *len);
int read_data_chunk2(backup_stream *stream, size_t *len);

/**
  Read next table data chunk.

  Read either a complete chunk using @c read_data_chunk1() function
  or just an initial fragment woith @c read_data_chunk2() function.

  After reading the chunk, global pointer @c data_buffer is set to point
  at the data payload. The size of the data is stored in @c data_size
  which should not be NULL.

  @retval BSTREAM_OK  Data chunk has been read and there are more in the stream.
  @retval BSTREAM_EOS Data chunk has been read and end of stream has been 
                      reached.
  @retcal BSTREAM_ERROR Error when reading data chunk.
*/

int read_data_chunk(backup_stream *stream, size_t *len)
{
  struct st_bstream_data_chunk dchunk;
  memset(&dchunk, 0, sizeof(dchunk));

  ++chunk_count;

  /* every now and then read data chunk using different logic */
  if (0 == chunk_count % 7)
    return read_data_chunk2(stream, len);  /* read only fragment of the chunk */
  else
    return read_data_chunk1(stream, len);  /* read complete chunk */
}


/**
  Read complete table data chunk.

  Read chunk using @c bstream_rd_data_chunk() function. If the size of the
  chunk stored as table number does not equal the actual size then error is
  reported.

  @retval BSTREAM_OK  Data chunk has been read and there are more in the stream.
  @retval BSTREAM_EOS Data chunk has been read and end of stream has been 
                      reached.
  @retcal BSTREAM_ERROR Error when reading data chunk.
*/

int read_data_chunk1(backup_stream *stream, size_t *len)
{
  struct st_bstream_data_chunk dchunk;
  int ret;
  size_t data_size;

  DBUG_ASSERT(len);

  memset(&dchunk, 0, sizeof(dchunk));

  ret= bstream_rd_data_chunk(stream, &dchunk);

  if (ret == BSTREAM_ERROR)
  {
    diag("Error reading data chunk #%lu.", chunk_count);
    return BSTREAM_ERROR;
  }

  if (ret == BSTREAM_EOC)
  {
    diag("Invalid chunk encountered (#%lu).", chunk_count);
    return BSTREAM_ERROR;
  }

  data_size= dchunk.data.end - dchunk.data.begin;

  if (dchunk.table_num != data_size)
  {
    diag("Invalid data size in chunk #%lu (%u): %lu instead of %lu",
         chunk_count, dchunk.snap_num, dchunk.table_num,
         (long unsigned)data_size);
    return BSTREAM_ERROR;
  }

  data_buffer= dchunk.data.begin;
  *len= data_size;

  return ret;
}


int bstream_read_part(backup_stream *s, blob *data, blob buf);

/**
  Read first 1024 bytes of a table data chunk and skip the rest.
*/

int read_data_chunk2(backup_stream *stream, size_t *len)
{
  short unsigned int       snap_num, flags;
  unsigned int             seq_num;
  long unsigned int        table_num;
  byte    buf[1024];
  blob    to_read, envelope;
  int ret= BSTREAM_OK;

  /* read table data chunk header */
  bstream_rd_byte(stream, &snap_num);
  bstream_rd_int2(stream, &seq_num);
  bstream_rd_byte(stream, &flags);
  bstream_rd_num(stream, &table_num);

  to_read.begin= buf;
  to_read.end=   buf+1024;
  envelope= to_read;
  /*
    Read bytes until current buffer is full or end of chunk is reached.
   */
  while ((ret == BSTREAM_OK) && (to_read.begin < to_read.end))
    ret= bstream_read_part(stream,&to_read,envelope);

  data_buffer= buf;
  *len= to_read.begin - buf;

  /* Move to next chunk */
  return ret==BSTREAM_ERROR ? BSTREAM_ERROR : bstream_next_chunk(stream);
}
