/* 
  @file

  Test bstream_end_chunk() function in situations where the current output
  block is completely or almost completely filled. This lead to assertion 
  failure reported in BUG#44569.

  We output to a stream several chunks of data, each time writing enough data
  to fill complete output block, leaving but a small gap at the end. Initially 
  there is no gap and then it is increased by one byte until it reaches the
  size given by @c MAX_GAP constant.
*/

#include <tap.h>

#include "stream_test.h"

#define MAX_GAP  3

/* import some low-level functions from bstream library */
int  bstream_write_blob(backup_stream*, blob);
int  bstream_flush(backup_stream*);
int  bstream_end_chunk(backup_stream *s);
int  bstream_read_part(backup_stream *s, bstream_blob *data, bstream_blob buf);

int read_chunk(backup_stream *s, size_t *len);


void run_test()
{
  test_backup_stream stream;
  backup_stream *bstr= (backup_stream*)&stream;
  struct st_bstream_buffer *buf= &(bstr->buf);

  /*
   Size of data_buf array determines the size of the output block. It should
   be smaller than 64 and bigger than 10.
  */ 
  byte data_buf[60];
  blob data;
  int i,j;
  size_t remains, howmuch;
  size_t chunk_size[2*(MAX_GAP+1)];    /* for saving chunk sizes */
  int ret= BSTREAM_OK;

  stream_block_size= sizeof(data_buf); /* set output block size */

  /* Fill data buffer and setup data blob */
  memset(data_buf, 0x01, sizeof(data_buf));
  data.begin= data_buf;
  data.end=   data_buf + sizeof(data_buf);

  plan(2 + 6*(MAX_GAP+1));

  /* Open stream for writing */

  ret= backup_stream_open_wr(&stream, "bstr_eoc.bst");
  if (ret)
  {
    diag("Failed to open stream for writing");
    return;
  }

  ok(TRUE, "Stream opened for writing");

  /*
    Variable i loops over gap sizes varying from 0 to MAX_GAP. For each gap 
    size we write two data chunks (j=0,1) to exercise different paths in 
    bstream_end_chunk().
  */
  for (i=0; i <= MAX_GAP; ++i)
  for (j=0; j <= 1; ++j)
  {
    /* Write first 10 bytes of a chunk and flush the stream */
    data.end= data_buf + 10;
    ret= bstream_write_blob(bstr, data);
    bstream_flush(bstr);

    if (ret != BSTREAM_OK)
    {
      diag("Error %d when writing first part of a chunk", ret);
      return;
    }

    /*
      Calculate how much bytes to output to the stream to leave a gap of desired
      size at the end of the current output block.

      Calculations use the fact that buf->pos points one byte after the last 
      byte output to the stream and buf->pos points one byte after the end of
      the current output block.
    */ 
    remains= buf->end - buf->pos;
    howmuch= remains - 1 - i;   /* add one byte for fragment header */

    data.end= data_buf + howmuch;
    ret= bstream_write_blob(bstr, data);
    if (ret != BSTREAM_OK)
    {
      diag("Error %d when writing second part of a chunk", ret);
      return;
    }

    /* check that there is a gap of expected size */
    ok((buf->end - buf->pos) == i, 
       "Closing chunk which ends %d bytes from the end of the output block", i); 

    /*
      We close the chunk in two situations. First (j=0), when the current 
      fragment is still open so that the EOC flag can be set in its header 
      which is pointed by buf->header.

      If j=1 then we set buf->header to buf->pos which effectively closes the
      current fragment. In that case its header can not be modified and end
      of the chunk must be marked by appending EOC marker to the stream.
    */ 
    if (j)
      buf->header= buf->pos;

    ret= bstream_end_chunk(bstr);
    if (ret != BSTREAM_OK)
    {
      diag("Error %d when closing chunk", ret);
      return;
    }

    ok(TRUE, "Closed data chunk of length %lu", (unsigned long)howmuch+10);

    /* 
      Save the size of the chunk so that we can verify it when the stream is
      read below.
    */
    chunk_size[2*i+j]= howmuch+10;
  }

  backup_stream_close(&stream);

  /* Now read chunks. */

  ret= backup_stream_open_rd(&stream, "bstr_eoc.bst");
  if (ret)
  {
    diag("Failed to open stream for writing");
    return;
  }

  ok(TRUE, "Stream opened for reading");

  i=0;  /* For counting chunks which are read from the stream */

  do 
  {  
    ret= read_chunk(bstr, &howmuch);

    if (ret == BSTREAM_ERROR)
    {
      diag("Error reading data chunk");
      return;
    }
    
    ok(howmuch == chunk_size[i++], "Read next chunk of size %lu", 
                                   (unsigned long)howmuch);
  }
  while (ret == BSTREAM_OK);

  backup_stream_close(&stream);
}


int read_chunk(backup_stream *s, size_t *len)
{
  byte buf[1024];
  blob space, to_read;
  size_t  howmuch;
  int  ret;

  DBUG_ASSERT(len);

  *len= 0;
  space.begin= buf;
  space.end= buf + sizeof(buf);

  do {
    to_read= space;
    ret= bstream_read_part(s, &to_read, space);
    howmuch= to_read.begin - buf;
    *len+= howmuch;
  } while (ret == BSTREAM_OK);

  if (ret == BSTREAM_EOC)
  {
    return bstream_next_chunk(s);
  }

  return ret;
}
