/** @file

  Prematurely report end of stream when reading backup image. EOS is
  reported after reading given amount of stream chunks.

  Test uses in-memory stream. When backup image is written, it is stored
  in a memory buffer. Then the image is analysed to find chunk boundaries.
  Finally, image is read from memory several times. Each time read is
  interrupted by EOS signal at different stages.
*/

#include <tap.h>

#include "stream_test.h"
#include "catalog.h"

#define FR_EOC    0x80  /**< bits for EOC fragment */
#define FR_EOS    0xC0  /**< bits for EOS fragment */
#define FR_MORE   0x00  /**< bits for MORE fragment */
#define FR_LAST   0x40  /**< bits for LAST fragment */

static int mem_write_image(image_header *hdr);
static int mem_read_image(image_header *hdr);
static int mem_read_lowlevel();
static void get_chunk_boundaries();

#define BSIZE   1024    /**< Size of stram i/o block */

/**
  @brief Buffer for in-memory stream.

  We will store only the first block of the image, hence only that much
  space is needed.
*/

byte image[BSIZE];
byte *stream_pos= image;  /**< current position in the buffer */
byte *stop_at= NULL;      /**< position at which reading will be interrupted */
int  early_eos= BSTREAM_EOS; /**< value to return when reding is interrupted */

/**
  @brief Array of pointers to chunk boundaries.

  This is filled by @c get_chunk_boundaries().
*/

byte *chunk[128];

/**
  @brief The number of chunks found in the image.

  This is set by @c get_chunk_boundaries().
*/

uint chunk_count= 0;


void run_test()
{
  image_header  hdr;
  int  i;
  byte header;
 /*
   Number of tests to be planned are not known at this point.
 */
  plan(NO_PLAN);

  init_catalog(&hdr);

  ok(!mem_write_image(&hdr), "Writing backup image");

  get_chunk_boundaries();

  /*
    In what follows we will read the image interrupting low-level read at each
    chunk boundary. First we instruct low-level read to report ERROR at
    interruption. Then, it will report EOS.
  */

  early_eos= BSTREAM_ERROR;

  for (i= chunk_count-1; i >= 0; --i)
  {
    stop_at= chunk[i];
    ok(BSTREAM_ERROR == mem_read_image(&hdr),
       "ERROR at chunk %d when reading image", i);
  }

  /*
    The low-level read subroutine does not interpret the image. It merely
    scans all data chunks.
  */
  for (i= chunk_count-1; i >= 0; --i)
  {
    stop_at= chunk[i];
    ok(BSTREAM_ERROR == mem_read_lowlevel(&hdr),
       "ERROR at chunk %d when reading chunks", i);
  }

  early_eos= BSTREAM_EOS;

  /*
    There are "chunks" at the end of the stream which function as end
    markers. The reading logic works fine if these are missing
    due to early EOS. Hence we don't expect bstream library to report
    errors for the few last chunks and we skip them.
   */

  for (i= chunk_count-4; i >= 0; --i)
  {
    stop_at= chunk[i];
    ok(BSTREAM_ERROR == mem_read_image(&hdr),
       "EOS at chunk %d when reading image", i);
  }

  /*
    In the next loop we store EOS marker at the end of each chunk, so that the
    end of stream is detected by bstream logic before it is reported by
    low-level I/O.
  */

  for (i= chunk_count-4; i >= 0; --i)
  {
    header= *chunk[i];
    *chunk[i]= FR_EOS;
    stop_at= chunk[i]+1;
    ok(BSTREAM_ERROR == mem_read_image(&hdr),
       "EOS marker at chunk %d when reading image", i);
    *chunk[i]= header;
  }

  for (i= chunk_count-4; i >= 0; --i)
  {
    header= *chunk[i];
    *chunk[i]= FR_EOS;
    stop_at= chunk[i]+1;
    ok(BSTREAM_EOS == mem_read_lowlevel(&hdr),
       "EOS marker at chunk %d when reading chunks", i);
    *chunk[i]= header;
  }

}


/*********************************************************************

  Helper functions for reading/writing backup image.

*********************************************************************/

static int mem_write(void*, blob*, blob);
static int mem_read(void*, blob*, blob);
static int mem_forward(void*, unsigned long int*);

static int mem_write_image(image_header *hdr)
{
  backup_stream stream;
  int ret;

  stream.stream.write= mem_write;
  stream.stream.read= mem_read;
  stream.stream.forward= mem_forward;
  stream_pos= image;

  reset_counts();

  if (bstream_open_wr(&stream, 1024, 0))
    return OPEN_ERROR;

  ret= bstream_wr_preamble(&stream, hdr);
  bstream_close(&stream);

  return check_iterators() ? BSTREAM_ERROR : ret;
}


static int mem_read_image(image_header *hdr)
{
  backup_stream stream;
  int ret;

  stream.stream.write= mem_write;
  stream.stream.read= mem_read;
  stream.stream.forward= mem_forward;
  stream_pos= image;

  reset_counts();

  if (bstream_open_rd(&stream, 0))
    return OPEN_ERROR;

  ret= bstream_rd_preamble(&stream, hdr);
  bstream_close(&stream);

  return check_iterators() ? BSTREAM_ERROR : ret;
}


static int mem_read_lowlevel()
{
  backup_stream stream;
  unsigned short int x;
  int ret;

  stream.stream.write= mem_write;
  stream.stream.read= mem_read;
  stream.stream.forward= mem_forward;
  stream_pos= image;

  if (bstream_open_rd(&stream, 0))
    return OPEN_ERROR;

  do {
    ret= bstream_rd_byte(&stream, &x);
    ret= bstream_next_chunk(&stream);
  } while (ret == BSTREAM_OK);

  bstream_close(&stream);

  return ret;
}


/*********************************************************************

  Implementation of in-memory stream.

*********************************************************************/

static
int mem_write(void *stream __attribute__ ((__unused__)),
              bstream_blob *buf,
              bstream_blob envelope __attribute__ ((__unused__)))
{
  int ret= BSTREAM_OK;

  DBUG_ASSERT(buf);

  if (!buf->begin || buf->begin == buf->end)
    return BSTREAM_OK;

  DBUG_ASSERT(buf->end);

  size_t howmuch = buf->end - buf->begin;
  if (stream_pos + howmuch >= image + 1024)
  {
    howmuch= image + 1024 - stream_pos;
    ret= BSTREAM_EOS;
  }

  memmove(stream_pos, buf->begin, howmuch);

  stream_pos+= howmuch;
  buf->begin+= howmuch;

  return ret;
}


static
int mem_read(void *stream __attribute__ ((__unused__)),
             bstream_blob *buf,
             bstream_blob envelope __attribute__ ((__unused__)))
{
  size_t howmuch;
  int ret= BSTREAM_OK;

  DBUG_ASSERT(buf);

  if (!buf->begin || buf->begin == buf->end)
    return BSTREAM_OK;

  DBUG_ASSERT(buf->end);

  if (stream_pos == stop_at)
    return early_eos;

  howmuch= buf->end - buf->begin;
  if (stream_pos + howmuch >= image + 1024)
  {
    howmuch= image + 1024 - stream_pos;
    ret= BSTREAM_EOS;
  }
  if (stop_at && stream_pos + howmuch > stop_at)
  {
    howmuch= stop_at - stream_pos;
    ret= BSTREAM_EOS;
  }

  memmove(buf->begin, stream_pos, howmuch);

  stream_pos += howmuch;
  buf->begin += howmuch;

  return ret;
}


static
int mem_forward(void *stream __attribute__ ((__unused__)),
                unsigned long int *offset)
{
  int ret= BSTREAM_OK;
  size_t howmuch;

  DBUG_ASSERT(offset);

  if (*offset == 0)
    return BSTREAM_OK;

  howmuch= *offset;
  if (stream_pos + howmuch >= image + 1024)
  {
    howmuch= image + 1024 - stream_pos;
    ret= BSTREAM_EOS;
  }

  stream_pos+= howmuch;
  *offset-= howmuch;

  return ret;
}


/*********************************************************************

  Code for analysing chunk boundaries (taken from stream_v1_transport.c).

*********************************************************************/

#define FR_TYPE_MASK  0xC0  /**< type bits for mask */
#define FR_LEN_MASK   0x3F  /**< type length for mask */

/** biggest size of small fragment */
#define FR_SMALL_MAX  ((size_t)FR_LEN_MASK)
#define FR_BIG        0x80        /**< type bits for big fragment */
#define FR_HUGE       0xC0        /**< type bits for huge fragment */
#define FR_BIG_SHIFT  6           /**< value shift for big fragment */
#define FR_HUGE_SHIFT 12          /**< value shift for huge fragment */
/** header for the biggest possible chunk */
#define FR_HUGE_MAX_HDR (FR_HUGE|FR_LEN_MASK)
/** size of the biggest possible chunk */
#define FR_HUGE_MAX_LEN ((size_t)FR_LEN_MASK << FR_HUGE_SHIFT)


static
int read_fragment_header(byte **header)
{
  byte hdr= *((*header)++);
  size_t len= hdr & FR_LEN_MASK; /* 6 bit value stored in the header */
  int last;

  if (hdr == FR_EOS)
    return BSTREAM_EOS;

  if (hdr == FR_EOC)
    return BSTREAM_EOC;

  last= hdr & FR_LAST ? FR_LAST : FR_MORE; /* is it last fragment of a chunk? */

  /*
    If len == 0 then we have a fragment which extends to the end of a block,
    thus we restore *header to signal that fact to the caller.
   */
  if (len == 0)
  {
    (*header)--;
    return last;
  }

  /*
    If the highest bit is set, we are looking at a big or huge fragment. Such
    fragment is never the last fragment of a chunk.
   */
  if (hdr & 0x80)
  {
    last= FR_MORE;

    if ((hdr & FR_TYPE_MASK) == FR_BIG)
      len <<= FR_BIG_SHIFT;

    if ((hdr & FR_TYPE_MASK) == FR_HUGE)
      len <<= FR_HUGE_SHIFT;
  }

  /* move *header pointer to the first byte after the fragment */
  *header += len;

  return last;
}


void get_chunk_boundaries()
{
  byte *fhead= image+5;
  byte *saved_fhead;

  chunk_count=0;

  chunk[chunk_count++]= fhead;

  while (fhead < image + 1024 && chunk_count < 127)
  {
    saved_fhead= fhead;
    switch (read_fragment_header(&fhead)) {

    case BSTREAM_EOC:
    case FR_LAST:
      if (fhead > saved_fhead)
        chunk[chunk_count++]= fhead;
      else
        fhead = image + 1024;
      break;

    case BSTREAM_EOS:
      chunk[chunk_count++]= saved_fhead;
      fhead= image + 1024;

    case FR_MORE: break;

    }
  }

  chunk[chunk_count]= NULL;
}
