/** @file

  Execute some "border cases" to exercise bstream library code. This includes
  incomplete backup image preamble, incorrect snapshot/item types and
  writing/reading image with no objects.
*/
#include <tap.h>
#include "stream_test.h"
#include "catalog.h"

/* Import some low-level functions from bstream library */
int bstream_wr_header(backup_stream*, struct st_bstream_image_header*);
int bstream_wr_catalogue(backup_stream*, struct st_bstream_image_header*);
int bstream_end_chunk(backup_stream*);

image_header  img_header;


void run_test()
{
  test_backup_stream stream;
  int ret1, ret2;

  plan(5);

  init_catalog(&img_header);

  /* write/read only image header */

  reset_counts();
  if (backup_stream_open_wr(&stream, "extra.bst"))
  {
    diag("Failed to open stream for writing");
    return;
  };
  ret1= bstream_wr_header((backup_stream*)&stream, &img_header);
  backup_stream_close(&stream);
  ret1= ret1 || check_iterators();

  reset_counts();
  if (backup_stream_open_rd(&stream, "extra.bst"))
  {
    diag("Failed to open stream for reading");
    return;
  };
  ret2= bstream_rd_preamble((backup_stream*)&stream, &img_header);
  backup_stream_close(&stream);
  ret1= ret1 || check_iterators();

  ok(!ret1 && ret2, "Incomplete preamble: only header");

  /* write/read preamble with metadata section missing */

  reset_counts();
  if (backup_stream_open_wr(&stream, "extra.bst"))
  {
    diag("Failed to open stream for writing");
    return;
  };
  ret1= bstream_wr_header((backup_stream*)&stream, &img_header);
  ret1= ret1 || bstream_end_chunk((backup_stream*)&stream);
  ret1= ret1 || bstream_wr_catalogue((backup_stream*)&stream, &img_header);
  backup_stream_close(&stream);
  ret1= ret1 || check_iterators();

  reset_counts();
  if (backup_stream_open_rd(&stream, "extra.bst"))
  {
    diag("Failed to open stream for reading");
    return;
  };
  ret2= bstream_rd_preamble((backup_stream*)&stream, &img_header);
  backup_stream_close(&stream);
  ret1= ret1 || check_iterators();

  ok(!ret1 && ret2, "Incomplete preamble: no metadata");


  /* write/read snapshot of unknown type */

  snapshots[0]= 333;      /* unknown snapshot type */
  init_catalog(&img_header);

  reset_counts();
  ret1= write_image(&img_header, "unknown_snapshot.bst") || check_iterators();
  reset_counts();
  ret2= read_image(&img_header, "unknown_snapshot.bst");
  ret1= ret1 || check_iterators();

  ok(!ret1 && ret2, "Image with unknown snapshot type");

  /* try writing image with unknown item type (it should fail) */

  objects_perdb[0].type= 333;
  init_catalog(&img_header);

  reset_counts();
  ret1= write_image(&img_header, "unknown_item.bst");
  ret2= check_iterators();

  ok(ret1 && !ret2, "Writing image with unknown item type");

  /* write/read empty backup image (also no databases) */
  opt_empty_catalog= 1;
  img_header.snap_count= 0;

  reset_counts();
  ret1= write_image(&img_header, "empty.bst") || check_iterators();
  reset_counts();
  ret2= read_image(&img_header, "empty.bst") || check_iterators();

  ok(!ret1 && !ret2, "Empty image");

}
