/* @file

  Provoke errors when reading basic data types, so that error handling branch
  in the code is executed. This is done only for datatypes for which error
  branches are not covered in other tests.
*/

#include <tap.h>

#include "stream_test.h"

/* import some low-level functions from bstream library */

int  bstream_rd_item_type(backup_stream*, item_type*);
int  bstream_write_blob(backup_stream*, blob);
int  bstream_end_chunk(backup_stream*);
void bstream_free(byte*);

/**
  Dummy low-level read function which always returns error.
*/

int dummy_read(void *stream, blob *buf, blob envelope);

void run_test()
{
  test_backup_stream stream;
  backup_stream *bstr= (backup_stream*)&stream;

  /*
    Note: buffer size must be odd as only then expected error branch is taken.
  */
  byte buf[127];

  blob data;
  unsigned int num;
  long unsigned int lnum;
  int ret= BSTREAM_OK;

  /* fill buffer with 0x01 bytes */
  memset(buf, 0x01, sizeof(buf));
  data.begin= buf;
  data.end=   buf + sizeof(buf);

  stream_block_size= 256;

  plan(3 + 3 + 3);

  /* write 0x01 bytes to stream */

  ret= backup_stream_open_wr(&stream, "extra1.bst");
  if (ret)
  {
    diag("Failed to open stream for writing");
    return;
  }

  ret= bstream_write_blob(bstr, data);
  ok(BSTREAM_OK == ret, "Writing bytes to stream");

  /*
    execute flush twice to enter branch where flush is made with empty output
    buffer
  */
  ret= bstream_flush(bstr);
  ret= ret || bstream_flush(bstr);
  ok(!ret, "Flushing flushed stream");

  backup_stream_close(&stream);

  /*
     call close on already closed stream to cover corresponding branch in the
     code
  */
  ret= bstream_close(bstr);
  ok(BSTREAM_OK == ret, "Closing already closed stream");

  ret= backup_stream_open_rd(&stream, "extra1.bst");
  if (ret)
  {
    diag("Failed to open stream for reading");
    return;
  }

  /*
    Call writing functions on stream opened for reading.
   */
  ret= bstream_write_blob(bstr, data);
  ok(BSTREAM_ERROR == ret, "Writing bytes to stream opened for reading");
  ret= bstream_end_chunk(bstr);
  ok(BSTREAM_ERROR == ret, "Closing chunk in stream opened for reading");
  ret= bstream_flush(bstr);
  ok(BSTREAM_ERROR == ret, "Flushing stream opened for reading");

  /*
    After setting read function to dummy_read, error will be hit by bstream
    library next time it tries to reload input buffer.
  */
  stream.base.stream.read= dummy_read;

  /*
   In what follows, various basic types of data are read from the stream. This
   is done in a loop, so that eventually the bstream input buffer will be
   reloaded and error will be hit.
  */

  do
  {
    ret= bstream_rd_string(bstr, &data);
    if (!ret)
      bstream_free(data.begin);
  } while (ret == BSTREAM_OK);

  backup_stream_close(&stream);
  ok(BSTREAM_ERROR == ret, "Error when reading strings");

  ret= backup_stream_open_rd(&stream, "extra1.bst");
  if (ret)
  {
    diag("Failed to open stream for reading");
    return;
  }
  stream.base.stream.read= dummy_read;

  do
  {
    ret= bstream_rd_int4(bstr, &lnum);
  } while (ret == BSTREAM_OK);

  backup_stream_close(&stream);
  ok(BSTREAM_ERROR == ret, "Error when reading int4 numbers");

  ret= backup_stream_open_rd(&stream, "extra1.bst");
  if (ret)
  {
    diag("Failed to open stream for reading");
    return;
  }
  stream.base.stream.read= dummy_read;

  do
  {
    ret= bstream_rd_int2(bstr, &num);
  } while (ret == BSTREAM_OK);

  bstream_next_chunk(bstr);
  bstream_next_chunk(bstr);

  backup_stream_close(&stream);
  ok(BSTREAM_ERROR == ret, "Error when reading int2 numbers");

}


int dummy_read(void *stream  __attribute__ ((__unused__)),
               blob *buf     __attribute__ ((__unused__)),
               blob envelope __attribute__ ((__unused__)))
{
  return BSTREAM_ERROR;
}
