/* @file

  In order to excersise bstream library code, read and write a pseudo-random
  sequence of chunks of varied sizes.
*/
#include <tap.h>
#include "stream_test.h"

/* import some low-level function from bstream library */

int bstream_flush(backup_stream *s);
int bstream_end_chunk(backup_stream *s);
int bstream_write_blob(backup_stream *s, bstream_blob buf);
int bstream_wr_int4(backup_stream *s, unsigned long int x);
int bstream_read_part(backup_stream *s, bstream_blob *data, bstream_blob buf);

/** Determines how big image will be created */
#define DATA_AMOUNT 10*1024*1024
/** Stream block size to use */
#define BSIZE       4*1024

/**
  List of chunk sizes which will be used in the pseudo-random sequence.
*/
size_t chunk_size[] = { 10,
                        BSIZE/10,
                        (5*BSIZE)/7,
                        (7*BSIZE)/5,
                        500*1024 };
#define CSIZE_NUM   5   /**< Size of @c chunk_size[] array */


static uint hash(uint mod, size_t key);
static int write_chunk(backup_stream *stream, size_t chunk_size);
static int read_chunk(backup_stream *stream, size_t *len);
/**
  Chunk counter increased by @c write/read_chunk() functions.
*/
static unsigned long int chunk_count=0;

void run_test()
{
  test_backup_stream stream;
  size_t byte_count;
  size_t csize;
  int    ret;

  stream_block_size= BSIZE;
 /*
   Number of tests to be planned are not known at this point.
  */
  plan(NO_PLAN);

  diag("Writing");

  ret= backup_stream_open_wr(&stream, "random.bst");
  if (ret)
  {
    diag("Failed to open stream for writing");
    return;
  }

  chunk_count= 0;

  for(byte_count=0 ; byte_count < DATA_AMOUNT ; )
  {
    csize= chunk_size[hash(CSIZE_NUM, chunk_count)];

    ok(!write_chunk((backup_stream*)&stream, csize),
       "Wrote chunk #%lu of size %lu bytes", chunk_count, (long unsigned)csize);

    /*
      Flush stream every now and then to exercise more code in bstream library.
     */
    if (chunk_count % 7 == 0)
      bstream_flush((backup_stream*)&stream);

    byte_count+= csize;
  }

  backup_stream_close(&stream);

  diag("Reading");


  ret= backup_stream_open_rd(&stream, "random.bst"); 
  if (ret)
  {
    diag("Failed to open stream for reading");
    return;
  }

  chunk_count= 0;

  do {

    ret= read_chunk((backup_stream*)&stream, &byte_count);
    ok(BSTREAM_ERROR != ret,
       "Reading chunk #%lu of size %lu bytes",
       chunk_count, (long unsigned)byte_count);

    if (ret == BSTREAM_ERROR)
      goto end;

  } while (ret == BSTREAM_OK);

  /* try to read yet another chunk to hit EOS */

  ret= read_chunk((backup_stream*)&stream, &byte_count);
  ok(BSTREAM_ERROR == ret,"Reading another chunk after EOS was hit");

  ret= bstream_next_chunk((backup_stream*)&stream);
  ok(BSTREAM_EOS == ret,"Moving to next chunk after EOS was hit");

 end:

  backup_stream_close(&stream);
}

/**
  Hash function returning number in given range.

  @param[IN] mod   function will return number in range 0 .. (mod-1)
  @param[IN] key   function argument

  @note Function uses sdbm algorithm from http://www.cse.yorku.ca/~oz/hash.html
*/
uint hash(uint mod, size_t key)
{
  static size_t hash= 33;
  hash= key + (hash << 6) + (hash << 16) - hash;
  return hash % mod;
}

/**
  Write a chunk of given size to backup stream.

  First 4 bytes of the chunk contain its size. The remaining bytes are filled
  with 0xF5 byte.

  @returns Non-zero value in case of error, 0 otherwise.
*/
int write_chunk(backup_stream *stream, size_t chunk_size)
{
  byte buf[1024];
  blob to_write;
  int ret;

  memset(buf, 0xF5, 1024);
  ++chunk_count;

  bstream_wr_int4(stream, chunk_size);
  chunk_size-= 4;

  while (chunk_size)
  {
    to_write.begin= buf;
    to_write.end= buf + chunk_size;

    if (to_write.end > to_write.begin + 1024)
      to_write.end= to_write.begin + 1024;

    chunk_size-= (to_write.end - to_write.begin);

    ret= bstream_write_blob(stream, to_write);

    if (BSTREAM_OK != ret)
    {
      diag("Error when writing chunk #%lu", chunk_count);
      return ret;
    }
  }

  return bstream_end_chunk(stream);
}

/**
  Read chunk written by @c write_chunk.

  Reads the chunk, verifies data integrity and its length.

  @param[OUT] len  set to the number of bytes found in the chunk.

  @retval BSTREAM_OK    Chunk read successfuly and integrity tests passed.
  @retval BSTREAM_EOS   End of stream has been encountered.
  @retval BSTREAM_ERROR Error detected.
*/
int read_chunk(backup_stream *stream, size_t *len)
{
  unsigned long int chunk_size;
  byte buf[1024];
  byte *ptr;
  bstream_blob to_read, envelope;
  int ret= BSTREAM_OK;

  DBUG_ASSERT(len);

  ++chunk_count;

  ret= bstream_rd_int4(stream, &chunk_size);
  if (ret != BSTREAM_OK)
    return ret;

  *len= chunk_size;
  chunk_size-= 4;

  while (chunk_size && ret == BSTREAM_OK)
  {
    to_read.begin= buf;
    to_read.end=   buf + chunk_size;

    if (to_read.end > to_read.begin + 1024)
      to_read.end= to_read.begin + 1024;

    envelope= to_read;
    ret= bstream_read_part(stream, &to_read, envelope);

    if (ret == BSTREAM_ERROR)
    {
      diag("Error reading data in chunk #%lu.", chunk_count);
      return BSTREAM_ERROR;
    }

    /* check that all bytes are correct */

    for (ptr= buf; ptr < to_read.begin; ++ptr)
      if (0xF5 != *ptr)
      {
        diag("Incorrect byte read in chunk #%lu", chunk_count);
        return BSTREAM_ERROR;
      }

    chunk_size-= to_read.begin - buf;
  }

  (*len)-= chunk_size;
  ret= bstream_next_chunk(stream);

  if (chunk_size)
  {
    diag("Not enough bytes found in chunk #%lu: %lu are missing",
         chunk_count, chunk_size);
    return BSTREAM_ERROR;
  }

  return ret;
}
