/* 
  @file

  Provoke a situation which led to the assertion failure reported in BUG#44569.
  This is a situation when the output buffer is empty (buf.begin == buf.header
  == buf.pos) and at the same time is at the very end of the current output
  block. Calling bstream_write_part() in such situation lead to assertion 
  failure.
*/

#include <tap.h>

#include "stream_test.h"

/* import some low-level functions from bstream library */

int  bstream_write_blob(backup_stream*, blob);
int  bstream_write_part(backup_stream*, blob*, blob);
int  bstream_flush(backup_stream*);


void run_test()
{
  test_backup_stream stream;
  backup_stream *bstr= (backup_stream*)&stream;
  struct st_bstream_buffer *buf= &(bstr->buf);

  byte data_buf[1089]; /* The size here is crucial - see below */
  blob data;
  int i;
  size_t remains;
  int ret= BSTREAM_OK;

  stream_block_size= sizeof(data_buf); /* set output block size */

  /* Fill data buffer and setup data blob */
  memset(data_buf, 0x01, sizeof(data_buf));
  data.begin= data_buf;
  data.end=   data_buf + sizeof(data_buf);

  plan(4);

  /* Open stream for writing */

  ret= backup_stream_open_wr(&stream, "bug45737.bst");
  if (ret)
  {
    diag("Failed to open stream for writing");
    return;
  }

  /*
    In the following loop we try to arrive at a situation where output stream
    is at a beginning of next output block. If we are not there yet, we write 
    enough data to fill a complete block so that we move to next one (we must 
    leave 2 bytes for fragment headers - with block size > 64 the data will be 
    split into at least 2 fragments).
    
    This must be repeated because first couple of blocks contains additional
    data (block size and init block count).  
  */

  for(i=0; i<10; ++i)
  {
    remains= buf->end - buf->pos;
    data.end = data_buf + remains - 2; // leave 2 bytes for fragment headers
    ret= bstream_write_blob(bstr, data);
    ret= (ret == BSTREAM_OK) ? bstream_flush(bstr) : ret;

    if (ret != BSTREAM_OK)
    {
      diag("Error %d after %d-th write when positioning stream", ret, i+1); 
      return;
    }

    if (buf->end == buf->pos + bstr->block_size)
      break;
  }

  ok(buf->end == buf->pos + bstr->block_size,
     "Stream at the beginning of the next output block");

  /*
    It is important that we write 1088 bytes. This is an amount which is
    bigger than MIN_WRITE_SIZE (1024) so that bstream_write() logic will 
    flush output buffer and write data directly to the stream. It is also 
    important that least 6 bits are 0, so that the whole chunk of data will
    be written as a single fragment. Stream block size must equal the size of
    the data chunk + 1 byte for fragment header.
    
    Only under these conditions we arrive at the situation where output
    buffer is empty and at the end of the current output block.
  */ 

  data.end= data_buf + 1088;
  ret= bstream_write_blob(bstr, data);

  if (ret != BSTREAM_OK)
  {
    diag("Writting error %d when preparing final stream position", ret);
    return;
  }

  /*
    Now we should have the desired situation where
    bstr->buf.begin == bstr->buf.header == bstr->buf.pos == bstr->buf.end
  */ 

  ok(buf->begin  == buf->header && buf->header == buf->pos,
     "Output buffer is empty");
  ok(buf->pos == buf->end, "At the end of the output block");

  /*
    Call bstream_write_part() - it should correctly handle this situation and
    not trigger assertion failure.
  */
  data.end= data_buf + 60;
  ret= bstream_write_part(bstr, &data, data);

  ok(ret == BSTREAM_OK, "Final write");

  backup_stream_close(&stream);
}
