/** @file

  Define backup/restore catalogue callbacks to be used for testing backup
  stream library.

  The contents of the catalogue is determined by arrays @c snapshots[],
  @c objects_global[], @c objects_perdb[]. Additionally, the catalogue will 
  contain DB_COUNT databases named db1, ..., dbN and TABLE_COUNT tables named 
  t1, ..., tM. Tables are distributed among available databases and snapshots.

  For each object in the catalogue, the serialization string and extra metadata
  string have form "serialization of <object name>" and
  "metadata for <object name>", respectively.

  Calls to all callback functions are counted. There are mechanisms for
  reporting error from each callback function after a given number of calls.
*/

#include <my_global.h>
#include <m_string.h>
#include <stream_v1_services.h>

#include "catalog.h"

/**
  List of table data snapshots reported in the catalogue.
*/
snapshot_type snapshots[] = {BI_NATIVE, BI_DEFAULT, BI_CS, BI_NODATA};
const uint snap_count= sizeof(snapshots)/sizeof(snapshot_type);

/**
  List of global objects other than databases.

  Databases are declared separately with DB_COUNT constant.
*/
item_info objects_global[]=
{
  {BSTREAM_IT_CHARSET,    {(byte*)"charset", NULL},   0},
  {BSTREAM_IT_USER,       {(byte*)"user", NULL},      0},
  {BSTREAM_IT_TABLESPACE, {(byte*)"tablespace",NULL}, 0},
};
const uint global_count= sizeof(objects_global)/sizeof(item_info);

/*
  List of per-database objects other than tables.

  Tables are declared separately with TABLE_COUNT constant. These objects
  are reported inside each database.
*/
item_info objects_perdb[]=
{
  {BSTREAM_IT_VIEW,       {(byte*)"view",NULL},       0},
  {BSTREAM_IT_SPROC,      {(byte*)"procedure",NULL},  1},
  {BSTREAM_IT_SFUNC,      {(byte*)"function",NULL},   2},
  {BSTREAM_IT_EVENT,      {(byte*)"event",NULL},      3},
  {BSTREAM_IT_PRIVILEGE,  {(byte*)"privilege",NULL},  4}
};
const uint perdb_count= sizeof(objects_perdb)/sizeof(item_info);

/**
  Defines number N of databases present in the catalogue.

  The databases are called db1, ... , dbN.
*/
#define DB_COUNT        1

/**
  Defines number M of tables present in the catalogue.

  The tables are called t1, ... , tM. They are evenly distributed among the
  databases. That is, table t1 belongs to db1, ... tN belongs to dbN, tN+1
  belongs to db1 and so on.
*/
#define TABLE_COUNT     4

table_info tables[TABLE_COUNT];
db_info    dbs[DB_COUNT];

const uint db_count= DB_COUNT;
const uint table_count= TABLE_COUNT;

/**
  Structure containing bstream callback counters.

  Whenever one of the low-level I/O functions is called, the corresponding
  counter will be increased.

  Setting counter to a negative value will cause the corresponding function
  to report ERROR or EOS when counter reaches the value of 0.
*/

struct st_callback_counts callback_counts;


/**
  Initialize data structures implementing the catalogue.

  This function must be called before any callback is used.
*/
void init_catalog(image_header *hdr)
{
  uint i;
  char *name;
  blob *engine_name;
  static char db_name[DB_COUNT][10];
  static char table_name[TABLE_COUNT][10];

  bzero(hdr, sizeof(image_header));

  hdr->version= 1;
  hdr->snap_count= snap_count;
  hdr->flags= BSTREAM_FLAG_BINLOG;

  /* Fill snapshot information */

  for(i=0; i < snap_count; ++i)
  {
    hdr->snapshot[i].type= snapshots[i];
    if (snapshots[i] == BI_NATIVE)
    {
      engine_name=  &(hdr->snapshot[i].engine.name);
      engine_name->begin= (byte*)"Storage_engine";
      engine_name->end=   engine_name->begin + 14;
    }
  }

  /* Initialize database items */

  for (i=0; i < db_count; ++i)
  {
    name= db_name[i];

    snprintf(name, 10, "db%d", i+1);

    dbs[i].base.type= BSTREAM_IT_DB;
    dbs[i].base.pos= i;
    dbs[i].base.name.begin= (byte*)name;
    dbs[i].base.name.end=   (byte*)name + strlen(name);
  }

  /* Initialize other global items */

  for (i=0; i < global_count; ++i)
  {
    name= (char*)objects_global[i].name.begin;
    objects_global[i].name.end= (byte*)name + strlen(name);
  }

  /* Initialize table items */

  for (i=0; i < table_count; ++i)
  {
    name= table_name[i];

    snprintf(name, 10, "t%d", i+1);

    tables[i].base.base.type= BSTREAM_IT_TABLE;
    tables[i].base.base.pos= i;
    tables[i].base.base.name.begin= (byte*)name;
    tables[i].base.base.name.end= (byte*)name + strlen(name);
    tables[i].base.db=  &dbs[i % db_count];
    tables[i].snap_num= i % snap_count;
    ++(hdr->snapshot[i % snap_count].table_count);
  }

  /* Initialize other per-db items */

  for (i=0; i < perdb_count; ++i)
  {
    name= (char*)objects_perdb[i].name.begin;
    objects_perdb[i].name.end= (byte*)name + strlen(name);
  }

}

/** Reset all callback counters. */
void reset_counts()
{
   bzero(&callback_counts, sizeof(callback_counts));
}

/**
  Check that each iterator_get() call was matched with iterator_free().

  @return 0 if call counts match, non-zero otherwise.
*/
int check_iterators()
{
  uint i;

  for(i=0; i<128; ++i)
    if (callback_counts.iterator_get[i] > 0
        && callback_counts.iterator_get[i] 
           != callback_counts.iterator_free[i])
      return 1;

  for(i=0; i<128; ++i)
    if (callback_counts.db_iterator_get[i] > 0
        && callback_counts.db_iterator_get[i] 
           != callback_counts.db_iterator_free[i])
      return 1;

  return 0;
}

/**********************************************************************

  Functions for iterating over contents of the catalogue.


  Backup stream library asks for different kinds of iterators:

  a) iterator over a given type of global objects (BSTREAM_IT_CHARSET,
     BSTRERAM_IT_USER, etc).
  b) iterator determining order in which global objects' metadata will be
     stored (BSTREAM_IT_GLOBAL).
  c) iterator determining order in which per-database objects' metadata will
     be stored (BSTREAM_IT_PERDB).

  The iterators are represented by pointers to the corresponding @c it_*
  counter. In case of iterators a) and b) they are both represented by a pointer
  to @c it_global. However, for iterator a), the value of @c it_global
  is positive and indicates the type of the object over which we should iterate.
  For iterator b) the value if @c it_global is negative.

  Note: Since always only one instance of an iterator is used, we do not bother
  to dynamically allocate/free iterators. Instead, we always use a single, 
  static instance of each iterator represented by an appropriate pointer to one
  of the @c it_* variables.

*********************************************************************/


int  it_global;
int  it_db;
int  it_perdb;

/** Iterate over databases */
item_info* db_iterator_next()
{
 DBUG_ASSERT(it_db >= 0);

 if ((unsigned)it_db >= db_count)
   return NULL;

 return (item_info*)&dbs[it_db++];
}

/** 
  Iterate over global objects of type @c it_global.

  This function is used for iterator represented by @c it_global when its
  value is positive. The value of @c it_global indicates the type of global
  objects over which we shuld iterate. A separate counter @c it_db is used to
  keep track of iterator's position.
*/
item_info* other_iterator_next()
{
  uint pos;
  item_info *item= NULL;

  DBUG_ASSERT(it_global >= 0);

  /* there are no more global objects to look at - return NULL */
  if ((unsigned)it_db >= global_count)
    return NULL;

  /* 
    Search objects_global[] array for next global objects whose type equals
    it_global.
  */
  for(pos=it_db++; pos < global_count; ++pos)
    if (objects_global[pos].type == (item_type)it_global)
      item= &objects_global[pos];

  if (!item)
    return NULL;

  /*
    For objects of type CHARSET and USER bstream library expects only the name
    of the object. For other objects we should return a pointer to item_info
    structure describing it.
  */ 
  switch (item->type) {
  case BSTREAM_IT_CHARSET:
  case BSTREAM_IT_USER: return (item_info*)&(item->name);
  default: return item;
  }
}

/**
  Iterate over all global objecs.

  This function is used for iterator represented by @c it_global when its
  value is negative. In that case we should iterate over all global objects 
  from the catalogue. We start with databases which are enumerated using 
  @c it_db counter. Then we enumerate other global objects from 
  @c objects_global[] array using @c it_global to track iterator's position.
*/
item_info* global_iterator_next()
{
  unsigned pos= it_db;

  DBUG_ASSERT(it_global < 0);

  /*
    Return next database if not all of them enumerated yet.
  */ 
  if (pos < db_count)
  {
    ++it_db;
    return (item_info*)&dbs[pos];
  }

  /*
    Otherwise iterate over all global objects from objects_global[] array
    using the value of it_global for keeping the current position within
    the array. Since we want it_global to be negative, we use convention that
    values -1,-2,-3,... correspond to positions 0,1,2,...
  */ 
  pos= -it_global-1;

  /*
    Return NULL if all global objects have been already enumerated.
  */ 
  if (pos >= global_count)
    return NULL;

  --it_global;

  return &objects_global[pos];
}

/**
  Iterate over all objects belonging to a given database.

  First all tables belonging to that database are returned, then other
  per-database objects as defined by @c objects_perdb[] array. Counter
  @c it_perdb indicates current postition of the iterator. First table_count
  values correspond to the tables while the following values indicate position
  within @c objects_perdb[] array.
*/
dbitem_info* perdb_iterator_next(struct st_bstream_db_info *db)
{
  uint pos;
  static dbitem_info item_buf;

  DBUG_ASSERT(it_perdb >= 0);

  /* Find next table beloging to this database */

  for (; (unsigned)it_perdb < table_count; ++it_perdb)
    if (tables[it_db].base.db == db)
      return (dbitem_info*)&(tables[it_perdb++]);

  /* Otherwise return next non-table object */
  pos= it_perdb - table_count;

  if (pos >= perdb_count)
    return NULL;

  ++it_perdb;

  item_buf.db= db;
  item_buf.base= objects_perdb[pos];

  return &item_buf;
}


/**********************************************************************

  Implementation of catalogue iterator callbacks.

*********************************************************************/

/**
  Return pointer to given iterator instance.

  The type of the iterator is indicated by @a type parameter:
  
  BSTREAM_IT_CHARSET, BSTREAM_IT_USER, etc - an iterator over a given type
     of global objects.

  BSTREAM_IT_GLOBAL - an iterator determining order in which global objects' 
     metadata will be stored.

  BSTREAM_IT_PERDB - an iterator determining order in which per-database 
     objects' metadata will be stored.

  @return Pointer to the @c it_* counter representing given iterator or NULL
  in case of error.
*/
void* bcat_iterator_get(struct st_bstream_image_header
                        *catalogue  __attribute__ ((__unused__)),
                        unsigned int type)
{
  STEP_AND_CHECK_COUNT(callback_counts.iterator_get[type], return NULL);

  switch (type) {

  case BSTREAM_IT_GLOBAL:
    it_db= 0;
    it_global= -1;
    return &it_global;

  case BSTREAM_IT_PERDB:
    it_db= 0;
    it_perdb= 0;
    return &it_perdb;

  case BSTREAM_IT_DB:
    it_db= 0;
    return &it_db;

  default:
    it_global= type;
    it_db= 0;
    return &it_global;
  }
}

/**
  Free resources used by iterator.

  Since we do not allocate any resources, here we only step the appropriate
  callback counter. Later these conuters can be used to verify that the number
  of @c bcat_iterator_free() calls matches that of @c bcat_iterator_get(). This
  is done in @check_iterators() helper function.
*/
void  bcat_iterator_free(struct st_bstream_image_header
                         *catalogue  __attribute__ ((__unused__)),
                         void *iter)
{
  uint UNINIT_VAR(type);

  if (iter == &it_global)
    type= it_global < 0 ? BSTREAM_IT_GLOBAL : it_global;
  else if (iter == &it_perdb)
    type= BSTREAM_IT_PERDB;
  else if (iter == &it_db)
    type= BSTREAM_IT_DB;

  ++(callback_counts.iterator_free[type]);
}

int opt_empty_catalog= 0;

/**
  Return next object from a given iterator.

  Call appropriate @c *_iterator_next() function to return next object pointed
  by the iterator.
*/
struct st_bstream_item_info*
bcat_iterator_next(struct st_bstream_image_header
                   *catalogue __attribute__ ((__unused__)), void *iter)
{
  dbitem_info *info;

  if (opt_empty_catalog)
     return NULL;

  if (iter == &it_global && it_global < 0)
    return global_iterator_next();

  if (iter == &it_global)
    return other_iterator_next();

  if (iter == &it_db)
    return db_iterator_next();

  /*
    If iter == &it_perdb we want to enumerate all per-database objects. We
    iterate over all databases using it_db counter and for each database use
    perdb_iterator_next() until it returns NULL indicating that there are
    no more objects in that database.
  */

  if (iter == &it_perdb)
  {
    DBUG_ASSERT(it_db >= 0);

    if ((unsigned)it_db >= db_count)
      return NULL;

    info= perdb_iterator_next(&dbs[it_db]);

    if (!info)
    {
      ++it_db;
      it_perdb=0;
      return bcat_iterator_next(NULL, iter);
    }

    return (item_info*)info;
  }

  return NULL;
}

/**
  Return iterator enumerating all objects belonging to a given database.

  It is not important what pointer is returned as long as it is not NULL.
  The enumeration is always performed by @c perdb_iterator_next() function.
*/
void* bcat_db_iterator_get(struct st_bstream_image_header
                           *catalogue  __attribute__ ((__unused__)),
                           struct st_bstream_db_info *db)
{
  STEP_AND_CHECK_COUNT(callback_counts.db_iterator_get[db->base.pos],
                       return NULL);
  it_perdb=0;
  return &it_perdb;
}

/** Return next object from database iterator */
struct st_bstream_dbitem_info*
bcat_db_iterator_next(struct st_bstream_image_header
                      *catalogue  __attribute__ ((__unused__)),
                      struct st_bstream_db_info *db,
                      void *iter  __attribute__ ((__unused__)))
{
  return perdb_iterator_next(db);
}

/** Free database iterator resources */
void  bcat_db_iterator_free(struct st_bstream_image_header
                            *catalogue  __attribute__ ((__unused__)),
                            struct st_bstream_db_info *db,
                            void *iter  __attribute__ ((__unused__)))
{
  ++(callback_counts.db_iterator_free[db->base.pos]);
}

/**********************************************************************

  Functions for getting object's metadata.

*********************************************************************/

int bcat_get_item_create_query(struct st_bstream_image_header
                               *catalogue  __attribute__ ((__unused__)),
                               struct st_bstream_item_info *item,
                               bstream_blob *stmt)
{
  static char buf[]= "serialization of xxxxxxxxxxxxxxxx";
  int len;

  STEP_AND_CHECK_COUNT(callback_counts.get_serialization[item->type], 
                       return BSTREAM_ERROR);

  len= snprintf(buf+17, 16, "%s", item->name.begin);

  stmt->begin= (byte*) buf;
  stmt->end= stmt->begin + 17 + len;

  return BSTREAM_OK;
}

int bcat_get_item_create_data(struct st_bstream_image_header
                              *catalogue  __attribute__ ((__unused__)),
                              struct st_bstream_item_info *item,
                              bstream_blob *data)
{
  static char buf[]= "metadata for xxxxxxxxxxxxxxxx";
  int len;

  STEP_AND_CHECK_COUNT(callback_counts.get_metadata[item->type], 
                       return BSTREAM_ERROR);

  len= snprintf(buf+13, 16, "%s", item->name.begin);

  data->begin= (byte*) buf;
  data->end= data->begin + 13 + len;

  return BSTREAM_OK;
}

/*********************************************************************

  Other catalogue callback functions.

*********************************************************************/

int bcat_create_item(struct st_bstream_image_header
                     *catalogue  __attribute__ ((__unused__)),
                     struct st_bstream_item_info *item,
                     bstream_blob create_stmt  __attribute__ ((__unused__)),
                     bstream_blob other_meta_data  __attribute__ ((__unused__)))
{
  STEP_AND_CHECK_COUNT(callback_counts.create_item[item->type], 
                       return BSTREAM_ERROR);
  return BSTREAM_OK;
}

int bcat_reset(struct st_bstream_image_header
               *catalogue __attribute__ ((__unused__)))
{
  STEP_AND_CHECK_COUNT(callback_counts.cat_reset, return BSTREAM_ERROR);
  return BSTREAM_OK;
}

int bcat_close(struct st_bstream_image_header
               *catalogue  __attribute__ ((__unused__)))
{
  STEP_AND_CHECK_COUNT(callback_counts.cat_close, return BSTREAM_ERROR);
  return BSTREAM_OK;
}

int bcat_add_item(struct st_bstream_image_header
                  *catalogue  __attribute__ ((__unused__)),
                  struct st_bstream_item_info *item)
{
  STEP_AND_CHECK_COUNT(callback_counts.add_item[item->type], 
                       return BSTREAM_ERROR);
  return BSTREAM_OK;
}


/*********************************************************************

  Memory allocation callbacks.

*********************************************************************/

byte* bstream_alloc(unsigned long int size)
{
  STEP_AND_CHECK_COUNT(callback_counts.alloc, return NULL);
  return malloc(size);
}

void bstream_free(byte *ptr)
{
  ++callback_counts.free;
  free(ptr);
}
