/** @file

  Implementation of low-level I/O file stream for testing backup stream
  library.
*/

#include <my_global.h>
#include <my_sys.h>

#include "stream_test.h"

/** I/O block size to be used when creating new stream. */

size_t stream_block_size= 1024;

static int stream_write(void *stream, bstream_blob *buf, bstream_blob envelope);
static int stream_read(void *stream, bstream_blob *buf, bstream_blob);
static int stream_forward(void *stream, unsigned long int *offset);

/**
  Structure containing I/O callback counters.

  Whenever one of the low-level I/O functions is called, the corresponding
  counter will be increased.

  Setting counter to a negative value will cause the corresponding function
  to report ERROR or EOS when counter reaches the value of 0.
*/
struct st_io_counts io_counts;

static const char* files_to_delete[128];
static unsigned int files_to_delete_count= 0;

extern void (*test_exit_hook)();


void delete_files_at_exit()
{
  unsigned i;

  for (i=0; i < files_to_delete_count && i < 128; ++i)
  {
    DBUG_ASSERT(files_to_delete[i]);
    my_delete(files_to_delete[i], MYF(0));
  }
}


/** Reset all callback counters */

void reset_io_counts()
{
  bzero(&io_counts, sizeof(io_counts));
}


/**
  Open test stream for writing.

  Data will be written to @a path file. If that file already exists, it will be
  overwritten. The stream will use I/O blocks of size specified by the current
  value of @c stream_block_size.
*/

int backup_stream_open_wr(test_backup_stream *stream, const char *path)
{
  DBUG_ASSERT(stream);

  /* open file for writing */
  int fd= my_open(path, O_WRONLY|O_CREAT|O_TRUNC, MYF(0));
  if (fd < 0)
    return BSTREAM_ERROR;

  /* Register file path for deletion at the end of the test */
  files_to_delete[files_to_delete_count++]= path;
  test_exit_hook= delete_files_at_exit;

  stream->m_fd= fd;
  stream->base.stream.write= stream_write;
  stream->base.stream.forward= stream_forward;

  return bstream_open_wr(&(stream->base), stream_block_size, 0);
}


/**
  Open test stream for reading.

  The stream will get bytes from @a path file.
*/

int backup_stream_open_rd(test_backup_stream *stream, const char *path)
{
  DBUG_ASSERT(stream);

  /* open file for reading */
  int fd= my_open(path, O_RDONLY, MYF(0));
  if (fd < 0)
    return BSTREAM_ERROR;

  stream->m_fd= fd;
  stream->base.stream.read=  stream_read;
  stream->base.stream.forward= stream_forward;

  return bstream_open_rd(&(stream->base), 0);
}


/** Close a test stream. */

int backup_stream_close(test_backup_stream *stream)
{
  DBUG_ASSERT(stream);
  int fd= stream->m_fd;

  (void)bstream_close(&(stream->base));

  /* close the stream */
  if (fd >= 0)
    my_close(fd, MYF(0));

  return BSTREAM_OK;
}


/**
  Helper function which writes a backup image to given file.

  The caller of this function should implement the @c bcat_ callbacks which
  will determine the set of objects to be included in the image and their
  metadata. The image will contain no table data chunks, only preamble and
  summary section.

  @returns 0 on success, non-zero otherwise.
*/

int write_image(image_header *hdr, const char *path)
{
  test_backup_stream stream;
  int ret1, ret2;

  ret1= backup_stream_open_wr(&stream, path);
  if (ret1)
    return OPEN_ERROR;

  ret1= bstream_wr_preamble((backup_stream*)&stream, hdr);
  ret2= bstream_wr_summary((backup_stream*)&stream, hdr);

  backup_stream_close(&stream);

  return ret1 || ret2;
}


/**
  Helper fuction which reads backup image from a file.

  Any table data chunks present in the image are ignored.

  @returns 0 on success, non-zero otherwise.
*/

int read_image(image_header *hdr, const char *path)
{
  test_backup_stream stream;
  table_data_chunk   dchunk;
  int ret1, UNINIT_VAR(ret2);

  ret1= backup_stream_open_rd(&stream, path);
  if (ret1)
    return OPEN_ERROR;

  ret1= bstream_rd_preamble((backup_stream*)&stream, hdr);

  if (ret1 == BSTREAM_ERROR)
    goto end;

  /* Read and skip table data chunks */
  do {
    ret2= bstream_rd_data_chunk((backup_stream*)&stream, &dchunk);
  } while (ret2 == BSTREAM_OK);

  /*
    Now, there should be the summary block at the end of the image.
  */
  if (ret2 != BSTREAM_EOC)
    goto end;

  /* Read the summary block and check that stream ends after it */

  ret2= bstream_rd_summary((backup_stream*)&stream, hdr);
  ret2= (ret2 != BSTREAM_EOS) ? BSTREAM_ERROR : BSTREAM_OK;

end:

  backup_stream_close(&stream);

  return ret1 || ret2;
}


/*******************************************************************

  Implementation of low-level I/O operations.

*******************************************************************/


/** Low-level write function */

static
int stream_write(void *stream, bstream_blob *buf,
                 bstream_blob envelope __attribute__ ((__unused__)))
{
  int fd;
  int res;
  size_t howmuch;

  DBUG_ASSERT(stream);
  DBUG_ASSERT(buf);

  test_backup_stream *s= stream;

  fd= s->m_fd;

  if (fd < 0)
    return BSTREAM_ERROR;

  if (!buf->begin || buf->begin == buf->end)
    return BSTREAM_OK;

  DBUG_ASSERT(buf->end);

  howmuch = buf->end - buf->begin;
  res= my_write(fd, buf->begin, howmuch,
                MYF(MY_NABP) /* error if not all bytes written */ );

  if (++io_counts.write == 0)
    goto error;

  if (res)
    goto error;

  buf->begin= buf->end;
  io_counts.write_bytes += howmuch;

  return BSTREAM_OK;

 error:
 
   my_close(fd, MYF(0));
   s->m_fd= -1;
   return BSTREAM_ERROR;   
}


/** Low-level read function */
static
int stream_read(void *stream, bstream_blob *buf,
                bstream_blob envelope  __attribute__ ((__unused__)))
{
  int fd;
  size_t howmuch;

  DBUG_ASSERT(stream);
  DBUG_ASSERT(buf);

  test_backup_stream *s= stream;

  fd= s->m_fd;

  DBUG_ASSERT(fd >= 0);

  if (!buf->begin || buf->begin == buf->end)
    return BSTREAM_OK;

  DBUG_ASSERT(buf->end);

  howmuch= buf->end - buf->begin;
  howmuch= my_read(fd, buf->begin, howmuch, MYF(0));

  if (0 == ++io_counts.read)
    goto error;

  if (howmuch == (size_t) -1)
    goto error;

  if (howmuch == 0)
    return BSTREAM_EOS;

  buf->begin += howmuch;

  io_counts.read_bytes += howmuch;

  return BSTREAM_OK;

 error:
 
   my_close(fd, MYF(0));
   s->m_fd= -1;
   return BSTREAM_ERROR;   
}


/** Low-level forward function */
static
int stream_forward(void *stream, unsigned long int *offset)
{
  int fd;
  size_t howmuch;

  DBUG_ASSERT(stream);
  DBUG_ASSERT(offset);

  test_backup_stream *s= stream;

  fd= s->m_fd;

  DBUG_ASSERT(fd >= 0);

  if (*offset == 0)
    return BSTREAM_OK;

  howmuch= *offset;

  lseek(fd, howmuch, SEEK_CUR);
  *offset= 0;

  return BSTREAM_OK;
}
