/* Copyright (C) 2009 Sun Microsystems, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA */

#include <my_global.h>
#include <my_sys.h>
#include <tap.h>
#include <string.h>
#if defined(__WIN__)
#include <direct.h>
#define chdir _chdir
#define rmdir _rmdir
#endif

#define DST_SIZE 80

static char   *concat_buf= NULL;
static size_t concat_size= 0;


/**
  Concatenate two strings.

  @param[in]    str1            string 1
  @param[in]    str2            string 2

  @return       string in allocated storage
*/

static char*
concat_str(const char *str1, const char *str2)
{
  size_t len1;
  size_t len2;
  size_t sum;

  len1= strlen(str1);
  len2= strlen(str2);
  sum= len1 + len2 + 1; /* Include terminating '\0'. */

  if (sum > concat_size)
  {
    concat_buf= my_realloc(concat_buf, sum, MYF(MY_FAE | MY_ALLOW_ZERO_PTR));
    if (!concat_buf)
      exit(3); /* purecov: inspected */
    concat_size= sum;
  }
  memcpy(concat_buf, str1, len1);
  memcpy(concat_buf + len1, str2, len2 + 1); /* Include terminating '\0'. */
  return concat_buf;
}


/**
  Do a test for a path name.

  @param[in]    path            input path name
  @param[out]   expect          expected output path name
*/

static void
check_result(const char *path1, const char *path2,
             const char *path3, const char *result, const char *expect)
{
  const char    *exp_src;
  char          *exp_dst;
  char          *exp_end;
  char          exp[DST_SIZE];

  /* Convert FN_LIBCHAR to slash for comparable results. */
  for (exp_src= expect, exp_dst= exp, exp_end= exp + DST_SIZE - 1;
       *exp_src && (exp_dst < exp_end);
       exp_src++, exp_dst++)
  {
    if (*exp_src == '/')
      *exp_dst= FN_LIBCHAR;
    else
      *exp_dst= *exp_src;
  }
  *exp_dst= '\0';

  /* Check and print result. */
  ok(!strcmp(result, exp),
     "input: '%s' '%s' '%s'  result: '%s'  expected: '%s'",
     path1 ? path1 : "[NullS]",
     path2 ? path2 : "[NullS]",
     path3 ? path3 : "[NullS]",
     result, exp);
}


/**
  Do a test for a path name.

  @param[in]    path1           input path name
  @param[in]    path2           input path name, optional NullS
  @param[in]    path3           input path name, optional NullS
  @param[out]   expect          expected output path name
*/

static void
do_multi_test(const char *path1, const char *path2,
              const char *path3, const char *expect)
{
  size_t dst_len;
  char   dst[DST_SIZE];

  /* Cleanup the path. That's the core test. */
  dst_len= safe_cleanup_cat_path(dst, DST_SIZE, path1, path2, path3, NullS);

  /* If path too long, dst_len == 0 but dst has partial result. Clear it. */
  if (!dst_len)
    dst[0]= '\0';

  /* Check and print result. */
  check_result(path1, path2, path3, dst, expect);
}


#define do_test(path, expect)   do_multi_test(path, NullS, NullS, expect)


int
main(void)
{
  int chdir_res;

  MY_INIT("safe_cleanup_cat_path-t");

  /* Set HOME for ~/ test. Trailing slash is for coverage test. */
  putenv((char*) "HOME=/home/unittest/");

  /* Create a directory with unique name and change to it. */
  if (my_mkdir("safe_cleanup_cat_path_dir", 0777, MYF(MY_WME)))
    goto end; /* purecov: inspected */
  if (my_setwd("safe_cleanup_cat_path_dir", MYF(MY_WME)))
    goto end1; /* purecov: inspected */

  /* Create some test directories. */
  if (my_mkdir("dir1", 0777, MYF(MY_WME)) ||
      my_mkdir("dir1/dir2", 0777, MYF(MY_WME)) ||
      my_mkdir("dir1/dir2/dir3", 0777, MYF(MY_WME)))
    goto end2;

  /* Use "file" for a non-existent file. */

  plan(70);

  do_test("", "");

  do_test("file", "file");
  do_test("file.file", "file.file");
  do_test("file.file.file.file.file.file.file.file.file.file.file",
          "file.file.file.file.file.file.file.file.file.file.file");

  do_test("/", "/");
  do_test("///", "/");
  do_test("/file", "/file");
#ifdef FN_NETWORK_DRIVES
  do_test("//node/pipe/file///", "//node/pipe/file");
  do_test("//node/file///", "//node/file");
  do_test("//./pipe/file", "//./pipe/file");
  do_test("//./file///", "//./file");
  do_test("//file///", "//file");
  do_test("///file///", "//file");
#else
  do_test("//dir1/dir2/file///", "/dir1/dir2/file");
  do_test("//dir/file///", "/dir/file");
  do_test("//./file///", "/file");
  do_test("//file///", "/file");
  do_test("///file///", "/file");
#endif

  do_test("dir1/file", "dir1/file");
  do_test("dir1///file", "dir1/file");
  do_test("dir1/dir2/", "dir1/dir2");
  do_test("dir1/dir2///", "dir1/dir2");

  do_test("file/file", "file/file");
  do_test("file///file", "file/file");
  do_test("file/file/", "file/file");
  do_test("file/file///", "file/file");

  do_test(".", ".");
  do_test("./", ".");
  do_test("./.", ".");
  do_test("././", ".");
  do_test("./dir1", "dir1");
  do_test("./file", "file");
  do_test("././dir1", "dir1");
  do_test("././file", "file");

  do_test("dir1/./file", "dir1/file");
  do_test("dir1/././file", "dir1/file");
  do_test("dir1/.", "dir1");
  do_test("dir1/./.", "dir1");

#if defined(__WIN__)
  /* Windows does path cleanup differently. */
  do_test("file/./file", "file/file");
  do_test("file/././file", "file/file");
  do_test("file/.", "file");
  do_test("file/./.", "file");
#else
  do_test("file/./file", "file/./file");
  do_test("file/././file", "file/././file");
  do_test("file/.", "file/.");
  do_test("file/./.", "file/./.");
#endif

  do_test("..", "..");
  do_test("../..", "../..");
  do_test("../../", "../..");
  do_test("../../file", "../../file");
  do_test("/../..", "/");
  do_test("/../file", "/file");
  do_test("/../../file", "/file");
  do_test("/../../", "/");

  do_test("dir1/../file", "file");
  do_test("dir1/dir2/dir3/../../file", "dir1/file");
  do_test("dir1/dir2/..", "dir1");
  do_test("dir1/..", ".");

  do_test(".././file///", "../file");
  do_test("..///./file///", "../file");

#if defined(__WIN__)
  /* Windows does path cleanup differently. */
  do_test("file/../file", "file");
  do_test("dir1/dir2/file/../../file", "dir1/file");
  do_test("file/file/..", "file");
  do_test("file/..", ".");
  do_test("dir1/../dir2/non-existent-file.txt/../existent-file.txt",
          "dir2/existent-file.txt");
  do_test("dir1/../dir2/non-existent-file.txt/./.././existent-file.txt",
          "dir2/existent-file.txt");
#else
  do_test("file/../file", "file/../file");
  do_test("dir1/dir2/file/../../file", "dir1/dir2/file/../../file");
  do_test("file/file/..", "file/file/..");
  do_test("file/..", "file/..");
  do_test("dir1/../dir2/non-existent-file.txt/../existent-file.txt",
          "dir2/non-existent-file.txt/../existent-file.txt");
  do_test("dir1/../dir2/non-existent-file.txt/./.././existent-file.txt",
          "dir2/non-existent-file.txt/./.././existent-file.txt");
#endif

  do_test("~/file", concat_str(home_dir, "/file"));
  do_test("dir1/~/file", "dir1/~/file");
  /* Exceed DST_SIZE. Result should be empty string. */
  do_test("file.file.file.file.file.file.file.file.file.file.file"
          "file.file.file.file.file.file.file.file.file.file.file"
          "file.file.file.file.file.file.file.file.file.file.file",
          "");

#ifdef HAVE_READLINK
  {
    int rc;
    (void) my_delete("symlink1", MYF(0));
    rc= my_symlink("dir1/dir2", "symlink1", MYF(MY_WME));
    do_test("symlink1/../../dir1/././dir2/../../file",
            "symlink1/../../file");
    (void) my_delete("symlink1", MYF(0));
  }
#else
  {
    do_test("dummy_test_to_please_the_test_counter",
            "dummy_test_to_please_the_test_counter");
  }
#endif
  do_test("dir1/dir2/../../dir1/././dir2/../../file",
          "file");

  do_multi_test(NullS,   NullS,     NullS,  "");
  do_multi_test("",      NullS,     NullS,  "");
  do_multi_test("",      "",        NullS,  "");
  do_multi_test("",      "/file",   NullS,  "/file");
  do_multi_test("dir1/", "/dir2",   NullS,  "dir1/dir2");
  do_multi_test("dir1",  "dir2",    NullS,  "dir1/dir2");
  do_multi_test("dir1",  "",        "dir2", "dir1/dir2");
  do_multi_test("dir1/", "../file", NullS,  "file");
  do_multi_test("",      "file",    NullS,  "file");

end2:
  my_free(concat_buf, MYF(MY_ALLOW_ZERO_PTR));
  rmdir("dir1/dir2/dir3");
  rmdir("dir1/dir2");
  rmdir("dir1");
  chdir_res= chdir("..");
end1:
  rmdir("safe_cleanup_cat_path_dir");
end:
  return exit_status();
}
